(*-----------#pc-version-----------------------------------------------------*)
{$N+,O+,F+,E+}
(* 
unit numeric;
interface
uses tt1,tt2, tt3,tt4,tt5,prior,mon;

const
   tiny=1.0e-20;

type
  fvec = array[1..maxvertex] of real;
  fivec = array[1..maxvertex] of integer;

var
   k,j,imax,i: integer;
   som,dum,big: real;
   vv: fvec;

function trace(var mat: covarray; dim: vertextype): real;
procedure inverse(var a: covarray; var y: covarray; dim: vertextype);
function determ(mat: covarray; dim: integer): real;
procedure sqmatmult(var x,y,result: covarray; dim: vertextype);
function brent(param: integer;
               ax,bx,cx,tol: real;
               var xmin: real;
               var rejfile: text;
               var paramrecalcind:calclistindarrt;
	       var covtermindex:hcovheadindext): real;

function FML(var sampcovin, impcovin: covarray): real;
function neglogpost(param: integer; x: real;
                    var paramrecalcind:calclistindarrt;
	            var covtermindex:hcovheadindext): real;


implementation
uses io,sys,calctet,gib,util;

*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "mon.h"
#include "calctet.h"
#include "calctet1.h"
#include "io.h"
#include "sys.h"
#include "rsutil.h"
#include "cspecial.h"
#include "prior.h"
#include "util.h"
#include "gib.h"
#include "recalc.h"
#include "eqs.h"
#include "numeric.h"


(*-----------#unix-version-----------------------------------------------------*)

function trace(*var mat: covarray; n: vertextype): real;  *);
var i: vertextype;
   tr: real;
begin
 tr:= 0.0;
 for i:= 1 to dim do
  tr:= tr + mat^[i,i];
 trace:= tr;
end;


(*NR:needed for matrix inverse and determinant*)
procedure ludcmp(var a: covarray; n: integer;
       var indx: fivec; var d: real; var ok: boolean); 
label 999;
var
   k,j,imax,i: integer;
   som,dum,big: real;
   vv: fvec;
begin
   ok := true;
   d := 1.0;
   for i := 1 to n do begin
      big := 0.0;
      for j := 1 to n do if (abs(a^[i,j]) > big) then big := abs(a^[i,j]);
      if (big = 0.0) then begin
         writeln('pause in LUDCMP - singular matrix'); 
         d := 10000000;
         ok := false;
         goto 999;
      end;
      vv[i] := 1.0/big
   end;
   for j := 1 to n do begin
      if (j > 1) then begin
         for i := 1 to j-1 do begin
            som := a^[i,j];
            if (i > 1) then begin
               for k := 1 to i-1 do begin
                  som := som-a^[i,k]*a^[k,j]
               end;
               a^[i,j] := som
            end
         end
      end;
      big := 0.0;
      for i := j to n do begin
         som := a^[i,j];
         if (j > 1) then begin
            for k := 1 to j-1 do begin
               som := som-a^[i,k]*a^[k,j]
            end;
            a^[i,j] := som
         end;
         dum := vv[i]*abs(som);
         if (dum > big) then begin
            big := dum;
            imax := i
         end
      end;
      if (j <> imax) then begin
         for k := 1 to n do begin
            dum := a^[imax,k];
            a^[imax,k] := a^[j,k];
            a^[j,k] := dum
         end;
         d := -d;
         vv[imax] := vv[j]
      end;
      indx[j] := imax;
      if (j <> n) then begin
         if (a^[j,j] = 0.0) then a^[j,j] := tiny;
         dum := 1.0/a^[j,j];
         for i := j+1 to n do begin
            a^[i,j] := a^[i,j]*dum
         end
      end
   end;
   if (a^[n,n] = 0.0) then a^[n,n] := tiny;
999: end;


procedure lubksb(var a: covarray; n: integer; var indx: fivec; var b: fvec);
(*NR: needed for matrix inverse*)
var
   j,ip,ii,i: integer;
   som: real;
begin
   ii := 0;
   for i := 1 to n do begin
      ip := indx[i];
      som := b[ip];
      b[ip] := b[i];
      if  (ii <> 0) then begin
         for j := ii to i-1 do begin
            som := som-a^[i,j]*b[j]
         end
      end else if (som <> 0.0) then begin
         ii := i
      end;
      b[i] := som
   end;
   for i := n downto 1 do begin
      som := b[i];
      if (i < n) then begin
         for j := i+1 to n do begin
            som := som-a^[i,j]*b[j]
         end
      end;
      b[i] := som/a^[i,i]
   end
end;

 (* From Numerical Recipes in Pascal, 
     Press, et. al, 1989, p. 46
     This destroys the a matrix passed in,
     and puts the answer in y matrix. *)
procedure inverse(*var a: covarray; var y: covarray; dim: vertextype; var ok: boolean *);
(*calculates matrix inverse*)
label 999;
var i,j: integer;
    vecy: fvec;
    indx: fivec;
    d: real;

begin 
      ok := true;
      for i := 1 to dim
          do for j := 1 to dim
                 do begin y^[i,j] := 0;
                          y^[i,i] := 1;
                    end;
      ludcmp(a,dim,indx,d,ok);
      if not ok then goto 999;
      for j := 1 to dim
          do begin for i := 1 to dim
                       do begin vecy[i]:=0;
                                vecy[j]:=1;
                          end;
                   lubksb(a,dim,indx,vecy);
                   for i:= 1 to dim
                       do y^[i,j] := vecy[i];
            end;
999: end;  (* Inverse *)


 (* From Numerical Recipes in Pascal, 
     Press, et. al, 1989, p. 47
     This destroys the mat passed in, so it
     cannot be a var parameter. *)
function determ(*mat: covarray; dim: integer; var ok: boolean): real*);
  var j: integer;
      indx: fivec;
      d: real;
begin
  ludcmp(mat,dim,indx,d,ok);
  for j:= 1 to dim do 
     d:= d*mat^[j,j];
  determ:= d;
end;

procedure sqmatmult(*var x,y,result: covarray; dim: vertextype*);
  var i,j,k: vertextype;
  begin
    for i:= 1 to dim do
     for j:= 1 to dim do
      begin
        result^[i,j]:= 0;
        for k:= 1 to dim do
          result^[i,j]:= result^[i,j] + (x^[i,k] * y^[k,j]);
      end;
  end;

(*******************************************)
(*  The Brent optimizer in one dimension 
    From Numerical Recipes in Pascal, 
    Press, et. al, 1989, p. 320
    The function being minimized is called neglogpost,
    which is the negative of the log(likelihood) if the
    prior is flat, and the negative of the 
    log(prior X likelihood) otherwise.
    the . - RS *)

function brent(*param: integer;
               ax,bx,cx,tol: real;
               var xmin: real;
               rejfile: text;
               var paramrecalcind:calclistindarrt;
	       var covtermindex:hcovheadindext): real*);

label 99;
const
    itmax=100;
    cgold=0.3819660;
    zeps= 1.0e-10;
var
    a,b,d,e,etemp: real;
    fu,fv,fw,fx: real;
    iter,numchars: integer;
    p,q,r,tol1,tol2: real;
    u,v,w,x,xm: real;

function sign(a,b: real): real;
begin
  if b >= 0.0 then sign := abs(a) else sign := -abs(a)
end;

begin
  if ax < cx then a := ax else a := cx;
  if ax > cx then b := ax else b := cx;
  v := bx;
  w := v;
  x := v;
  e := 0.0;
  fx := neglogpost(param,x,paramrecalcind, covtermindex);
  fv := fx;
  fw := fx;
  for iter := 1 to itmax do begin
     xm := 0.5 * (a+b);
     tol1 := tol*abs(x)+zeps;
     tol2 := 2.0*tol1;
     if abs(x-xm) <= tol2-0.5*(b-a) then goto 99;
     if abs(e) > tol1 then begin
        r := (x-w)*(fx-fv);
        q := (x-v)*(fx-fw);
        p := (x-v)*q-(x-w)*r;
        q := 2.0*(q-r);
        if q > 0.0 then p := -p;
        q := abs(q);
        etemp := e;
        e := d;
        if (abs(p) >= abs(0.5*q*etemp)) or (p <= q*(a-x))
           or (p >= q*(b-x)) then begin
           if x >= xm then
              e := a-x
           else
              e := b-x;
           d := cgold*e
        end
        else begin
              d := p/q;
              u := x+d;
              if (u-a < tol2) or (b-u < tol2) then
                 d := sign(tol1,xm-x)
        end
     end
     else begin
        if x >= xm then
           e := a-x
        else
           e := b-x;
        d := cgold*e
     end;
     if abs(d) >= tol1 then
        u := x+d
     else
        u := x+sign(tol1,d);
        fu := neglogpost(param,u,paramrecalcind,covtermindex);
     if fu <= fx then begin
        if u >= x then a := x else b := x;
        v := w;
        fv := fw;
        w := x;
        fw := fx;
        x := u;
        fx := fu
     end
     else begin
        if u < x then a := u else b := u;
        if (fu <= fw) or (w=x) then begin
           v := w;
           fv := fw;
           w := u;
           fw := fu
        end
        else if (fu <= fv) or (v=x) or (v=w) then begin
           v := u;
           fv := fu
        end
     end
  end;
  twrite(rejfile,'BRENT - too many iters |');
  writeparm(param, rejfile, numchars); 
  twriteeoln(rejfile);
99:
   xmin := x;
   brent := fx
end;
(***  end of brent optimizer *)

(********************************************************)
(** functions for calculating the posterior density **)

  
function getparmval(param: integer): real;
  var cause, effect: vertextype;
  begin
    cause:= priormean[param].cause;
    effect:= priormean[param].effect;
    if param <= max 
      then getparmval:= exogd[cause].parm2
    else getparmval:= parm^[cause,effect];
  end;

    (* This takes a vector of the prior means for the
       free parameters minus the current estimates, where 
       x is the estimate under consideration, and pre and 
       post multiplies it by the inverse of the prior cov matrix
       over the free parameters. *)
function negchi2(param: integer; x: real): real;
 var col,k,i,n: integer;
     answer: real;
     xvec,temp: array[1..maxparms] of real;
 begin
  answer:= 0;
   (* form xvec: the vector of current parameter estimates minus
      their prior means *)
  n:= 0;
  for i:= 1 to numparms do
   if priormean[i].free then 
     begin
       n:= n+1;
       if (i=param) then xvec[n]:= x - priormean[i].val
       else xvec[n]:= getparmval(i) - priormean[i].val;
     end;

  (* multiply xvec as a column vector times invprior, the
     inverse of the prior cov matrix over the parameter estimates *)
   for i:= 1 to numfree do
     temp[i]:= 0;
   for col:= 1 to numfree do
    for k:= 1 to numfree do
     temp[col]:= temp[col] + (xvec[k] * invprior^[k,col]);

  (* multiply the result of the previous step times
     xvec as a column vector *)
   for k:= 1 to numfree do
    answer:= answer + (temp[k] * xvec[k]);
   negchi2:= -1*answer;
 end;  (* negchi2 *)

(* Prior density of the data from the prior over the 
   model's parameters *)
function neglogprior(param: integer; x: real): real;
begin
 neglogprior:= -negchi2(param,x)/2;
 (* neglogprior:= -ln( pi2thep * (exp(negchi2(param,x)/2) / sqdetprior)); *)
end; (* neglogprior *)

(*********)
 (* necessary because taking the inverse and determinant etc. 
    change the matrices passed in, even though they are not 
    declared as var parameters *)
procedure copymat(*var source, copy: covarray*);
 var i,j: integer;
 begin
  for i:= 1 to numob do
    for j:=1 to numob do
      copy^[i,j]:= source^[i,j];
 end;

   (* This takes a sample covariance matrix and an
      implied covariance matrix and returns, up to a
      constant the Fitting function FML that LISREL
      etc. minimizes *)
function FML(*var sampcovin, impcovin: covarray): real*);
  label 99;
  var impldcov,impinv,sampcov,tempcov: covarray;
      term1,term2: real; ok: boolean;
begin
  tnewcov(impldcov,fatal);  (* check if enough space *)
  tnewcov(impinv,fatal); 
  tnewcov(sampcov,fatal); 
  tnewcov(tempcov,fatal); 
  if fatal in errors then goto 99;
  copymat(sampcovin, sampcov);
  copymat(impcovin, impldcov);
  inverse(impldcov,impinv,numob,ok);  (* puts inverse of impldcov in impinv *)
  sqmatmult(sampcov,impinv,tempcov,numob);  (* stores answer in tempcov *)
  term1:= trace(tempcov,numob);
  term2:= ln(determ(sampcov,numob,ok));
  FML:=  term1 - term2;
  dispose(impldcov);
  dispose(impinv);
  dispose(sampcov);
  dispose(tempcov);
99: end; (* loglikelihood *)


(* The parameters of the full model are stored in the global
   variables: parm^[cause,effect] and exogd[var].parm2,
   where the first are the edge coefficients and the second
   the error variances.  This function is written with 
   parameters so that the Brent off the shelf minimizer
   can iteratively call it to find the value of a particular 
   parameter such that the -log likelihood is minimal. Thus
   the x parameter is the value the minimizer is trying and
   the other parameters identify which parameter is being
   optimized . *)
function negloglike(param: integer; x: real;
               var paramrecalcind:calclistindarrt;
	       var covtermindex:hcovheadindext; var ok: boolean): real;
  label 99;
  var tparm: real;
      tempcov,gcov2: covarray;
      cause, effect: vertextype;
      loctime: double;
begin
  cause:= priormean[param].cause;
  effect:= priormean[param].effect;
  (* store global parameter that will be changed *)
  if param <= max then  (* parameter is an error variance *)
    begin 
      tparm:= exogd[cause].parm2;
      exogd[cause].parm2:= x;
    end
  else  (* parameter is an edge coefficient *)
    begin
      tparm:= parm^[cause,effect];
      parm^[cause,effect]:= x;
    end;
  
(*  tnewcov(gcov2,fatal);
  loctime:= tgettime;
  formtrue(gcov2, false, true);  
  imptime:= imptime + (tgettime-loctime);  *)


  loctime:= tgettime;
  recalcmatrix(gcov, exogd, parm,param,paramrecalcind,covtermindex,priormean);
  trtime:= trtime + (tgettime-loctime);

(*  checkthomas(gcov, gcov2); 
  dispose(gcov2);  *)

  loctime:= tgettime;
  tnewcov(tempcov,fatal);  (* check if enough space *)
  if fatal in errors then goto 99;
  for i:= 1 to numob do
    for j:=1 to numob do
     tempcov^[i,j]:= gcov^[i,j];
  inverse(tempcov,ginverse,numob,ok);  (* puts inverse of imp.covs in ginverse *)
  sqmatmult(cov,ginverse,tempcov,numob);  (* stores answer in tempcov *)
  (* restore global parameter *)
  if (cause = effect) then exogd[cause].parm2:= tparm 
  else parm^[cause,effect]:= tparm;
  negloglike:= (sampsize/2)*(ln(determ(gcov,numob,ok))
                            + trace(tempcov,numob));
  matoptime:= matoptime + (tgettime-loctime);
  likecalcnum:= likecalcnum+1;
  dispose(tempcov);
99: end; (* negloglikelihood *)

function neglogpost(*param: integer; x: real;
               var paramrecalcind:calclistindarrt;
	       var covtermindex:hcovheadindext): real*);
  var a,b: real; ok: boolean
  begin
   a:= negloglike(param,x,paramrecalcind,covtermindex,ok);
   if (not flatprior) then b:= neglogprior(param,x);
   if flatprior then
     neglogpost:= a
   else neglogpost:= a + b;
  end;

(** posterior density calculation **)

procedure gaussj(var a,b: covarray; n: integer);
var big,dum,pivinv: real; i,icol,irow,j,k,l,ll: integer;
    indxc,indxr,ipiv: fixers;
begin
   for j := 1 to n do ipiv[j] := 0;
   for i := 1 to n do 
      begin
        big := 0.0;
        for j := 1 to n do 
           if (ipiv[j] <> 1) then
              for k := 1 to n do
                 if (ipiv[k] = 0) then 
                    begin
                      if (abs(a^[j,k]) >= big) then
                         begin
                              big := abs(a^[j,k]);
                              irow := j;
                              icol := k;
                         end;
                    end
                 else if (ipiv[k] > 1 ) then twriteln(ofilvar,'Pause 1 in Gaussj - singular matrix');
        ipiv[icol] := ipiv[icol] + 1;
        if (irow <> icol) then
          begin
           for l := 1 to n do 
               begin
                  dum := a^[irow,l];
                  a^[irow,l] := a^[icol,l];
                  a^[icol,l] := dum;
               end;
           for l := 1 to n do 
               begin
                  dum := b^[irow,l];
                  b^[irow,l] := b^[icol,l];
                  b^[icol,l] := dum;
               end;
          end;
        indxr[i] := irow;
        indxc[i] := icol;
        if (a^[icol,icol] = 0.0) then twriteln(ofilvar,'Pause 2 in Gaussj - singular matrix');        
        pivinv := 1.0/a^[icol,icol];
        a^[icol,icol] := 1.0;
        for l := 1 to n do a^[icol,l] := a^[icol,l] * pivinv;
        for l := 1 to n do b^[icol,l] := b^[icol,l] * pivinv;
        for ll := 1 to n do
           if (ll <> icol) then
               begin
                  dum := a^[ll,icol];
                  a^[ll,icol] := 0.0;
                  for l := 1 to n do
                      a^[ll,l] := a^[ll,l] - a^[icol,l] * dum;
                  for l := 1 to n do
                      b^[ll,l] := b^[ll,l] - b^[icol,l] * dum;
               end;
      end;
   for l := n downto 1 do
      if (indxr[l] <> indxc[l]) then
         for k := 1 to n do 
            begin
                dum := a^[k,indxr[l]];
                a^[k,indxr[l]] := a^[k,indxc[l]];
                a^[k,indxc[l]] := dum;
            end;
end;

(*-----------#unix-version-----------------------------------------------------*)
(* end.*)
(*-----------#unix-version-----------------------------------------------------*)



