(*-----------#pc-version-----------------------------------------------------*)
{$N+,O+,F+,E+}
(* unit prior;
interface
uses tt1,tt2,tt3,tt4,tt5, mon, calctet;

type
  relation = (lt, eq, gt, none2); 

  priormeanrec = record
    val: real;
    cause: vertextype;
    effect: vertextype;
    free: boolean;
   end;

  priorinfo = record 
    cov: real;
    relparm: relation;
    reltype: relation;
    relnum: real;
      end;

  priorarr = array[1..maxparms,1..maxparms] of priorinfo;
  priorarray = ^priorarr; 
  priormeansarray = array[1..maxparms] of priormeanrec;

var
  numparms,numfree: integer;
  priorcov: priorarray;
  priormean: priormeansarray;
  invprior: covarray;
  priorvar,sqdetprior,pi2thep: real;
  parmstoprint: numberset;


procedure priorinit;
procedure writeprior(stretch1,stretch2: real; niter: integer);
procedure listparms;
procedure listprintparms;
procedure writeoutputinfo;

implementation
uses io,util,rsutil,sys, gib,numeric;

*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "gib.h"
#include "mon.h"
#include "calctet.h"
#include "io.h"
#include "prior.h"
#include "recalc.h"
#include "numeric.h"
#include "util.h"
#include "cspecial.h"
#include "part.h"
#include "indep1.h"
#include "rsutil.h"
#include "sys.h"
#include "calctet1.h"


(*-----------#unix-version-----------------------------------------------------*)

(************)
  (* This takes priorcov, which is the full covariance matrix
     over all the parameters, and forms pcov which is the reduced
     covariance matrix over only the free parameters. *)
procedure formpcov(var pcov: covarray; var numfree: integer);
  var i,j,n: integer;
  begin
   numfree:= 0;
   for i:= 1 to numparms do 
    if priormean[i].free then
      begin
        numfree:= numfree+1;
        n:= 0;
        for j:= 1 to numparms do
          if priormean[j].free then
             begin
                n:= n+1;
                pcov^[numfree,n]:= priorcov^[i,j].cov;
             end;
      end;
  end;


  (* This determines constants that will be used in 
     calculating the prior density. invprior is the
     inverse of the prior covariance matrix over the
     parameters.  sqdetprior is 1/sqrt(|prior cov mat|)
     pi2thep is 2pi raised to the negative numparms over 2 *)
procedure calcpdconsts(var priorcov: priorarray;
                         var invprior: covarray;
                         var sqdetprior,pi2thep: real);
  label 99;
  var pcov: covarray; ok: boolean;
  begin
   tnewcov(pcov,fatal);
   tnewcov(invprior,fatal);
   if fatal in errors then goto 99;
   formpcov(pcov,numfree);
   sqdetprior:= sqrt(determ(pcov,numfree,ok));
     (* This is because determ has changed pcov *)
   formpcov(pcov,numfree);
   pi2thep:= exp(ln(2*pi)*(-numfree/2));
    (* destroys pcov, stores answer in invprior *)
   inverse(pcov,invprior,numfree,ok);
 99: end;

procedure listfixedparms;
  var i,numchars, parmscovered: integer;
  begin
   parmscovered:= 0;
   i:= 0;
   repeat
     i:= i+1;
     twriteint(output,i);
     twrite(output,': |');
     if i<10 then numchars:= 3 else numchars:= 4;
     writeparm(i,output,numchars);
     for j:= 1 to (25 - numchars) do twrite(output,' |');
     if priormean[i].free then twrite(output,'Free |')
        else twrite(output,'Fixed|');
     parmscovered:= parmscovered+1;
     if (numparms > 25) and (i+25 <= numparms) then
       begin
         twrite(output,'        |');
         twriteint(output,i+25);
         twrite(output,': |');
         numchars:= 35;
         writeparm(i+25,output,numchars);
         for j:= 1 to (46 - numchars) do twrite(output,' |');
         if priormean[i+25].free then twrite(output,'Free|')
            else twrite(output,'Fixed|');
         parmscovered:= parmscovered+1;
       end;
     twriteeoln(output);
   until (parmscovered = numparms);
   twriteeoln(output)
  end;

procedure writeparminequal(i: integer);
  begin
   case priorcov^[i,i].reltype of
     lt: twrite(output, ' < |');
     gt: twrite(output, ' > |');
     eq: twrite(output, ' = |');
     none2: twrite(output, ' none |');
   end;
  if priorcov^[i,i].reltype <> none2 then
    tpwritereal(output, priorcov^[i,i].relnum,5,decpoints,false);
 end;

procedure listinequals;
  var i,numchars, parmscovered: integer;
  begin
   parmscovered:= 0;
   i:= 0;
   repeat
     i:= i+1;
     twriteint(output,i);
     twrite(output,': |');
     if i<10 then numchars:= 3 else numchars:= 4;
     writeparm(i,output,numchars);
     for j:= 1 to (15 - numchars) do twrite(output,' |');
     writeparminequal(i);
     parmscovered:= parmscovered+1;
     if (numparms > 25) and (i+25 <= numparms) then
       begin
         twrite(output,'        |');
         twriteint(output,i+25);
         twrite(output,': |');
         numchars:= 35;
         writeparm(i+25,output,numchars);
         for j:= 1 to (46 - numchars) do twrite(output,' |');
         writeparminequal(i+25);
         parmscovered:= parmscovered+1;
       end;
     twriteeoln(output);
   until (parmscovered = numparms);
   twriteeoln(output)
  end;   (* listinequals *)

procedure getfixed;
 var chparms: numberset;
     i,j: integer;
     ok: boolean;
 begin
  listfixedparms;
  twriteln(output,'List the numbers of any parameters whose fixed/free|');
  twriteln(output,'status you wish to change from its default value.|');
  twriteln(output,'Separate them by at least one space, and use no delimiters. |');
  repeat
    twrite(output,'List to change = |');
    getnumbers(chparms, numparms);
    twriteeoln(output);
    if chparms <> [] then
      for i:= 1 to numparms do
        if (i in chparms)
          then
            begin
               if (priorcov^[i,i].cov = 0.0)
                   then priorcov^[i,i].cov:= priorvar
               else priorcov^[i,i].cov:= 0.0;
               priormean[i].free:= not(priormean[i].free);
            end;
    listfixedparms;
    twrite(output,'List OK? (y/n)  [y] |');
    ok := compareto('y');
    twriteeoln(output);
  until ok;
 end;

procedure getnewinequals(i: integer);
 var tempreltype: real; c: char; numchars: integer;
 begin
   case priorcov^[i,i].reltype of
     lt: tempreltype:= 1.0;
     gt: tempreltype:= 2.0;
     eq: tempreltype:= 3.0;
     none2: tempreltype:= 4.0;
   end;
   writeparm(i,output,numchars);
   writeparminequal(i);
   twriteeoln(output);
   twrite(output,'Type either "<",  ">", "=", or "none" : |');
   readttyline(buf); readwhitespace; readit(c);
   case c of
    '<': priorcov^[i,i].reltype:= lt;
    '>': priorcov^[i,i].reltype:= gt;
    '=': priorcov^[i,i].reltype:= eq;
    'n': priorcov^[i,i].reltype:= none2;
    end;
   if priorcov^[i,i].reltype <> none2
     then if i <= max then
      priorcov^[i,i].relnum:= getnumber(0,maxint, true,
               'Number: |             ',priorcov^[i,i].relnum)
     else priorcov^[i,i].relnum:= getnumber(-maxint,maxint, true,
               'Number: |             ',priorcov^[i,i].relnum);
   twriteeoln(output);
 end;

procedure getinequals;
 var chparms: numberset;
     i,j: integer;
     ok: boolean;
 begin
  listinequals;
  twriteln(output,'List the numbers of any parameters upon which|');
  twriteln(output,'you wish to impose inequality constraints.|');
  twriteln(output,'Separate them by at least one space, and use no delimiters. |');
  repeat
    twrite(output,'List to change = |');
    getnumbers(chparms, numparms);
    twriteeoln(output);
    if chparms <> [] then
      for i:= 1 to numparms do
        if (i in chparms)
          then getnewinequals(i);
    listinequals;
    twrite(output,'List OK? (y/n)  [y] |');
    ok := compareto('y');
    twriteeoln(output);
  until ok;
 end;

procedure listparms;
  var i,numchars: integer;
  begin
   twriteln(output,'                 Prior|');
   twriteln(output,'Param.       Mean    Var|');
   for i:= 1 to numparms do 
     begin
       numchars:= 0;
       twriteint(output,i);
       twrite(output,': |');
       writeparm(i,output,numchars);
       for j:= 1 to (9 - numchars) do twrite(output,' |');
       tpwritereal(output, priormean[i].val, 5,decpoints,false);
       twrite(output,'    |');
       tpwritereal(output, priorcov^[i,i].cov, 5,decpoints,false);
       twriteeoln(output);
     end;
   twriteeoln(output);
  end;

procedure getnewprior(i: integer);
  var numchars: integer;
  begin
    twrite(output,'Prior mean for |');
    writeparm(i,output,numchars);
    if (i <= max) then 
       priormean[i].val:= getnumber(0,100,false,' |',priormean[i].val)
    else priormean[i].val:= getnumber(-100,100,false,' |',priormean[i].val);
    if priormean[i].free then
     begin
       twrite(output,'Variance for |');
       writeparm(i,output,numchars);
       priorcov^[i,i].cov:= getnumber(0,100,false,' |',priorcov^[i,i].cov);
       twriteeoln(output);
     end;
  end;

procedure getnewcov(i,j: integer);
  var temp: real;
      numchars: integer;
  begin
   if priormean[i].free and priormean[j].free then
    begin
      twrite(output,'Rho[|');
      writeparm(i,output,numchars);
      twrite(output,',|');
      writeparm(j,output,numchars);
      twrite(output,'] = |');
      temp:= priorcov^[i,j].cov/
              (sqrt(priorcov^[i,i].cov)*sqrt(priorcov^[j,j].cov));
      temp:= getnumber(-1,1,false,' |',temp);
      priorcov^[i,j].cov:= temp*(sqrt(priorcov^[i,i].cov)*
                                 sqrt(priorcov^[j,j].cov));
    end;
  end;

procedure getuserprior(var priorcov: priorarray);
 var chparms: numberset;
     i,j: integer;
     first: boolean;
 begin
   chparms:= []; first:= true;
   listparms;
   twriteln(output,'List the numbers of the parameters whose mean, variance, or|');
   twriteln(output,'covariance you wish to change from its default value in the prior.|');
   twriteln(output,'Separate them by at least one space, and use no delimiters. |');
   twrite(output,'List = |');
   getnumbers(chparms, numparms);
   twriteeoln(output);
   if chparms <> [] then
    begin
     for i:= 1 to numparms do
      if (i in chparms) then getnewprior(i);
     for i:= 1 to numparms do
      if (i in chparms) and priormean[i].free then
        for j:= i+1 to numparms do
          if (j in chparms) and priormean[j].free 
            then
              begin
                if first then
                  twriteln(output,'Enter as correlations, not covariances.|');
                getnewcov(i,j);
                first:= false;
              end;
    end;
 end; (* getuserprior *)

procedure listprintparms;
  var i,numchars: integer;
  begin
    for i:= 1 to numparms do
     if priormean[i].free or lrtest then
       begin
         twriteint(output,i);
         twrite(output,': |');
         if i<10 then numchars:= 3 else numchars:= 4;
         writeparm(i,output,numchars);
         twriteeoln(output);
       end;
   twriteeoln(output)
  end;

procedure getprintparms;
 var i: integer;
 begin
   parmstoprint:= [];
   twrite(output,'Print all free parameters in output file (y\n)? [y] |');
   if compareto('y') then
     begin
      for i:= 1 to numparms do
        if priormean[i].free then parmstoprint:= parmstoprint + [i];
      twriteeoln(output);
     end
   else 
      begin
        listprintparms;
        twriteln(output,'List the numbers of those parameters which|');
        twriteln(output,'you want printed in the output file. |');
        twrite(output,'List = |');
        getnumbers(parmstoprint, numparms);
        twriteeoln(output);
      end;
  end;


procedure priorinit;
label 99;
var  i,j,c,e: vertextype;
begin
 priorvar:= 16.0;
 twriteeoln(output);
 if (not lrtest) and (not flatprior) then
     priorvar:= getnumber(0.0, 500.0, true, 
               'Default variance for the Prior: |             ', priorvar);
 tnewpriorcov(priorcov, fatal);
 if fatal in errors then goto 99;
 numparms:= 0;
 for i:=1 to max do
  begin
   numparms:= numparms+1;
    (* default prior means for variances of error terms for 
       measured vars (i <= numob), for unmeasured vars (i <= max) *)
   priormean[i].val:= exogd[i].parm2;
   priormean[i].cause:= i; priormean[i].effect:= i;
      (* latent variances are fixed by default *)
   priormean[i].free:= (i <= numob);
   with priorcov^[i,i] do
     begin
       if priormean[i].free then cov:= priorvar else cov:= 0.0;
       relparm:= none2; reltype:= gt; relnum:= 0.0;
            (* variances are greater than 0 *)
     end;
  end; (* i:= 1 to max *)
 for c:= 1 to max do
  for e:= 1 to max do
   if (c <> e) and (adjacency^[c,e] <> notthere) then
    begin
      numparms:= numparms+1;
      with priormean[numparms] do 
        begin
          cause:= c; effect:= e;
          free:= true;
          val:= parm^[c,e];
        end;
      with priorcov^[numparms,numparms] do
       begin
         cov:= priorvar;
         relparm:= none2; reltype:= none2; relnum:= 0.0;
       end;
    end; (* if c <> e *)
  for i:= 1 to numparms do
   for j:= 1 to numparms do
    if (i <> j) then with priorcov^[i,j] do
      begin
         cov:= 0.0;
         relparm:= none2; reltype:= none2; relnum:= 0.0;
      end;
 if not lrtest then
   begin
    getfixed;
    getinequals;
    if not flatprior then
      begin
        getuserprior(priorcov);
        calcpdconsts(priorcov,invprior,sqdetprior,pi2thep);
      end;
    getprintparms;
   end;

99: end; (* priorinit *)

procedure writepriormarginal;
var i,j,numchars: integer;
begin
  if not flatprior then 
   begin
     twriteln(ofilvar,'MARGINALS|');
     twriteeoln(ofilvar);
     twriteln(ofilvar,'Parameter           Mean      Std. Dev   Free/Fixed|');
     twriteln(ofilvar,'---------------------------------------------------|');
   end
  else 
   begin
    twriteln(ofilvar,'Starting Values for Parameters|');
    twriteeoln(ofilvar);
    twriteln(ofilvar,'Parameter            Value     Free/Fixed|');
    twriteln(ofilvar,'-----------------------------------------|');
   end;
  for i:= 1 to numparms do
    begin
      numchars:= 0;
      writeparm(i,ofilvar,numchars);
      for j:= 1 to (20 - numchars) do twrite(ofilvar,' |');
      tpwritereal(ofilvar, priormean[i].val, 5,decpoints,false);
      twrite(ofilvar,'   |');
      if (not flatprior) then 
        begin
         tpwritereal(ofilvar, sqrt(priorcov^[i,i].cov), 5,decpoints,false);
         twrite(ofilvar,'   |');
        end;
      if priormean[i].free then twrite(ofilvar,' Free|')
        else twrite(ofilvar,' Fixed|');
      twriteeoln(ofilvar);
    end;
   twriteeoln(ofilvar);
 end; (* writepriormarginal *)


procedure writepriorcovs;
 var i,j,numchars: integer;
     tcor: real;
 found: boolean;
 begin
   numchars:= 0;
   twriteln(ofilvar,'Non-0 correlations between parameters in prior|');
   twriteln(ofilvar,'----------------------------------------------|');
   found:= false;
   for i:= 1 to numparms-1 do
    for j:= i+1 to numparms do
     if (priorcov^[i,j].cov <> 0) then
      begin
        found:= true;
        twrite(ofilvar,'Rho[|');
        writeparm(i,ofilvar,numchars);
        twrite(ofilvar,',|');
        writeparm(j,ofilvar,numchars);
        twrite(ofilvar,'] = |');
        tcor:= priorcov^[i,j].cov/
          (sqrt(priorcov^[i,i].cov)*sqrt(priorcov^[j,j].cov));
        tpwritereal(ofilvar, tcor,5,decpoints,false);
        twriteeoln(ofilvar);
      end;
   if not found then twriteln(ofilvar,' None|');
   twriteeoln(ofilvar);
 end;

procedure writeinequals;
 var i,j,numchars: integer;
 found: boolean;
 begin
   twriteln(ofilvar,'Equality and Inequality constraints|');
   twriteln(ofilvar,'   (All error variances > 0)|');
   twriteln(ofilvar,'-----------------------------------|');
   numchars:= 0; found:= false;
   for i:= 1 to numparms-1 do
    for j:= i to numparms do
     begin
      if (priorcov^[i,j].reltype <> none2) and
         ((i <> j) or ((i=j) and (i > max))) then
        begin
          found:= true;
          writeparm(i,ofilvar,numchars);
          case priorcov^[i,j].reltype of
             lt: twrite(ofilvar,' < |');
             gt: twrite(ofilvar,' > |');
             eq: twrite(ofilvar,' = |');
           end;
          tpwritereal(ofilvar, priorcov^[i,j].relnum,5,decpoints,false);
          twriteeoln(ofilvar);
        end;
      if (priorcov^[i,j].relparm <> none2) then
        begin
          found:= true;
          writeparm(i,ofilvar,numchars);
          case priorcov^[i,j].reltype of
             lt: twrite(ofilvar,' < |');
             gt: twrite(ofilvar,' > |');
             eq: twrite(ofilvar,' = |');
           end;
          writeparm(j,ofilvar,numchars);
        end;
     end;
   if not found then twriteln(ofilvar,' None|');
   twriteeoln(ofilvar);
end;

procedure writeoutputinfo;
var i,j: integer;
begin
    twritefilemess(ofilvar,'Output file: |',ofilename,totty);
    twritefilemess(ofilvar,'Graph file: |',gfilename,totty);
    twritefilemess(ofilvar,'Data file: |',dfilename,totty);
    twrite(ofilvar,'Graph:  |');
    if max > 0 then writegraph(adjacency,ofilvar,false,true);
    twriteeoln(ofilvar);
end; 

procedure writeprior(* stretch1,stretch2: real; niter: integer *);
 var  numchars: integer;
 begin
   twrite(ofilvar, '{|');
   writeoutputinfo;
   twrite(ofilvar, 'Stretch: |');
   tpwritereal(ofilvar, stretch1,5,decpoints,false);
   twriteeoln(ofilvar);
   twriteint(ofilvar, niter);
   twrite(ofilvar, ' iterations |');
   twriteeoln(ofilvar); twriteeoln(ofilvar);
   twriteln(ofilvar,'Prior distribution over Model Parameters |');
   if flatprior then
     twriteln(ofilvar,' is uninformative. |')
   else   twriteln(ofilvar,'    (multivariate normal)|');
   twriteln(ofilvar,'--------------------------|');
   twriteeoln(ofilvar);
   writepriormarginal;
   if not flatprior then
     writepriorcovs;
   writeinequals;
 end;

(*-----------#pc-version-----------------------------------------------------*)
(* end. *)
(*-----------#pc-version-----------------------------------------------------*)
