(*-----------#pc-version-----------------------------------------------------*)
 {$N+,O+,F+,E+}

(*
unit gib;
interface
uses tt1,tt2,tt3,tt4,tt5, mon, numeric, prior;

var gcov,ginverse: covarray;
    flatprior,lrtest: boolean;

procedure dolrt;
procedure dogibbs;
procedure writeheapleft(message: inbuf);
procedure writeparm(parmnum: integer; var ofile: text;
                    var numchars: integer);

implementation
uses io,util,rsutil,calctet,calctet1,sys; *)
(*-----------#pc-version-----------------------------------------------------*)

(* ------------ unix version  ------ *)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "prior.h"
#include "cspecial.h"
#include "mon.h"
#include "calctet.h"
#include "calctet1.h"
#include "io.h"
#include "sys.h"
#include "util.h"
#include "rsutil.h"
#include "recalc.h"
#include "numeric.h"
#include "gib.h"

(* ------------ unix version  ------ *)


var
 havevalue, novalueyet: vertexset;
 tiers : tiersptr;
 rejfile,bugfile: text;
 rejfilename: inbuf;


procedure writeheapleft(*message: inbuf*);
  begin
    twrite(output,message);
    twriteint(output,heapavail);
    twriteeoln(output);
  end;


(********************)
(* initializing routines *)

procedure getgibsprefs(var niter: longint; 
                       var stretch1,stretch2,tol: real;
                       var rejfilename: inbuf);
  var changed : boolean;
      seed: integer;
begin
 twriteeoln(output);
 twriteln(output,'Gibbs Parameter Estimation|');
 flatprior:= true;
 twrite(output,'Flat or informative prior (f\i)? [f] |');
 flatprior:= compareto('f');
 niter:= 10;
 niter:= round(getnumber(1, tmaxlongint, true,
               'Number of iterations: |             ', niter));
 stretch1:= 2.0;
 stretch1:= getnumber(0.0, 50.0, true, 
               'stretch1: |             ', stretch1);
 stretch2:= 1.0;
 tol:= 0.0001;
 initrand(seed);
(*  RandSeed:=  round(getnumber(-2147483647, 2147483647, false,
               'Random Seed: |             ', RandSeed)); *)
 rejfilename := 'gibbs.rej|';
 getfilename('Rejection Filename: |',rejfilename, changed);
 twriteeoln(output);
end;  (* getgibsprefs *)

procedure checkthomas(*var gcov,gcov2: covarray*);
 label 99;
 var i,j: vertextype;
     numchars: integer;
 begin 
  for i:= 1 to numob do
    for j:= 1 to numob do
      if abs(gcov^[i,j] - gcov2^[i,j]) > abs(0.001*gcov2^[i,j])  then 
       begin
         twriteln(output, 'TR impcov no good.|');
         rswritename(i, output, numchars); 
         twrite(output,'  |');
         rswritename(j, output, numchars); 
         twriteeoln(output);
         goto 99;
       end;
   twriteln(output,'TR impcov and formtrue match.|');
  99: end;

(* Notice the local parameters from Thomas' good style - RS 1/10/96 *)
procedure gibbsinit(var paramrecalcind: calclistindarrt;
                    var htermindex,covtermindex:hcovheadindext;
                    var gcov: covarray);
label 99;
var i,j: vertextype;
begin
  if not lrtest then rewritefile(rejfile, rejfilename);
  tnewcov(ginverse,fatal);  (* check if enough space *)
  tnewcov(gcov,fatal);  (* check if enough space *)
  tnewhcovheadindex(covtermindex,fatal); 
  tnewhcovheadindex(htermindex,fatal);
  if fatal in errors then goto 99;
  undercontinuous := true; 
  havevalue := [];       
  novalueyet := graphvertices;
  tiers := nil;
  new(varorder);
  formeqsorder(varorder,adjacency);
  (* formtrue(gcov, true, true); *)
  makeimplcov(gcov,adjacency, priormean, exogd, parm, numparms,
                                 paramrecalcind, htermindex, covtermindex);
99:  end;  (* gibbsinit *)

(***************************)
(*** Printing Routines  ****)

(* numchars is to keep track of how many characters are
   written by this procedure, so we can justify columns, etc. *)
procedure writeparm(*parmnum: integer; var ofile: text;
                    var numchars: integer*);
  begin
    if parmnum <= max then
      begin
        twrite(ofile,'eV(|');
        numchars:= numchars+3;
        rswritename(priormean[parmnum].cause, ofile, numchars);
        twrite(ofile,')|'); numchars:= numchars+1;
      end
    else
      begin
        rswritename(priormean[parmnum].cause, ofile,numchars);
        twrite(ofile,'->|'); numchars:=numchars+2;
        rswritename(priormean[parmnum].effect,ofile,numchars);
      end;
  end;

procedure writeparmheadings(var ofilvar: text);
  var i,j,numchars: integer;
  begin
   twrite(ofilvar,'Iter   |');
     for i:= 1 to numparms do
       if priormean[i].free and (i in parmstoprint)
        then
         begin
           numchars:= 0;
           writeparm(i,ofilvar,numchars);
           for j:= 1 to ((decpoints+7) - numchars) do twrite(ofilvar,' |');
         end;
     twriteln(ofilvar,' }|');
  end; 

procedure writerej(var rejfile: text; iter: integer;
                   parm: integer; cand,rejcons: real);
var numchars: integer;
begin
  writeparm(parm,rejfile, numchars);
  twrite(rejfile,': |');
  tpwritereal(rejfile, cand, 6, decpoints,false);
  twrite(rejfile,'  rj: |');
  tpwritereal(rejfile, rejcons, 6,decpoints,false);
  twriteeoln(rejfile);
end;

procedure postprogress(iter: longint; var thefile: text);
  begin
    twrite(thefile, 'Before Iter |');
    twriteint(thefile,iter);
(*    twrite(thefile, ' Rseed = |');write(thefile,RandSeed); *)
    twriteeoln(thefile);
  end;

(********End of Printing Routines **)

(***  Utilities for Approximating the Posterior  **)
function maxof(a,b: real): real;
  begin
    if (a < b) then maxof:= b else maxof:= a;
  end;

procedure getbrentstarts(param: integer;
                         var ax,bx,cx: real;
                         first: boolean);
 begin
  with priorcov^[param,param] do
    case reltype of
     none2: begin bx:= 0.0; ax:= -500.0; cx:= 500.0; end;
     gt: begin ax:= relnum; 
               cx:= relnum + 500;
               bx:= (ax+cx)/2;
         end;
     lt: begin cx:= relnum; 
               ax:= relnum - 500.0; 
               bx:= (ax + cx)/2;
         end;
     eq: begin bx:= relnum; ax:= relnum-500; cx:= relnum+500; end;
     end;
 end;

  (* Here we are using the brent minimizer to find the
     value of mean for which the posterior 
     density is maximal. This will become the mean of
     the approximating distribution from which we 
     will draw candidates. We then calculate the vr
     (the variance) for the approximating distribution.
     Dmean is the density at the mean *)
procedure formapproxdist(stretch1, tol: real;
                         param: integer;
                         var mean, vr, dmean: real;
                         first: boolean;
                         paramrecalcind:calclistindarrt;
                         covtermindex: hcovheadindext);
  var ax,bx,cx,dmeanplus,denom,gap: real;
  begin
    getbrentstarts(param,ax,bx,cx,first);
    dmean:= -(brent(param,ax,bx,cx,tol,mean,rejfile,paramrecalcind, covtermindex));
     (* we increase the gap between the value for which 
        the Brent routine found the posterior dens to be
        maximal (mean) and a sample point a "gap" away
        until there is some difference in posterior density,
        which we use to estimate the variance of the
        approximating distribution. *)
    gap:= -0.005;
    repeat
      gap:= -2*gap;
      dmeanplus:= neglogpost(param,(mean + gap),paramrecalcind, covtermindex);
      denom:= dmean + dmeanplus;
    until denom > 0.0 ;
    vr := stretch1*(0.5 * gap*gap) / denom ;
  end;

procedure isrealdraw(cand: real; param:integer;
                    var realdraw: boolean);
  begin
   with priorcov^[param,param] do
     case reltype of
       none2: realdraw:= true;
       lt: realdraw:= (cand < relnum);
       gt: realdraw:= (cand > relnum);
     end;
  end; (* is realdraw *)

  (* We draw cand from a distribution with mean and vr
     (variance) that we got from formapprox dist.  We then 
     calculate dcand, which is the density of cand, and compare 
     the densities of the candidate under the approximating
     distribution and under the real posterior, keeping the
     candidate if a uniform [0..1] draw exceeds the proportion
     of their densities. *)
procedure drawfromapprox(var cand: real; param, iter: integer;
                         mean,vr,dmean,stretch2: real;
                         paramrecalcind: calclistindarrt;
                         covtermindex:hcovheadindext);
  var realdraw: boolean;
      dcand,rejcons,rj,numer,denom,accept: real;
  begin
    repeat
      cand:= randnorm(mean,vr);
      isrealdraw(cand,param,realdraw);
      if realdraw then
        begin
          dcand:= -1*neglogpost(param,cand,paramrecalcind, covtermindex);
          numer:= dcand-dmean;
          denom:= (-1*sqr(cand-mean)/(2*vr)) - ln(stretch2);
          rj:= numer - denom;
          accept:= ln(randunif(0,1));
        end;
    until realdraw and (rj > accept);
    if (rj > 5.0) then rj:= 5.0;
    if (exp(rj) > 1.0001) then 
       writerej(rejfile,iter,param,cand,exp(rj));
 end;

procedure updateparm(param: integer; cand: real);
  var cause, effect: vertextype;
  begin
   cause:= priormean[param].cause;
   effect:= priormean[param].effect;
   if (not lrtest) and (param in parmstoprint) then
     begin
       twrite(ofilvar,' |');   
       tpwritereal(ofilvar, cand, decpoints+6, decpoints, false);
     end;
   if param <= max then
     exogd[cause].parm2:= cand
   else parm^[cause,effect]:= cand;
  end;


procedure writebugfile(iter: longint);
  var value: real;
      param: integer;
      bugfilename: inbuf;
      cause, effect: vertextype;
  begin
    bugfilename := 'gibbs.bug';
(*    assign(bugfile,bugfilename); *)
    rewrite(bugfile,bugfilename);
    postprogress(iter, bugfile);
    writeparmheadings(bugfile);
    twriteint(bugfile,iter-1); twrite(bugfile,'  |');
    for param:= 1 to numparms do
       if priormean[param].free then
         begin
           cause:= priormean[param].cause;
           effect:= priormean[param].effect;
           if param <= max then value:= exogd[cause].parm2
           else value:= parm^[cause,effect];
           tpwritereal(bugfile, value, 9 , decpoints, false);
         end; 
    finishfile(bugfile,bugfilename);
  end;

(*****************************)

(***************************************************)
(****   Main Control for Gibbs    ****)
procedure getparmsprinted;
 var i,j: integer;
 begin
   listprintparms;
   twriteln(output,'List the numbers of those parameters which|');
   twriteln(output,'are printed in the gibbs sample file. |');
   twrite(output,'List = |');
   getnumbers(parmstoprint, numparms);
   twriteeoln(output);
 end;

procedure getparmvals;
 var param,numchars,lociter: integer;
      value: real;
 begin
   readwhitespace;
   lociter:= round(readnum(false,fatal));
   for param:= 1 to numparms do
    if (param in parmstoprint) and not(myeoln) then 
      begin
        readwhitespace;  
        value:= readnum(false,fatal);
        if value= 0.0 then value:= 0.0000001;
        updateparm(param,value);
      end;
 end;


procedure dolikestats(var scov,cov,gcov: covarray; iternum: integer;
                      var simworse: boolean);
 var origFML,simFML: real;
 begin
  (* twriteln(ofilvar,'Implied Cov|');
   writemat(gcov,ofilvar,2);
   twriteln(ofilvar,'Simulated sample|');
   writemat(scov,ofilvar,2);
   twriteln(ofilvar,'Original sample|');
   writemat(cov,ofilvar,2);
   twriteint(ofilvar,iternum);
   twrite(ofilvar,'  |');  *)
   origFML:= FML(cov, gcov);
   simFML:=  FML(scov, gcov);
 (*  tpwritereal(ofilvar, origFML, 9 , decpoints, false);
   twrite(ofilvar,'  |');
   tpwritereal(ofilvar, simFML, 9 , decpoints, false);
   twrite(ofilvar,'  |');
   if (simFML > origFML) then
     begin twriteint(ofilvar, 1); simworse:= true; end
    else begin twriteint(ofilvar, 0); simworse:= false; end; *)
   simworse:= (simFML > origFML);
 end;

procedure lrtinit(var numperiter: integer;
                  var gibbsfilename: inbuf; var gibbsfile: text);
 var changed: boolean; 
 begin
   gibbsfilename := 'gibbs.out|';
   getfilename('File that contains gibbs sample: |',gibbsfilename, changed);
   twriteeoln(output);
   resetfile(gibbsfile, gibbsfilename,montef);
   (* skipping the comment in the gibbs sample file *)
   readcomment(gibbsfile);
   twritefilemess(ofilvar,'Gibbs sample file: |',gibbsfilename,totty);
   writeoutputinfo;
   twriteln(ofilvar, 'Likelihood Ratio Test|');
   twriteln(ofilvar, '---------------------|');
   twriteeoln(ofilvar);
   twrite(ofilvar,'Sample Size = |');
   twriteint(ofilvar, round(sampsize));   
   getparmsprinted;
   tnewcov(scov,fatal);
   numperiter:= 1;
   numperiter:= round(getnumber(1, 50, true, 
         'Number of samples per population: |             ', numperiter));
 end;

procedure reportresults(iternum,numperiter,totalscore: integer);
begin
 twriteeoln(ofilvar);
 twrite(ofilvar,'Number of parameter vectors in Gibbs sample: |');
 twriteint(ofilvar, iternum);   twriteeoln(ofilvar);
 twriteeoln(ofilvar);
 twriteln(ofilvar,'The empirical p-value is defined as follows. For each|');
 twriteln(ofilvar,'population P(i) defined by the graph and the parameter|');
 twriteln(ofilvar,'vector theta(i) from the Gibbs sample, n pseudo-random|');
 twriteln(ofilvar,'samples S(i,n) are drawn. The p-value is the relative|');
 twriteln(ofilvar,'frequency with which the fit of P(i) to the original|');
 twriteln(ofilvar,'sample is superior to the fit of P(i) to S(i,n).|');
 twriteeoln(ofilvar);
 twrite(ofilvar,'Empirical p-value : |');
 tpwritereal(ofilvar, totalscore/(numperiter*iternum), 8 , 5, false);
end;

procedure printgibbsprofile;
 var totaltime: integer;
 begin
  totaltime:= round(tgettime - gibstart); 
  twrite(output,'The likelihood was calculated a total of |');
  twriteint(output, likecalcnum); twrite(output,' times.|');  twriteeoln(output);
  twrite(output,'  Total seconds  = |'); twriteint(output, totaltime); 
  twriteeoln(output);twriteeoln(output);
  twriteln(output,'Calculating the implied covariances (New method)|');
  twrite(output,'  - % of tot. time = |'); twriteint(output, round(100*trtime/totaltime));
  twriteeoln(output);
  twrite(output,'  - total seconds  = |'); twriteint(output, round(trtime)); twriteeoln(output);
  twrite(output,'  - Trek calculations. Seconds  = |'); twriteint(output, round(trektime)); twriteeoln(output);
  twrite(output,'  - % of Formtrue from treks = |'); twriteint(output, round(100*trektime/imptime));
  twriteeoln(output);  twriteeoln(output);
  twrite(output,'Calculating the likelihood (traces, matrix inversion, determinants)|');
  twriteeoln(output);
  twrite(output,'  - % of tot. time = |'); twriteint(output, round(100*matoptime/totaltime));
  twriteeoln(output);
  twrite(output,'  - total seconds  = |'); twriteint(output, round(matoptime)); twriteeoln(output);
  twriteeoln(output);
 end;

procedure printlrtprofile;
 var totaltime: integer;
 begin
  totaltime:= round(tgettime - gibstart); 
  twrite(output,'  Total seconds  = |'); twriteint(output, totaltime); 
  twriteeoln(output);twriteeoln(output);
  twriteln(output,'Calculating the implied covariances|');
  twrite(output,'  - % of tot. time = |'); twriteint(output, round(100*imptime/totaltime));
  twriteeoln(output);
  twrite(output,'Drawing samples|');
  twriteeoln(output);
  twrite(output,'  - % of tot. time = |'); twriteint(output, round(100*trtime/totaltime));
  twriteeoln(output);
  twrite(output,'Calculating FML (traces, matrix inversion, determinants)|');
  twriteeoln(output);
  twrite(output,'  - % of tot. time = |'); twriteint(output, round(100*matoptime/totaltime));
  twriteeoln(output);
 end;

procedure timeinit;
 begin
   matoptime:= 0.0; imptime:= 0.0; trtime:= 0.0; gibstart:= tgettime; 
   likecalcnum:= 0; trektime:= 0.0;
 end;

procedure timewarn(niter: longint; var wantstop: boolean);
 var  first, estimate: real;
 begin
  first:= tgettime - gibstart;
  estimate:= (first * niter)/60;
  if estimate > 5.0 then
    begin
      twrite(output,'The first iteration took about |');  
      tpwritereal(output, first, 4, 2, false); twrite(output, ' seconds.|');
      twriteeoln(output);
      twrite(output,'The estimated time is thus |');
      if estimate > 60.0 then
        begin
          twriteint(output, round(estimate/60.0)); 
          twrite(output, ' hours.|');
          twriteeoln(output);
        end
      else
        begin
          twriteint(output, round(estimate)); 
          twrite(output, ' minutes.|');
          twriteeoln(output);
        end;
      wantstop:= false;
      twrite(output,'Continue (y\n)? [y] |');
      wantstop:= compareto('n');
    end;
 end;


  (* This computes an empirical version of a p-value for
     the likelihood ratio test *)
procedure dolrt;
  var gibbsfilename: inbuf; 
      gibbsfile: text; simworse: boolean;
      iternum,totalscore,numperiter,i: integer;
      paramrecalcind:calclistindarrt;
      htermindex,covtermindex:hcovheadindext;
      loctime: double;
  begin
   lrtest:= true;
   priorinit; (* notice the lack of parameters *)
   gibbsinit(paramrecalcind,htermindex,covtermindex,gcov);
   lrtinit(numperiter,gibbsfilename,gibbsfile);
   iternum:= 0;
   totalscore:= 0;
   timeinit;
   twrite(output,'Finished iteration |');
   repeat
     iternum:= iternum+1;
     readfileline(gibbsfile,buf);
     getparmvals;
     loctime:= tgettime;
     formtrue(gcov,true,true);
     imptime:= imptime + (tgettime-loctime);
     for i:= 1 to numperiter do
       begin
         loctime:= tgettime;
         getsamplecov(scov);
         trtime:= trtime + (tgettime-loctime);
         loctime:= tgettime;
         dolikestats(scov,cov,gcov,iternum,simworse);
         matoptime:= matoptime + (tgettime - loctime);
         if simworse then totalscore:= totalscore+1;
       end;
     twriteint(output,iternum); twrite(output,' |');
   until eof(gibbsfile);
   twriteeoln(output);
   reportresults(iternum, numperiter,totalscore);
   dispose(gcov); dispose(scov); dispose(ginverse);
   disposehcovind(htermindex,covtermindex);
   disposerecalclist(paramrecalcind,numparms);
   finishfile(ofilvar,ofilename);
   printlrtprofile;
  end; (* dolrt *)


procedure dogibbs;
label 99;
var param: integer;
    iter,niter: longint;
    stretch1,stretch2,tol: real;
    mean,vr,dmean,cand: real;
    first,realdraw,wantbug,wantstop: boolean;
    paramrecalcind:calclistindarrt;
    htermindex,covtermindex:hcovheadindext;
begin
  lrtest:= false; wantbug:= false; wantstop:= false;
  getgibsprefs(niter,stretch1,stretch2,tol,rejfilename);
  priorinit;
  gibbsinit(paramrecalcind,htermindex,covtermindex,gcov);
  writeprior(stretch1,stretch2,niter);
  writeparmheadings(ofilvar);
  timeinit;
  for iter:= 1 to niter do
    begin
      first:= (iter = 1);
      if wantbug then writebugfile(iter);
      postprogress(iter,rejfile);
      tpwriteint(ofilvar,iter, 4,false); twrite(ofilvar,' |');
      for param:= 1 to numparms do
       if priormean[param].free and (not aaabort) then
         begin
           formapproxdist(stretch1,tol,param,mean,vr,dmean,first,paramrecalcind, covtermindex);
           drawfromapprox(cand,param,iter,mean,vr,dmean,stretch2,paramrecalcind, covtermindex);
           updateparm(param,cand);
         end
       else if aaabort then goto 99;
      twriteeoln(ofilvar);
      if (iter mod 10) = 0 then
        begin
          twrite(output,'Iter # |');
          twriteint(output,iter);
          twriteeoln(output);
        end;
      if first then begin timewarn(niter, wantstop); if wantstop then goto 99; end;
    end; (* iterations *)
  dispose(gcov);
  dispose(ginverse);
  disposepointrs(tiers,varorder);
  disposehcovind(htermindex,covtermindex); 
  disposerecalclist(paramrecalcind,numparms);
  finishfile(ofilvar,ofilename);
  finishfile(rejfile,rejfilename);
  printgibbsprofile;
99: end;


(*-----------#pc-version-----------------------------------------------------*)
(* end. *)
(*-----------#pc-version-----------------------------------------------------*)
