//Copyright (C) 1997 Frank C. Wimberly.  All rights reserved.
import java.io.*;
import java.util.*;

class bayesNetSample{

  protected int sampleSize;
  protected int cases[][];
  protected bayesNet B;
  private int values[];       // new variables in bayesNetSample class
  nrRandom tst;               // new variables in bayesNetSample class

  public bayesNetSample(bayesNet B, int sampleSize) 
  {
     this.sampleSize = sampleSize;
     this.B = B;
     cases = new int[sampleSize][B.D.nvertices];
     values = new int[B.D.nvertices];  //goes in constructor
  }

  public void readSample(String fileName)
  {
    InputStream s;
    StringTokenizer st;

    try { s = new FileInputStream(fileName); } catch (IOException e) {
      System.out.println("Cannot open file " + fileName);
      return;
      }
    DataInputStream in = new DataInputStream(s);
    for (int k = 0; k < sampleSize; k++)  {
      try { 
        st = new StringTokenizer(in.readLine( ));
        for(int j = 0; j < B.D.nvertices; j++) cases[k][j] = Integer.parseInt(st.nextToken( ));
        } 
      catch (IOException e) {
        System.out.println("Read error in " + fileName);
        return;
        }
      //System.out.println("Read " + value[0] + " " + value[1] + " " + value[2] + " " + DEBUG
      //                  value[3] + " " + value[4] + " " + value[5]);                  DEBUG
      }
  }

  public void genSample (long seed) 
  {

    tst = new nrRandom(seed);  //goes in genSample
    //debug print
    //System.out.println("ntiers = " + ntiers);
    //System.out.println(" Level 0:  nvars = " + nvars[0] + " " + tiers[0][0]);
    //System.out.println(" Level 1:  nvars = " + nvars[1] + " " + tiers[1][0]);
    //System.out.println(" Level 2:  nvars = " + nvars[2] + " " + tiers[2][0] + " " + tiers[2][1]);
    for(int i = 0; i < sampleSize; i++) {
      //Generate case(i);
      int j;
      for(int level = 0; level < B.D.ntiers; level++) {
        if(B.D.nvars[level] > 0) {
          for(int var = 0; var < B.D.nvars[level]; var++) {
            int v = B.D.tiers[level][var];
            if(level != 0) {
              int np = B.D.getNumParents(v);
              for(int par = 0; par < np; par++) {
                int parent = B.D.getParent(v, par);
                values[par] = cases[i][parent];
                }
              }
            float P[] = new float[B.nvalues[v]]; 
            for(j = 0; j < B.nvalues[v]; j++) {
              float p = B.conditionalProb(v, values, j);
              P[j] = p;
              }
            int val = toss(P, B.nvalues[v]);
            cases[i][v] = val;
            }
          }
        }
      }
  }

  public void writeSample (String fn) 
  {
    OutputStream s;
    String fileName = "/tmp/" + fn;
    //String fileName = "/usr/user/wimberly/op.tst";
    try { s = new FileOutputStream(fileName); } catch (IOException e) {
      System.out.println("Cannot open file " + fileName);
      return;
      }
    DataOutputStream d = new DataOutputStream(s);


    //try { 
    //  d.writeBytes("/raw\n");
    //  d.writeBytes(sampleSize + "\n");
    //  for(int j = 0; j < B.D.nvertices; j++) d.writeBytes(B.D.vertexNames[j] + " ");
    //  d.writeBytes("\n");
    //  } catch (IOException e) {
    //    System.out.println("Write error in file " + fileName);
    //    }
    for(int i = 0; i < sampleSize; i++) {
      //System.out.println("Case " + i + " " + cases[i][0] + " " + cases[i][1] + " " +
      //cases[i][2] + " " + cases[i][3]);
      try { 
        for(int j = 0; j < B.D.nvertices; j++) d.writeBytes(" " + cases[i][j]);
        d.writeBytes("\n");
        } catch (IOException e) {
          System.out.println("Write error in file " + fileName);
          }
      }
  }


  public int toss(float p[], int n)
  { int i;
    float pp[] = new float[n+1];
    float ran = tst.nrnextFloat( );
    pp[0] = (float) 0.0;
    pp[n] = (float) 1.0;
    for(i = 1; i < n; i++) 
      pp[i] = pp[i-1] + p[i - 1];
    for(i = 0; i < n; i++) {
      //System.out.println("i = " + i + " range = " + pp[i] + " " + pp[i+1] + " " + " ran = " + ran);
      if(ran >= pp[i] && ran < pp[i+1]) return i;
      }
    return n-1;
  }
}
