//Copyright (C) 1997 Frank C. Wimberly.  All rights reserved.

class nrRandom{

  private final int IM = 2147483647;
  private final int IA = 16807;
  private final float AM = (float) (1.0 / IM);
  private final int IQ = 127773;
  private final int IR = 2836;
  private final int NTAB = 32;
  private final int NDIV = (1 + (IM - 1)/NTAB);
  private final float EPS = (float) 1.2e-7;
  private final float RNMX = (float) (1.0 - EPS);

  protected long seed;
  protected long iy = 0;
  protected long iv[] = new long[NTAB];
  protected int iset = 0;
  protected float gset;

  public nrRandom(long sd)
  {
    this.seed = sd;
  }

  public float nrnextFloat( )
  {
     int j;
     long k;
     float temp;

     if(seed <= 0 || iy == 0)
       {
         if( -seed < 1 ) seed = 1;
         else seed = -seed;
         for (j = NTAB + 7; j >= 0; j--)
           {
              k = seed/IQ;
              seed = IA * (seed - k*IQ) - IR*k;
              if(seed < 0) seed += IM;
              if(j < NTAB) iv[j] = seed;
           }
         iy = iv[0];
        }
      k = seed / IQ;
      seed = IA * (seed - k*IQ) - IR*k;
      j = (int) iy/NDIV;
      iy = iv[j];
      iv[j] = seed;
      if((temp = AM*iy) > RNMX) return RNMX;
      else return temp;
  }

  public float nrnextGaussian( )
  {
     float fac, rsq, v1, v2;
    
     if (iset == 0) {
       do {
          v1 = (float) (2.0*this.nrnextFloat( ) - 1.0);
          v2 = (float) (2.0*this.nrnextFloat( ) - 1.0);
          rsq = v1*v1 + v2*v2;
          } while (rsq >= 1.0 || rsq == 0.0);
       fac = (float) (Math.sqrt(-2.0*Math.log((double) rsq)/rsq));
       gset = v1*fac;
       iset = 1;
       return v2*fac;
       } else {
     iset = 0;
     return gset;
     }
  }
}
