(****** recalc.p ******)

(* ------------ unix version  ------ *)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "prior.h"
#include "mon.h"
#include "util.h"
#include "sys.h"
#include "recalc.h"

(* ------------ unix version  ------ *)

procedure tnewhcovheadindex(* var ptr: hcovheadindext; err: errortype *);
var size: word;

begin
   size := sizeof(ptr) * maxvertex * maxvertex;
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create covariance calculation 	datastructure|',err,false);
end;


procedure initheadrec(var headrec:headptr);

begin
   new(headrec);
   with headrec^ do
	begin
   	tail := nil;
   	val := 0;
   	params := [];
	end;
end;

procedure inittailrec(var tailrec:tailptr);

begin
   new(tailrec);
   with tailrec^ do
     begin
      next := nil;
      coeff := 0;
      hcov := nil;
    end;
end;

procedure initdblrec(var dblord:dblptr);

begin
	new(dblord);
	dblord^.vert:= 0;
	dblord^.next:= nil;
	dblord^.prev:= nil;
end;

procedure initrecalcrec(var calcrec:calcptr);

begin
	new(calcrec);
	calcrec^.next:=nil;
	calcrec^.hcov:=nil;
end;

procedure disposerecalclist(*var calclist:calclistindarrt; numparms:integer*);

var curptr,tmptr:calcptr;
    j:integer;

begin

for j:=1 to numparms do
	begin
 	curptr:=calclist[j];
	while curptr<>nil do
	  begin
	  tmptr:=curptr^.next;
	  dispose(curptr);
	  curptr:=tmptr;
	  end;
	end;
end;

procedure disposehcovind(*var hind,covind:hcovheadindext*);

var curptr,tmptr:tailptr;
    	j,k:integer;

begin

for j:=1 to origmax do
	for k:=1 to origmax  do
 	if hind^[j,k]<>nil then
		begin
		curptr:=hind^[j,k]^.tail;
		dispose(hind^[j,k]);
		while curptr<>nil do
		  begin
		  tmptr:=curptr^.next;
		  dispose(curptr);
		  curptr:=tmptr;
		  end;
		end;
dispose(hind);


for j:=1 to origmax do
	for k:=1 to j  do
 	if covind^[j,k]<>nil then
		begin
		curptr:=covind^[j,k]^.tail;
		dispose(covind^[j,k]);
		while curptr<>nil do
		  begin
		  tmptr:=curptr^.next;
		  dispose(curptr);
		  curptr:=tmptr;
		  end;
	end;
dispose(covind);

end;	



procedure disposedblorder(var stdblord:dblptr);


var tmptr,curptr:dblptr;

begin

curptr:=stdblord;

while curptr<>nil do
	begin
	tmptr:=curptr;
	curptr:=curptr^.next;
	dispose(tmptr);
	end
end;

procedure formdblorder(var tmporder:eqorder; var stdblord, endblord: dblptr);

var prevptr, nxtptr:dblptr;

begin
	prevptr:= nil;
	initdblrec(stdblord);
	nxtptr:=stdblord;
	while tmporder^.next^.next <> nil do 
		begin
		nxtptr^.vert:= tmporder^.varbl;
		nxtptr^.prev:= prevptr;
		prevptr:= nxtptr;
		initdblrec(nxtptr^.next);
		nxtptr:=nxtptr^.next;
		tmporder:=tmporder^.next
		end;
	nxtptr^.vert:= tmporder^.varbl;
	nxtptr^.prev:= prevptr;
	endblord:= nxtptr;
end;



procedure lslice1(var adj1: adjtype; v1: vertextype;var temp: vertexset);
{adj1 - the input graph; v1 - a vertex in that graph
 temp - the set of vertices which is output
 lslice1 gives the set of vertices (x | adj1^[x,v1] is there)
 If supplied with an ordinary graph matrix as input, lslice1 will give the set
 of parents of v1 as output}
var i: vertextype;

begin
temp := [] ;
for i :=1 to origmax do
    if adj1^[i,v1]<>notthere then temp := (temp + [i]);
end;


procedure rslice1(var adj1: adjtype; v1: vertextype;var temp: vertexset);
{adj1 - the input graph; v1 - a vertex in that graph
 temp - the set of vertices which is output
 rslice1 gives the set of vertices (x | adj1^[v1,x] is there)
 If supplied with an ordinary graph matrix as input, rslice1 will give the set
 of children of v1 as output}
var i: vertextype; 

begin
  temp := [];
  for i :=1 to origmax do
  if adj1^[v1,i]<>notthere then temp := (temp + [i]);
end;


procedure buildrecalclist(paramsocc: vertexset;
			var hcovind:hcovheadindext; 
			var paramrecalcind,endlistptrs:calclistindarrt; 
			xptr,yptr:vertextype;
			numparms:integer);

var j:integer;

begin
for j:=1 to numparms do
	if j in paramsocc then
		begin
   	   if (paramrecalcind[j] = nil) then
		begin
		initrecalcrec(paramrecalcind[j]);
		endlistptrs[j]:=paramrecalcind[j];
		end
	   else
		begin
		initrecalcrec(endlistptrs[j]^.next);
		endlistptrs[j]:= endlistptrs[j]^.next;
		end;
		endlistptrs[j]^.hcov := hcovind^[xptr,yptr];

		end;
end;

procedure sumproduct(var hcovar:headptr; var exogd:exogarray; var parm:covarray;
                      var priormean:priormeansarray);
		
var curptr:tailptr;
    temp,total: real;

(* NB if we stored exogd[j].parm2 in parm[j,j] we could speed this up 
	by removing the conditional test! *)

begin

curptr:= hcovar^.tail;

total:=0;
while curptr<>nil do
	begin
	if curptr^.hcov<>nil then temp:= curptr^.hcov^.val else temp:=1;
total:= total + parm^[priormean[curptr^.coeff].cause, priormean[curptr^.coeff].effect]*temp;
	curptr:=curptr^.next;
	end;
hcovar^.val := total;

end;

(*if priormean[curptr^.coeff]^.cause = priormean[curptr^.coeff].effect then
		total:= total + exogd[priormean[curptr^.coeff].cause].parm2*temp *)
 		
	
	
(* Data structure Initialization Routines 
   Called once at beginning of run			  *)


procedure makeimplcov(*var gcov:covarray; var graph:adjtype; 
			 		  var priormean:priormeansarray; 
					  var exogd:exogarray;
					  var parm:covarray;
					  numparms:integer;
					  var paramrecalcind:calclistindarrt
					  var htermindex,covtermindex:hcovheadindext*);

var j,k: integer;
    parsocc : set of vertextype;
    parentset,childset: vertexset;
    tmporder: eqorder;
    parmind: adjtype;
    fromptr,toptr,xptr,yptr,stdblord,endblord:dblptr;
    fromvert,tovert,xvert,yvert: vertextype;
    headdone:boolean;
    endptrarr: calclistindarrt;
    curtailrec:tailptr;
    trsptr:tiersptr;

begin	 	 
tnewhcovheadindex(htermindex,fatal);
tnewhcovheadindex(covtermindex, fatal);
tnewadj(parmind,fatal);

for j:=1 to numparms do
	begin
		(* Construct parameter index *)
			parmind^[priormean[j].cause, priormean[j].effect]:=j;
		(* Initialize recalculation list *)
			paramrecalcind[j]:=nil;
			endptrarr[j]:=nil
	end;

for j:=1 to origmax do
	for k:=1 to origmax do
		begin
			htermindex^[j,k]:=nil;
			covtermindex^[j,k]:=nil;
		end;
		
for k:=1 to origmax do
  parm^[k,k]:=exogd[k].parm2;
		
(* A: Form Consistent Ordering - use same method as in 'formtrue' *)
new(tmporder);
formeqsorder(tmporder,graph);

formdblorder(tmporder, stdblord,endblord);
trsptr:=nil;
disposepointrs(trsptr,tmporder);

(* B: Build H term Lists *)

toptr:=endblord;

while toptr^.prev <> nil do
 begin
 	tovert:=toptr^.vert;
	fromptr:=toptr^.prev;
 	while fromptr <> nil do
 	begin
		fromvert:=fromptr^.vert;
		headdone:=false;
		parsocc:=[];
		rslice1(graph,fromvert,childset); (* Find Children of 'from' vertex *)
			(*for each Child of fromvert, which is an ancestor of tovert
							put an entry in the linked list for H(from;to) *) 	
 		 if childset<>[] then	
 			for j:=1 to origmax do
	 		    if (j in childset) and 
				((htermindex^[j,tovert]<>nil) or (j=tovert)) then 
 					begin
 					if not headdone then begin
 						initheadrec(htermindex^[fromvert,tovert]);
						inittailrec(htermindex^[fromvert,tovert]^.tail);
	 					curtailrec:=htermindex^[fromvert,tovert]^.tail;
 						headdone:=true;
 						end
 					else begin
 						inittailrec(curtailrec^.next);
 						curtailrec:=curtailrec^.next;
 						end;
 					curtailrec^.coeff:=parmind^[fromvert,j];
 					parsocc:= parsocc + [parmind^[fromvert,j]];
 					if j<>tovert then begin
 					curtailrec^.hcov:=htermindex^[j,tovert];
 					parsocc:= parsocc + htermindex^[j,tovert]^.params;
 					end
 						else curtailrec^.hcov:=nil;
 					end;
 		if headdone then 
 		begin
 			htermindex^[fromvert,tovert]^.params:=parsocc;
 		 	buildrecalclist(parsocc, htermindex, paramrecalcind,endptrarr, 
 		 							fromvert, tovert, numparms);
 		 	sumproduct(htermindex^[fromvert,tovert],exogd,parm,priormean);
		end;			
		fromptr:=fromptr^.prev
	end;
	toptr:=toptr^.prev
 end;
 
(* C: Build Cov term Lists *)

xptr:=stdblord;
while xptr<>nil do
   begin
	yptr:=xptr;
	xvert:=xptr^.vert;
	while yptr <> nil do
	 begin
		yvert:=yptr^.vert;
		headdone:=false;
		parsocc:=[];
		if (htermindex^[xvert,yvert] <> nil) or (xvert=yvert) then begin
 						initheadrec(covtermindex^[xvert,yvert]);
				covtermindex^[yvert,xvert]:=covtermindex^[xvert,yvert];
 						inittailrec(covtermindex^[xvert,yvert]^.tail);
 						curtailrec:=covtermindex^[xvert,yvert]^.tail;
 						headdone:=true;
 						curtailrec^.coeff:=parmind^[xvert,xvert];
 						parsocc:= parsocc + [parmind^[xvert,xvert]];
 						if xvert<>yvert then 
 							begin
 							curtailrec^.hcov:= htermindex^[xvert,yvert];
 							parsocc:= parsocc + htermindex^[xvert,yvert]^.params;
 							end
 							else curtailrec^.hcov:= nil;
 						end;
 		lslice1(graph,xvert,parentset);
 		if parentset<>[] then	
 			for j:=1 to origmax do
 				if (j in parentset) and (covtermindex^[j,yvert]<>nil) then
 					begin
 					if (headdone=false) then	
 						begin
 						initheadrec(covtermindex^[xvert,yvert]);
					covtermindex^[yvert,xvert]:=covtermindex^[xvert,yvert];
						inittailrec(covtermindex^[xvert,yvert]^.tail);
 						curtailrec:=covtermindex^[xvert,yvert]^.tail;
 						headdone:=true;
 						end
 					else begin
 						inittailrec(curtailrec^.next);
 						curtailrec:=curtailrec^.next;
 						end;
 					curtailrec^.coeff:=parmind^[j,xvert];
 					curtailrec^.hcov:=covtermindex^[j,yvert];
 					parsocc:= parsocc + [parmind^[j,xvert]] 
 								+ covtermindex^[j,yvert]^.params;
 					end;
 	if (headdone=true) then 
 		begin
 		covtermindex^[xvert,yvert]^.params:=parsocc;
 	 	buildrecalclist(parsocc, covtermindex, paramrecalcind,endptrarr,xvert, yvert, numparms);
 		sumproduct(covtermindex^[xvert,yvert],exogd,parm,priormean);
		end;	 					
 	yptr:=yptr^.next;
   	end;
 xptr:=xptr^.next;
 end;

for j:=1 to numob do
for k:=1 to numob do
if covtermindex^[j,k]<>nil then gcov^[j,k]:=covtermindex^[j,k]^.val
   else gcov^[j,k]:=0;

for k:=1 to origmax do
parm^[k,k]:=0;

dispose(parmind);
disposedblorder(stdblord);
			
end;

			
(* Re-calculation Routines *)		

procedure recalcmatrix(*var gcov:covarray; var exogd:exogarray;
					var parm:covarray;
					xparam: integer;
					var paramrecalcind:calclistindarrt;
					var covtermind:hcovheadindext
					var priormean:priormeansarray*);

var curcalptr:calcptr;
    hcovar:headptr;
    j,k:integer;

begin

curcalptr:=paramrecalcind[xparam];

for k:=1 to origmax do
parm^[k,k]:=exogd[k].parm2;

while curcalptr<>nil do
	begin
		hcovar:= curcalptr^.hcov;
	 	sumproduct(hcovar,exogd,parm,priormean);
		curcalptr:=curcalptr^.next;
	end;

for k:=1 to origmax do
parm^[k,k]:=0;

for j:=1 to numob do
for k:=1 to numob do
if covtermind^[j,k]<>nil then gcov^[j,k]:=covtermind^[j,k]^.val
 else gcov^[j,k]:=0;

(* NB it is necessary to reload every value into gcov even though 
   not all values have changed, because the values in gcov are modified
   in the routines determ etc., in numeric.p			*) 

end;
 	 
	 	 


