#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "part.h" 
#include "matrices.h" 
#include "util.h"
#include "sys.h" 
#include "calctet.h" 
#include "io.h" 
#include "mon.h" 
   
procedure matmult(* var a,b,c: realarraynpbynp; rowa,cola,colb: vertextype *);
var i,j,k: vertextype;
begin
   for i := 1 to rowa do
     for j := 1 to colb do
       begin
          c[i,j] := 0;
          for k := 1 to cola do
              c[i,j] := c[i,j] + a[i,k] * b[k,j];
       end;
end;

procedure fillmatrix(* tail,head: vertextype; partialedon: partialedonarray;
          var matrixa: realarraynpbynp *);
var i,j: vertextype;
begin
    for i := 1 to partialedon.length+2 do matrixa[i][i] := 1.0;
    matrixa[1,2] := rho^[tail,head];
    for i := 3 to partialedon.length+2 do
       matrixa[1,i] := rho^[tail,partialedon.arr[i-2]];
    for i := 3 to partialedon.length+2 do
       matrixa[2,i] := rho^[head,partialedon.arr[i-2]];
    for i := 3 to partialedon.length+1 do
      for j := i+1 to partialedon.length+2 do
        matrixa[i,j] := rho^[partialedon.arr[i-2],partialedon.arr[j-2]];
    for i := 1 to partialedon.length+1 do
      for j := i+1 to partialedon.length+2 do
         matrixa[j,i] := matrixa[i,j];
end;

procedure mludcmp(var a: realarraynpbynp; n: integer; var indx: integerarraynp;
                 var d: real; var ok: boolean);
label 999;
const
   tiny = 1.0e-20;
var 
   k,j,imax,i: integer;
   sum,dum,big: real;
   vv: ^realarraynp;
begin
   ok := true;
   new(vv);
   d := 1.0;
   for i := 1 to n do
      begin
         big := 0.0;
         for j := 1 to n do
            if abs(a[i,j]) > big then big := abs(a[i,j]);
         if big = 0.0 then 
           begin
             errormess('Singular matrx.|',fatal,false);
             ok := false;
             goto 999;
           end;
         vv^[i] := 1.0/big;
      end;
   for j := 1 to n do 
      begin
        for i := 1 to j-1 do 
           begin
             sum := a[i,j];
             for k := 1 to i-1 do 
               sum := sum - a[i,k]*a[k,j];
             a[i,j] := sum;
           end;
         big := 0.0;
         for i := j to n do 
            begin
               sum := a[i,j];
               for k := 1 to j-1 do 
                 sum := sum - a[i,k]*a[k,j];
               a[i,j] := sum;
               dum := vv^[i]*abs(sum);
               if dum >= big then
                  begin
                    big := dum;
                    imax := i;
                  end;
             end;
          if j <> imax then
            begin
               for k := 1 to n do 
                 begin
                   dum := a[imax,k];
                   a[imax,k] := a[j,k];
                   a[j,k] := dum;
                 end;
               d := -d;
               vv^[imax] := vv^[j];
            end;
        indx[j] := imax;
        if a[j,j] = 0.0 then 
           begin
             a[j,j] := tiny;
             ok := false;
           end;
        if j <> n then
           begin
            dum := 1.0/a[j,j];
            for i := j+1 to n do
              a[i,j] := a[i,j]*dum;
           end;
      end;
      dispose(vv);
999: end;

procedure mlubksb(var a: realarraynpbynp; n: integer; var indx: integerarraynp;
                 var b: realarraynp);
var j,ip,ii,i: integer;
    sum: real;
begin
   ii := 0;
   for i := 1 to n do
      begin
         ip := indx[i];
         sum := b[ip];
         b[ip] := b[i];
         if ii <> 0 then
           for j := ii to i-1 do
             sum := sum - a[i,j] * b[j]
         else if sum <> 0.0 then
            ii := i;
         b[i] := sum;
       end;
   for i := n downto 1 do 
      begin
        sum := b[i];
        for j := i+1 to n do
            sum := sum - a[i,j] * b[j];
        b[i] := sum/a[i,i];
      end;
end;

procedure invert(* var a,y: realarraynpbynp; n: vertextype; var ok: boolean; *);
var i,j: integer;
    d: real;
    indx: integerarraynp; 
    col: realarraynp;
begin
    mludcmp(a,n,indx,d,ok);
    for j := 1 to n do 
       begin
         for i := 1 to n do col[i] := 0.0;
         col[j] := 1.0;
         mlubksb(a,n,indx,col);
         for i := 1 to n do y[i,j] := col[i];
       end;
end;

procedure newdeterminant(* var a: realarraynpbynp; n: vertextype; var determ: real; var ok: boolean *);
var j: integer; d: real;
    indx: integerarraynp;
begin
   mludcmp(a,n,indx,d,ok);
   for j := 1 to n do d := d * a[j,j];
   determ := d;
end;

procedure copycov(var cov1: covarray; var v1: realarraynpbynp; n: vertextype);
var i,j: vertextype;   
begin
   for i := 1 to n do
     for j := 1 to n do
       v1[i,j] := cov1^[i,j];
end;

procedure getdivergence(* var cov1,cov2: covarray; var divergence: real; var ok: boolean *);
var v1,v2,v3: realarraynpbynp; atrace,adeterm: real; i: integer;
begin
   copycov(cov1,v1,max);
   copycov(cov2,v3,max);
   invert(v3,v2,max,ok);
   matmult(v1,v2,v3,max,max,max);
   atrace := 0.0;
   for i := 1 to max do atrace := atrace + v3[i,i] - 1;
   newdeterminant(v3,max,adeterm,ok);
   writeln(atrace,' ',adeterm,' ',divergence);
   divergence := 1/2 * (atrace - ln(adeterm));
end;

procedure formcov(i: vertextype; var v1: realarraynpbynp; 
                  var adj: adjtype; var cov: covarray; var cola: vertextype);
var j: vertextype;
begin
   cola := 1;
   for j := 1 to max do 
     if adj^[j,i] <> notthere then 
      begin
        v1[1,cola] := cov^[i,j];
        cola := cola + 1;
      end;
   cola := cola - 1;
end;

procedure formparinverse(i: vertextype; var v1: realarraynpbynp; var adj: adjtype;
                         var cov: covarray; var rowno: vertextype; var ok: boolean);
var j,k,colno: vertextype; v2: realarraynpbynp; 
begin
  rowno := 0;
  for j := 1 to max do
     if adj^[j,i] <> notthere then 
       begin
          rowno := rowno + 1;
          colno := 1;
          for k := 1 to max do
              if adj^[k,i] <> notthere then
                 begin
                   v2[rowno,colno] := cov^[j,k];
                   colno := colno + 1;
                 end;
       end;
   if (rowno <> 0) then invert(v2,v1,rowno,ok);
end;

procedure linestimate(* var adj: adjtype; var cov,parm: covarray; 
                      var ok: boolean *);
var v1,v2,v3: realarraynpbynp; i,j,k,cola,colb: vertextype; 
begin
  for i := 1 to max do
    begin
       formcov(i,v1,adj,cov,cola);
       formparinverse(i,v2,adj,cov,colb,ok);
       if (colb <> 0) then
          begin
            matmult(v1,v2,v3,1,cola,colb);
            k := 1;
            for j := 1 to max do
               if adj^[j,i] <> notthere then
                  begin
                    parm^[j,i] := v3[1,k];
                    k := k + 1;
                 end
               else parm^[j,i] := 0.0;
         end
       else for j := 1 to max do parm^[j,i] := 0.0;
    end;
end;

procedure comparetotrue;
var ok: boolean; divergence: real;
begin
   linestimate(adjacency,rho,parm,ok);
   minternal := true;
   tnewcov(truecov,fatal);
   tnewcov(truecor,fatal);
   datatypes := [5];
   domonte;
   getknowledge;
   getdivergence(rho,truecor,divergence,ok);
   writeln(divergence);
   dispose(truecov);
   dispose(truecor);
   minternal := false;
end;

procedure calcr(v: vertextype; var partialedon: partialedonarray; 
                      var cov,parm: covarray; var ok: boolean; var r: real);
var v1,v2,v3: realarraynpbynp; i,j: vertextype; regrss: real; 
begin
   for i := 1 to partialedon.length do
     begin
      v1[1,i] := cov^[v,partialedon.arr[i]];
      for j := i to partialedon.length do
         begin
            v3[i,j] := cov^[partialedon.arr[i],partialedon.arr[j]];
            v3[j,i] := v3[i,j];
         end;
     end;
    invert(v3,v2,partialedon.length,ok);
    matmult(v1,v2,v3,1,partialedon.length,partialedon.length);
    regrss := 0;
    for i := 1 to partialedon.length do
      for j := i to partialedon.length do
         if i = j then
            regrss := regrss + v3[1,i] *v3[1,i] * cov^[partialedon.arr[i],partialedon.arr[i]]
         else regrss := regrss + 2 * v3[1,i] * v3[1,j] * cov^[partialedon.arr[i],partialedon.arr[j]];
    r := sqrt(regrss/cov^[v,v]);
end;



