(*-----------#unix-version-----------------------------------------------------*)

(*-----------#pc-version-----------------------------------------------------*)
(* implementation uses calctet,calctet1,io,util,rsutil,sys,indep; *)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h" 
#include "tt1.h" 
#include "tt2.h" 
#include "tt3.h"
#include "tt4.h" 
#include "tt5.h" 
#include "part.h" 
#include "calctet.h" 
#include "io.h" 
#include "util.h" 
#include "cspecial.h"
#include "indep1.h" 
#include "sys.h" 
#include "calctet1.h" 
#include "mon.h" 
#include "mon1.h" 
#include "newcalc.h" 
#include "up.h" 
#include "matrices.h"
#include "linbayes.h" 
#include "cross.h" 
#include "eqs.h"
#include "testbuildm.h"
#include "powell.h"
(*-----------#unix-version-----------------------------------------------------*)

const
    edgepath = 1;
    nonedgepath = 3;

type
    datatype = 0..60;
    edgeptr = ^edgeslist;
    edgeslist = record
                 edge: integer;
                 next: edgeptr;
                 last: edgeptr;
               end;
    parentstype = array[1..maxvertex] of vertexset;
    nodeptr = ^node;
    node = record
              prob: real;
              edge: integer;
              fwd,back: nodeptr;
            end;
    smallarray = array[1..25,1..25] of real;
    alphtype = array[0..maxvertex] of integer;
    settype = array[0..maxvertex] of vertexset;
    corrarray = array[1..maxvertex] of nodeptr;
var
    partscore				 : real;
    outfile				 : text;
    zeroes				 : dseparray;
    parents				 : parentstype;
    debug				 : boolean;
    heuristic				 : boolean;
    numchecked				 : integer;
    start				 : double;
    toomuchtime				 : boolean;
    deftrilist,posstrilist: triangleptr;
    corrlists,corrlists1		 : corrarray; 
    tempadj: adjtype;
    oldmod: modeltypeptr;
    bestin,bestout: tempset;

procedure tnewend(* var ptr: endarray; err: errortype *); 
var size: word;
begin
   size := sizeof(endarraytype);
   if (size < heapavail) then new(ptr)
   else errormess('Out of memory: Cannot run build.|',err,false); 
end;

procedure tnewdsep(* var ptr: dseparray; err: errortype *); 
var size: word; 
begin
   size := sizeof(dseparraytype);
   if (size < heapavail) then new(ptr)
   else errormess('Out of memory: Cannot run build.|',err,false); 
end;

procedure checktime; 
var now: double; 
begin
         now := tgettime;
         if maxtime <> -1 then
             if (now - start) > maxtime then toomuchtime := true; 
end;

function setsize(aset: vertexset): vertextype; 
var i: vertextype; num: integer; 
begin
  num := 0;
  for i := 1 to maxvertex do
     if i in aset then num := num + 1;
  setsize := num; 
end;

function getpower(base: real; exponent: integer): real; 
var i: integer; pow: real;
 begin
   pow := 1;
   for i := 1 to exponent do pow := pow * base;
   getpower := pow; 
end;

procedure matinv(matrixa: smallarray; var matrixb: smallarray;  var ok: boolean; size: vertextype); 
label 999; 
const
errorbound = 1.0E-10; 
var i,j,k,m: vertextype;
    factor,temp: real; 
begin
   for i := 1 to size do
     for j := 1 to size do
        if i = j then matrixb[i,j] := 1.0
        else matrixb[i,j] := 0.0;
   for j := 1 to size do
     begin
       i := j;
       while abs(matrixa[i,j]) < errorbound do
          begin
             if i = size then
               begin
                  ok := false;
                  goto 999;
               end;
             i := i + 1;
          end;
       for k := 1 to size do
           begin
              temp := matrixa[i,k];
              matrixa[i,k] := matrixa[j,k];
              matrixa[j,k] := temp;
              temp := matrixb[i,k];
              matrixb[i,k] := matrixb[j,k];
              matrixb[j,k] := temp;
           end;
        factor := 1.0 / matrixa[j,j];
        for k := 1 to size do
           begin
              matrixa[j,k] := factor * matrixa[j,k];
              matrixb[j,k] := factor * matrixb[j,k];
           end;
        for m := 1 to size do
           if m <> j then
             begin
                factor := -1 * matrixa[m,j];
                for k := 1 to size do
                   begin
                     matrixa[m,k] := matrixa[m,k] + factor * matrixa[j,k];
                     matrixb[m,k] := matrixb[m,k] + factor * matrixb[j,k];
                   end
             end
    end; 999: 
end;

procedure fillinmatrix(tail,head: vertextype; partialedon: partialedonarray;
          var matrixa: smallarray);
 var i,j: vertextype; 
begin
    for i := 1 to partialedon.length+2 do matrixa[i][i] := 1.0;
    matrixa[1,2] := rho^[tail,head];
    for i := 3 to partialedon.length+2 do
       matrixa[1,i] := rho^[tail,partialedon.arr[i-2]];
    for i := 3 to partialedon.length+2 do
       matrixa[2,i] := rho^[head,partialedon.arr[i-2]];
    for i := 3 to partialedon.length+1 do
      for j := i+1 to partialedon.length+2 do
        matrixa[i,j] := rho^[partialedon.arr[i-2],partialedon.arr[j-2]];
    for i := 1 to partialedon.length+1 do
      for j := i+1 to partialedon.length+2 do
         matrixa[j,i] := matrixa[i,j]; 
end;

function calcpart1(* partial: partialpair; partialedon: partialedonarray; var ok: boolean): real; *); 
var temp: real; matrixa,matrixb: realarraynpbynp; i: vertextype; 
begin
    ok := true;
    if partialedon.length = 0 then
          calcpart1:= rho^[partial[1],partial[2]]
    else
       begin
         fillmatrix(partial[1],partial[2],partialedon,matrixa);
         invert(matrixa,matrixb,partialedon.length+2,ok);
         if ok then
               calcpart1 := -1 * matrixb[1,2] / (sqrt(abs(matrixb[1,1] * matrixb[2,2])))
         else
           begin
		writename(partial[1],ofilvar);
		twrite(ofilvar,' |');
		writename(partial[2],ofilvar);
		twrite(ofilvar,' .|');
                for i := 1 to partialedon.length do
		   begin
			twrite(ofilvar,' |');
			writename(partialedon.arr[i],ofilvar);
		   end;
                twrite(ofilvar,' does not exist.|');
		twriteeoln(ofilvar);
           end;
       end; 
end;
         
procedure calcpartialheld(var partial,aprob: real; var held: boolean; length: integer); 
var z: real; 
begin
  begin
   if ((numchecked mod 100) = 0) then
      begin
         checktime;
         numchecked := 1;
      end
   else numchecked := numchecked + 1;
   if abs(partial) <= 1 then
          begin
              z := (1/2) * (sqrt(sampsize - 3 - length)) * ln((1 + abs(partial)) /
                   (1 - abs(partial)));
              getprob(aprob,abs(z));
              held := aprob >= signif;
          end
   else held := false;
 end; end; { procedure }

procedure outpart(partial,pofstdev: real; part: partialpair;
                  partialedon: partialedonarray; impld,hold: boolean);
var field,i: integer; impldcheck,heldcheck: char; 
begin
         setcheck(impld,impldcheck);
         setcheck(hold,heldcheck);
         if partial >= 0 then field := 4 else field := 3;
	 writename(part[1],ofilvar);
         twrite(ofilvar,' |');
	 writename(part[2],ofilvar);
	 twrite(ofilvar,' . |');
         for i := 1 to numob-2 do
           if i in graphvertices then
             if i <= partialedon.length then
                writename(partialedon.arr[i],ofilvar)
             else twrite(ofilvar,' |');
         tpwritereal(ofilvar,partial,3,field,false);
	 twrite(ofilvar,' |');
	 twritechar(ofilvar,impldcheck);
	 twrite(ofilvar,' |');
	 twritechar(ofilvar,heldcheck);
	 twrite(ofilvar,' |');
	 tpwritereal(ofilvar,pofstdev,7,4,false) 
end;

function collider(i,j,k: integer; var ugraph: endarray): boolean;
begin
   if (knownin2 in ugraph^[i,k]) and (knownin2 in ugraph^[j,k]) then
      collider := true
   else if ((knownout2 in ugraph^[i,k]) or (knownout2 in ugraph^[j,k])) then
      collider := false
   else if (i < j) then 
      collider := (ugraph^[i,k] <> []) and (ugraph^[j,k] <> []) and
               not (k in dsep^[i,j])
   else collider := (ugraph^[i,k] <> []) and (ugraph^[j,k] <> []) and
               not (k in dsep^[j,i]); 
end;

procedure readarraytoset(var dsepcell: vertexset; partialedon:
partialedonarray); var i: vertextype; begin
   dsepcell := [];
   for i := 1 to partialedon.length do
      dsepcell := dsepcell + [partialedon.arr[i]]; 
end;

procedure setdirection(tail,head: vertextype); 
begin 
end;

function triangle(x,y,z: vertextype; var ugraph: endarray): boolean;
begin
   triangle := (ugraph^[x,y] <> []) and (ugraph^[y,z] <> []) and
               (ugraph^[x,z] <> []); 
end;

function innocollidelist(x,y,z: vertextype; var nocollidelist: triangleptr): triangleptr; 
var curpos: triangleptr; found: boolean;
begin
   innocollidelist := nil;
   found := false;
   curpos := nocollidelist;
   while (curpos <> nil) and (not found) do
     begin
       if (curpos^.unori = y) and
          (((curpos^.other = x) and (curpos^.inout = z)) or
           ((curpos^.other = z) and (curpos^.inout = x))) then found := true;
       if found then innocollidelist := curpos
       else curpos := curpos^.fwd;
     end; 
end;

procedure addtolist(x,y,z: vertextype; var trilist: triangleptr); 
label 999;
var newtri,curpos: triangleptr; 
begin
   new(newtri);
   newtri^.inout := x;
   newtri^.unori := y;
   newtri^.other := z;
   newtri^.fwd := nil;
   if trilist = nil then 
     begin
       trilist := newtri;
       trilist^.back := nil;
     end
   else
       begin
         curpos := trilist;
         repeat
            if (y = curpos^.unori) and 
               (((x = curpos^.inout) and (z = curpos^.other)) or
                ((z = curpos^.inout) and (x = curpos^.other))) then
                   begin
                     dispose(newtri);
                     goto 999;
                   end;
            if curpos^.fwd <> nil then curpos := curpos^.fwd;
        until curpos^.fwd = nil;
        newtri^.back := curpos;
        if curpos <> nil then curpos^.fwd := newtri;
      end;
999: end;

procedure putinlist(x,y,z: vertextype; var trilist: triangleptr); 
var newtri: triangleptr; 
begin
   new(newtri);
   newtri^.inout := x;
   newtri^.unori := y;
   newtri^.other := z;
   newtri^.back := nil;
   if trilist <> nil then trilist^.back := newtri;
   newtri^.fwd := trilist;
   trilist := newtri; 
end;

procedure formtri(x,y,z: vertextype; var deftrilist,posstrilist:triangleptr; var ugraph: endarray); 
begin
   if (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
       ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x]))) then
            if ((notknown in ugraph^[x,y]) or (notknown in ugraph^[z,y])) then
                putinlist(x,y,z,deftrilist);
   if (((in2 in ugraph^[z,x]) or (knownin2 in ugraph^[z,x])) and
       ((out2 in ugraph^[y,x]) or (knownout2 in ugraph^[y,x]))) then
            if ((notknown in ugraph^[x,z]) or (notknown in ugraph^[y,z])) then
                putinlist(x,z,y,deftrilist);
   if (not (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
           ((in2 in ugraph^[z,x]) or (knownin2 in ugraph^[z,x])))) and
      (not (((out2 in ugraph^[y,x]) or (knownout2 in ugraph^[y,x])) and
           ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x])))) then
         begin
            if ((notknown in ugraph^[x,y]) or (notknown in ugraph^[z,y])) then
                putinlist(x,y,z,posstrilist);
            if ((notknown in ugraph^[x,z]) or (notknown in ugraph^[y,z])) then
                putinlist(x,z,y,posstrilist);
         end; 
end;

procedure inittrilists(var deftrilist,posstrilist,nocollidelist,bothlist: triangleptr); 
var i,j,k: vertextype; 
begin
  deftrilist := nil;
  posstrilist := nil;
  nocollidelist := nil;
  bothlist := nil; 
end;

procedure removetri(var curtri,trilist: triangleptr); 
begin
    if curtri^.back <> nil then curtri^.back^.fwd := curtri^.fwd;
    if curtri^.fwd <> nil then curtri^.fwd^.back := curtri^.back;
    if curtri = trilist then trilist := trilist^.fwd; 
end;

procedure updateposstrilist(var deftrilist,posstrilist,nocollidelist: triangleptr;
                            var nochange: boolean; var ugraph: endarray); 
var curpos,next: triangleptr; x,y,z: vertextype; 
begin
   curpos := posstrilist;
   while (curpos <> nil) do
     begin
         x := curpos^.inout;
         y := curpos^.unori;
         z := curpos^.other;
         next := curpos^.fwd;
         if (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
             ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x]))) and
            ((notknown in ugraph^[x,y]) or (notknown in ugraph^[z,y])) and
            (innocollidelist(x,y,z,nocollidelist) = nil) then
                begin
                   removetri(curpos,posstrilist);
                   curpos^.fwd := deftrilist;
                   curpos^.back := nil;
                   if deftrilist <> nil then deftrilist^.back := curpos;
                   deftrilist := curpos;
                   nochange := false;
                end
         else if (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
             ((in2 in ugraph^[z,x]) or (knownin2 in ugraph^[z,x]))) or
            (((out2 in ugraph^[y,x]) or (knownout2 in ugraph^[y,x])) and
             ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x]))) or
            ((not (notknown in ugraph^[x,y])) and (not (notknown in ugraph^[z,y]))) or
            (innocollidelist(x,y,z,nocollidelist) <> nil) then
                begin
                  removetri(curpos,posstrilist);
                  dispose(curpos);
                end;
          curpos := next;
     end; 
end;

procedure setnocollide(i,k,j: vertextype; var deftrilist,nocollidelist: triangleptr); 
var curtri: triangleptr; 
begin
   new(curtri);
   curtri^.inout := i;
   curtri^.other := k;
   curtri^.unori := j;
   curtri^.fwd := nocollidelist;
   if nocollidelist <> nil then nocollidelist^.back := curtri;
   curtri^.back := nil;
   nocollidelist := curtri; 
end;

procedure adddedge(tail,head: edge; var adj: adjtype); 
var i,j: vertextype; newtailpaths: array[1..maxvertex] of vertextype;
begin
   for i := 1 to numob do
       for j := 1 to numob do
         if (i in graphvertices) and (j in graphvertices) then
          if ((i <> tail) or (adj^[i,tail] <> notthere)) and
             ((j <> head) or (adj^[head,j] <> notthere)) then
                adj^[i,j] := nonedgepath;
   adj^[tail,head] := edgepath; 
end;

procedure updatepaths(var localadj: adjtype; tail,head: vertextype; var ok: boolean; level: integer;
                      var ugraph: graphrec); 
var i,j: vertextype; 
begin
  for i := 1 to max do
    if (i in graphvertices) then
     for j := 1 to max do
       if (j in graphvertices) then
        if (localadj^[i,tail] <> notthere) and (localadj^[head,j] <> notthere) then
             begin
                 localadj^[i,j] := level;
                 ugraph.ancestset[j] := ugraph.ancestset[j] + [i];
                 ugraph.descend[i] := ugraph.ancestset[i] + [j];
                 if localadj^[j,i] <> notthere then 
                    ok := false;
             end;
  localadj^[tail,head] := level;
  if localadj^[head,tail] <> notthere then ok := false; 
end;

procedure recordchanges(tail,head: vertextype; bothends: boolean; 
                        var ugraph: graphrec); 
begin
  with ugraph do
   begin
    if bothends and (notknown in augraph^[head,tail]) then
        begin
           recchange^[curchange] := head;
           recchange^[curchange+1] := tail;
           curchange := curchange + 2;
        end;
    if (notknown in augraph^[tail,head]) then
        begin
           recchange^[curchange] := tail;
           recchange^[curchange+1] := head;
           curchange := curchange + 2;
        end;
   end;
end;

procedure propagate(tail,head: vertextype; bothends: boolean; var ok: boolean;
                   var ugraph: graphrec; level: integer); 
var i: integer; curpos: triangleptr;
begin
  with ugraph do
   begin
    if (out2 in augraph^[tail,head]) then ok := false;
    if (bothends and (in2 in augraph^[head,tail])) then ok := false;
    if ok then
      begin
        recordchanges(tail,head,bothends,ugraph);
       if bothends then augraph^[head,tail] := augraph^[head,tail] + [out2] - [notknown];
       augraph^[tail,head] := augraph^[tail,head] + [in2] - [notknown];
       if bothends then updatepaths(ancest,tail,head,ok,level,ugraph);
       for i := 1 to numob do
         if (i in graphvertices) then
           if (i <> tail) and (i <> head) and
              (not (out2 in augraph^[tail,head])) and
              (not (knownout2 in augraph^[tail,head])) and
              (augraph^[i,head] <> []) and
              (not (in2 in augraph^[i,head])) and
              (not (knownin2 in augraph^[i,head])) and
              (not (out2 in augraph^[head,i])) and
              (not (knownout2 in augraph^[head,i])) and
              (not ((in2 in augraph^[head,i]) and (out2 in augraph^[i,head]))) then
                  if augraph^[tail,i] = [] then
                     propagate(head,i,true,ok,ugraph,level)
                  else
                     begin
                        curpos := innocollidelist(tail,head,i,nocollidelist);
                        if curpos <> nil then
                           begin 
                             (* removetri(curpos,nocollidelist); *)
                             propagate(head,i,true,ok,ugraph,level);
                           end;
                     end; 
        end;
     end;
end;

procedure finddiscrim(a,b,c: vertextype; var ugraph: endarray; var foundone: boolean;
                      var endpoint: vertextype); 
label 999; 
var i,lastadded,curpos,curvertex: vertextype; expand: vertexset;
    expandlist: array[0..maxvertex] of vertextype; 
begin
  expand := [a];
  lastadded := 1;
  expandlist[lastadded] := a;
  curpos := 1;
  foundone := false;
  repeat
    curvertex := expandlist[curpos];
    for i := 1 to max do
       begin
         if (not (i in expand)) and (i <> b) and (i <> c) then
            if (in2 in ugraph^[i,curvertex]) and
               (((in2 in ugraph^[i,b]) and (out2 in ugraph^[b,i])) or (ugraph^[i,b] = [])) then
              begin
                if (in2 in ugraph^[curvertex,i]) and (not (i in expand)) then
                   begin
                      lastadded := lastadded + 1;
                      expandlist[lastadded] := i;
                      expand := expand + [i];
                   end;
                if ugraph^[i,b] = [] then
                     begin
                        foundone := true;
                        endpoint := i;
                        goto 999;
                    end;
              end;
        end;
     curpos := curpos + 1;
  until (curpos > lastadded) or foundone; 
999: end;


(* procedure discrimpath(var curtri,deftrilist,nocollidelist: triangleptr;
                      var curpath: vertexset; nearend,curend: vertextype;
                      var founddis,nochange,ok: boolean; var adj: adjtype; level: integer);
var i,other: vertextype; nextri: triangleptr; 
begin
     finddiscrim(curtri^.inout,curtri^.other,curtri^.unori,ugraph,founddis,i);
     if founddis then
        if (curtri^.unori in dsep^[i,curtri^.other]) then
          begin
            if (in2 in ugraph^[curtri^.inout,curtri^.unori]) or
               (knownin2 in ugraph^[curtri^.inout,curtri^.unori]) then
                  begin
                      nextri := curtri^.fwd;
                      propagate(curtri^.unori,curtri^.other,true,adj,ok,nocollidelist,level);
                      removetri(curtri,deftrilist);
                      dispose(curtri);
                      curtri := nextri;
                  end
            else if (in2 in ugraph^[curtri^.other,curtri^.unori]) or
               (knownin2 in ugraph^[curtri^.other,curtri^.unori]) then
                   begin
                       nextri := curtri^.fwd;
                       propagate(curtri^.unori,curtri^.inout,true,adj,ok,nocollidelist,level);
                       removetri(curtri,deftrilist);
                       dispose(curtri);
                       curtri := nextri;
                   end
            else setnocollide(curtri,deftrilist,nocollidelist);
            end
         else
            begin
               nextri := curtri^.fwd;
               propagate(curtri^.inout,curtri^.unori,false,adj,ok,nocollidelist,level);
               propagate(curtri^.other,curtri^.unori,false,adj,ok,nocollidelist,level);
               removetri(curtri,deftrilist);
               dispose(curtri);
               curtri := nextri;
            end; 
end; *)

procedure discrimpath(var nochange,ok: boolean; var ugraph: graphrec; setcoll: boolean; level: integer); 
var farend,other: vertextype; nextri: triangleptr; i,j,k,l: vertextype; founddis: boolean; 
begin
 with ugraph do
  for i := 1 to max do
     for j := i to max do
        for k := j to max do
           if (i <> j) and (j <> k) and (i <> k) and (out2 in augraph^[k,i]) and
              (in2 in augraph^[j,i]) and (augraph^[j,k] <> []) and 
                 ((notknown in augraph^[i,j]) or (notknown in augraph^[k,j])) then
              begin
                  finddiscrim(i,k,j,augraph,founddis,farend);
                  if founddis then
                     if setcoll then
                        begin
                         if (j in dsep^[farend,k]) then
                            begin
                              if (in2 in augraph^[i,j]) then 
                                 propagate(j,k,true,ok,ugraph,level)
                              else if (in2 in augraph^[k,j]) then
                                 propagate(j,i,true,ok,ugraph,level)
                              else setnocollide(i,k,j,deftrilist,nocollidelist);
                            end
                        else
                           begin
                              propagate(i,j,false,ok,ugraph,level);
                              propagate(k,j,false,ok,ugraph,level);
                          end
                        end
                     else addtolist(i,j,k,bothlist);
              end; 
end;

procedure orientfours(i,j,k,l: vertextype; var nochange,ok: boolean; 
                      var ugraph: graphrec; level: integer); 
begin
  with ugraph do
    begin
       if (augraph^[j,k] = []) and
          (augraph^[i,j] <> []) and (augraph^[i,k] <> []) and
          (not (in2 in augraph^[j,i]) or (not (in2 in augraph^[k,i]))) and
          (in2 in augraph^[j,l]) and (in2 in augraph^[k,l]) and
          (augraph^[i,l] <> []) then
              if (not (in2 in augraph^[i,l])) then
                 begin
                     propagate(i,l,false,ok,ugraph,level);
                     nochange := false;
                 end;
        if (augraph^[j,l] = []) and
           (augraph^[i,j] <> []) and (augraph^[j,l] <> []) and
           (not (in2 in augraph^[i,j]) or (not (in2 in augraph^[j,l]))) and
           (out2 in augraph^[j,l]) and (in2 in augraph^[k,l]) and
           (augraph^[i,l] <> []) then
               if (not (in2 in augraph^[i,j])) then
                 begin
                    nochange := false;
                    propagate(i,j,false,ok,ugraph,level);
                 end; 
    end;
end;

procedure orientfoursome(nochange,ok: boolean; var ugraph: graphrec; 
                         level: integer); 
var i,j,k,l: vertextype; 
begin
   for i := 1 to max-3 do
     for j := i+1 to max-2 do
        for k := j+1 to max-1 do
           for l := k+1 to max do
             begin
               orientfours(i,j,k,l,nochange,ok,ugraph,level);
               orientfours(i,j,l,k,nochange,ok,ugraph,level);
               orientfours(i,k,j,l,nochange,ok,ugraph,level);
               orientfours(i,k,l,j,nochange,ok,ugraph,level);
               orientfours(i,l,j,k,nochange,ok,ugraph,level);
               orientfours(i,l,k,j,nochange,ok,ugraph,level);
               orientfours(j,i,k,l,nochange,ok,ugraph,level);
               orientfours(j,i,l,k,nochange,ok,ugraph,level);
               orientfours(j,k,i,l,nochange,ok,ugraph,level);
               orientfours(j,k,l,i,nochange,ok,ugraph,level);
               orientfours(j,l,i,k,nochange,ok,ugraph,level);
               orientfours(j,l,k,i,nochange,ok,ugraph,level);
               orientfours(k,i,j,l,nochange,ok,ugraph,level);
               orientfours(k,i,l,j,nochange,ok,ugraph,level);
               orientfours(k,j,i,l,nochange,ok,ugraph,level);
               orientfours(k,j,l,i,nochange,ok,ugraph,level);
               orientfours(k,l,i,j,nochange,ok,ugraph,level);
               orientfours(k,l,j,i,nochange,ok,ugraph,level);
               orientfours(l,i,j,k,nochange,ok,ugraph,level);
               orientfours(l,i,k,j,nochange,ok,ugraph,level);
               orientfours(l,j,i,k,nochange,ok,ugraph,level);
               orientfours(l,j,k,i,nochange,ok,ugraph,level);
               orientfours(l,k,i,j,nochange,ok,ugraph,level);
               orientfours(l,k,j,i,nochange,ok,ugraph,level);
             end; 
end;

function zcycle(var newadj: adjtype; first,second: vertextype; var ugraph: endarray): boolean; 
label 999; 
var j,k,i: integer;
begin (* calcpath *)
    zcycle := false;
    if first = second then goto 999;
    for i := 1 to max do
       for j := 1 to max do newadj^[i,j] := notthere;
    for i := 1 to numobgraph - 1 do
      for j := i+1 to numobgraph do
         if (i <> j) then
              if (in2 in ugraph^[graphlist[i],graphlist[j]]) and (out2 in ugraph^[graphlist[j],graphlist[i]]) then
		 newadj^[graphlist[i],graphlist[j]] := 1
	      else if (in2 in ugraph^[graphlist[j],graphlist[i]]) and (out2 in ugraph^[graphlist[i],graphlist[j]]) then
		 newadj^[graphlist[j],graphlist[i]] := 1
              else
                 begin
                   newadj^[graphlist[i],graphlist[j]] := notthere;
                   newadj^[graphlist[j],graphlist[i]] := notthere;
                 end;
    newadj^[first,second] := 1;
    for k := 1 to numobgraph do
       for i := 1 to numobgraph do
           for j := 1 to numobgraph do
               if (newadj^[graphlist[i],graphlist[k]] <> notthere) and (newadj^[graphlist[k],graphlist[j]] <> notthere) then
                    begin
                      newadj^[graphlist[i],graphlist[j]] := 1;
                      if newadj^[graphlist[j],graphlist[i]] <> notthere then
                         begin
                            zcycle := true;
                            goto 999;
                         end;
                     end; 
999:end; (* Calcdpath *)

procedure suffprop(tail,head: vertextype; var adj: adjtype; var ugraph: endarray); 
label 999; 
var i: integer; 
begin
    if not zcycle(tempadj,tail,head,ugraph) then
        begin
            ugraph^[head,tail] := ugraph^[head,tail] + [out2] - [notknown];
            ugraph^[tail,head] := ugraph^[tail,head] + [in2] - [notknown];
        end
    else
        begin
            ugraph^[head,tail] := ugraph^[head,tail] + [in2] - [notknown];
            ugraph^[tail,head] := ugraph^[tail,head] + [out2] - [notknown];
        end;
    for i := 1 to numob do
     if (i in graphvertices) then
      if (i <> tail) and (i <> head) and
         (not (out2 in ugraph^[tail,head])) and
         (not (knownout2 in ugraph^[tail,head])) and
         (ugraph^[i,head] <> []) and (ugraph^[tail,i] = []) and
         (not (in2 in ugraph^[i,head])) and
         (not (knownin2 in ugraph^[i,head])) and
         (not (out2 in ugraph^[head,i])) and
         (not (knownout2 in ugraph^[head,i])) and
         (not ((in2 in ugraph^[head,i]) and (out2 in ugraph^[i,head]))) then
             suffprop(head,i,adj,ugraph); 
999: end;

procedure getadjacencies(var ugraph: endarray; v: vertextype; var adjset: vertexset;
                         var numadj: integer); 
var i: vertextype;
begin
   numadj := 0;
   adjset := [];
   for i := 1 to max do
     if (i in graphvertices) then
      if (i <> v) and (ugraph^[i,v] <> []) then
         begin
            adjset := adjset + [i];
            numadj := numadj + 1;
         end; 
end;
      
procedure calccolliders(adjset: vertexset; v: vertextype; var collidera: adjtype;
                        var dsep: dseparray; ugraph: endarray); 
var i,j,k,index: vertextype;  g2,aprob,aval: real; partial: partialpair;  partialedon: partialedonarray; 
    held: boolean; ok: boolean; 
(* collidera: 1 is non-collider, 2 is collider, 3 is ambiguous *)
begin
   for i := 1 to max do
     for j := 1 to max do
        if (i <> j) and (i in adjset) and (j in adjset)
           and (ugraph^[i,j] = []) then
              if v in dsep^[i,j] then collidera^[i,j] := 1
              else
                 begin
                    partial[1] := i;
                    partial[2] := j;
                    partialedon.length := 0;
                    index := 0;
                    for k := 1 to max do
                       if k in dsep^[i,j] then
                         begin
                           index := index + 1;
                           partialedon.length := partialedon.length + 1;
                           partialedon.arr[index] := k;
                         end;
                     partialedon.length := partialedon.length + 1;
                     partialedon.arr[index+1] := v;
                     if discrete then
                        held := calcindep1(partial,aprob,g2,partialedon,numchecked)
                     else if (not discrete) then
                        begin
                          aval := calcpart1(partial,partialedon,ok);
                          if ok then calcpartialheld(aval,aprob,held,partialedon.length)
                          else held := false;
                        end;
                     if held then collidera^[i,j] := 3
                     else collidera^[i,j] := 2;
                 end
        else collidera^[i,j] := 3;
 end;

procedure calcimplied(orientset,adjset: vertexset; var collidera: adjtype; var thescore: integer;
                      ugraph: endarray);
var i,j: vertextype; 
begin
  thescore := 0;
  for i := 1 to max-1 do
     for j := i+1 to max do
        if (i <> j) and (i in adjset) and (j in adjset) and (ugraph^[i,j] = []) then
           if (i in orientset) and (j in orientset) then
             begin
               if collidera^[i,j] = 2 then
                  thescore := thescore + 1
               else if collidera^[i,j] = 1 then
                  thescore := thescore - 1
             end
           else
             begin
               if collidera^[i,j] = 1 then
                  thescore := thescore + 1
               else if collidera^[i,j] = 2 then
                  thescore := thescore - 1;
             end;
end;

procedure findbest(var orientset,adjset: vertexset; var bestin,bestout: tempset;
                   var collidera: adjtype; var thescore,bestscore: integer; v: integer;
                   var ugraph: endarray); 
begin
    calcimplied(orientset,adjset,collidera,thescore,ugraph);
    if thescore > bestscore then
       begin
          bestscore := thescore;
          bestin[v] := orientset;
          bestout[v] := adjset - orientset
       end
    else if thescore = bestscore then
       begin
          bestin[v] := bestin[v] + orientset;
          bestout[v] := bestout[v] + (adjset - orientset);
       end; 
end;

procedure checkorientation(var orientset,adjset: vertexset; var bestin,bestout: tempset;
               var thescore,bestscore: integer; start,depth,numadj,v: integer;
               var collidera: adjtype; var ugraph: endarray); 
var i: integer; 
begin 
for i := start to max do
 if (i in adjset) then
   begin
     if depth <> numadj then
        checkorientation(orientset,adjset,bestin,bestout,
            thescore,bestscore,i+1,depth+1,numadj,v,collidera,ugraph)
     else findbest(orientset,adjset,bestin,bestout,collidera, thescore,bestscore,v,ugraph);
     orientset := orientset + [i];
     if depth <> numadj then
        checkorientation(orientset,adjset,bestin,bestout,
            thescore,bestscore,i+1,depth+1,numadj,v,collidera,ugraph)
     else findbest(orientset,adjset,bestin,bestout,collidera,thescore,bestscore,v,ugraph);
     orientset := orientset - [i];
   end; 
end;


procedure setugraph(var bestin,bestout: tempset; var ugraph: endarray; toomanyadj: vertexset); 
var i,j: vertextype; 
begin
for i := 1 to max do
 if (not (i in toomanyadj)) then
  for j := 1 to max do
    if j in bestin[i] - bestout[i] then
       begin
          tempadj^[j,i] := 1;
          if not zcycle(tempadj,j,i,ugraph) then
             begin
               if causalsuff then ugraph^[i,j] := ugraph^[i,j] + [out2] - [notknown];
               ugraph^[j,i] := ugraph^[j,i] + [in2] - [notknown];
            end
          else
             begin
               ugraph^[i,j] := ugraph^[i,j] + [in2] - [notknown];
               if causalsuff then ugraph^[j,i] := ugraph^[j,i] + [out2] - [notknown];
             end
       end
    else if j in bestout[i] - bestin[i] then 
       begin
          tempadj^[i,j] := 1;
          if not zcycle(tempadj,i,j,ugraph) then
             begin
               ugraph^[i,j] := ugraph^[i,j] + [in2] - [notknown];
               ugraph^[j,i] := ugraph^[j,i] + [out2] - [notknown];
             end
          else
             begin
               ugraph^[i,j] := ugraph^[i,j] + [out2] - [notknown];
               ugraph^[j,i] := ugraph^[j,i] + [in2] - [notknown];
            end
       end 
end;
             
procedure setcollider2(dopropagate: boolean; var nochange: boolean; 
                      var ok: boolean; var ugraph: graphrec; level: integer; k: integer); 
var i,j,start,thescore,bestscore,numadj: integer;
    adjset,orientset: vertexset; bestin,bestout: tempset; 
begin
with ugraph do
 begin
  for i := 1 to numob do
   for j := 1 to numob do
        if (i in graphvertices) and (j in graphvertices) and (k in graphvertices) then
           if (i <> j) and (j <> k) and (i <> k) then
              if augraph^[i,j] = [] then
                 if collider(i,j,k,augraph) then
                    begin
                       augraph^[i,k] := augraph^[i,k] + [in2] - [notknown];
                       augraph^[j,k] := augraph^[j,k] + [in2] - [notknown];
                    end; 
  end;
end;

procedure setcollider(dopropagate: boolean; var nochange: boolean; 
                      var ok: boolean; var ugraph: graphrec; level: integer); 
var i,j,k,start,thescore,bestscore,numadj: integer;
    adjset,orientset: vertexset; bestin,bestout: tempset; toomanyadj: vertexset;
begin
  toomanyadj := [];
  for i := 1 to max do 
    if (i in graphvertices) then
       begin
          bestin[i] := [];
          bestout[i] := [];
          getadjacencies(ugraph.augraph,i,adjset,numadj);
          if (numadj > 10) then
             begin
                 setcollider2(dopropagate,nochange,ok,ugraph,level,i);
                 toomanyadj := toomanyadj + [i];
             end
          else
             begin
                calccolliders(adjset,i,collidera,dsep,ugraph.augraph);
                orientset := [];
                start := 1;
                bestscore := -1000000;
                checkorientation(orientset,adjset,bestin,bestout,thescore,bestscore,start,1,numadj,i,collidera,ugraph.augraph);
             end;
       end;
  setugraph(bestin,bestout,ugraph.augraph,toomanyadj);
  with ugraph do
  begin
  if dopropagate then
   for i := 1 to numob do
     for j := 1 to numob do
       for k := 1 to numob do
        if (i in graphvertices) and (j in graphvertices) and (k in graphvertices) and 
           (not (k in toomanyadj)) then
         if (i <> j) and (j <> k) and (i <> k) then
           if ((in2 in augraph^[i,j]) or (knownin2 in augraph^[i,j])) and
              (not (out2 in augraph^[i,j])) and
              (not (knownout2 in augraph^[i,j])) and
              (augraph^[j,k] <> []) and (not (out2 in augraph^[j,k])) and
              (not (knownout2 in augraph^[j,k])) and
              (not (in2 in augraph^[k,j])) and
              (not (knownin2 in augraph^[k,j])) and
              (augraph^[i,k] =[]) then
                 if not causalsuff then
                    propagate(j,k,true,ok,ugraph,level)
                 else suffprop(j,k,ancest,augraph); 
  end;
end;

procedure setcollider1(dopropagate: boolean; var nochange: boolean; 
                      var ok: boolean; var ugraph: graphrec; level: integer); 
var i,j,k,start,thescore,bestscore,numadj: integer;
    adjset,orientset: vertexset; bestin,bestout: tempset; 
begin
with ugraph do
 begin
  for i := 1 to numob do
   for j := 1 to numob do
     for k := 1 to numob do
        if (i in graphvertices) and (j in graphvertices) and (k in graphvertices) then
           if (i <> j) and (j <> k) and (i <> k) then
              if augraph^[i,j] = [] then
                 if collider(i,j,k,augraph) then
                    begin
                       augraph^[i,k] := augraph^[i,k] + [in2] - [notknown];
                       augraph^[j,k] := augraph^[j,k] + [in2] - [notknown];
                    end; 
  if dopropagate then
   for i := 1 to numob do
     for j := 1 to numob do
       for k := 1 to numob do
        if (i in graphvertices) and (j in graphvertices) and (k in graphvertices) then
         if (i <> j) and (j <> k) and (i <> k) then
           if ((in2 in augraph^[i,j]) or (knownin2 in augraph^[i,j])) and
              (not (out2 in augraph^[i,j])) and
              (not (knownout2 in augraph^[i,j])) and
              (augraph^[j,k] <> []) and (not (out2 in augraph^[j,k])) and
              (not (knownout2 in augraph^[j,k])) and
              (not (in2 in augraph^[k,j])) and
              (not (knownin2 in augraph^[k,j])) and
              (augraph^[i,k] =[]) then
                 if not causalsuff then
                    propagate(j,k,true,ok,ugraph,level)
                 else suffprop(j,k,ancest,augraph); 
  end;
end;

(*********************************************************)

procedure calcdpath(var ugraph: graphrec; var ok: boolean); 
label 999; 
var j,k,i: integer;
begin (* calcpath *)
    for i := 1 to numob do 
       begin
         ugraph.ancestset[i] := [i];
         ugraph.descend[i] := [i];
       end;
    for i := 1 to numob do
      for j := 1 to numob do
       if (i in graphvertices) and (j in graphvertices) then
        if i <> j then
           begin
              if (in2 in ugraph.augraph^[i,j]) and (out2 in ugraph.augraph^[j,i]) then 
                 begin
                   ugraph.ancest^[i,j] := edgepath;
                   ugraph.ancestset[j] := ugraph.ancestset[j] + [i];
                   ugraph.descend[i] := ugraph.descend[i] + [j];
                 end
              else ugraph.ancest^[i,j] := notthere;
           end
       else ugraph.ancest^[i,j] := nonedgepath;
    for k := 1 to numob do
       for i := 1 to numob do
           for j := 1 to numob do
             if (i in graphvertices) and (j in graphvertices) and (k in graphvertices)
                and (i <> j) and (j <> k) and (i <> k) then
               if (ugraph.ancest^[i,k] <> notthere) and
                  (ugraph.ancest^[k,j] <> notthere) then
                      begin
                         ugraph.ancest^[i,j] := nonedgepath;
			 ugraph.ancestset[j] := ugraph.ancestset[j] + [i];
			 ugraph.descend[i] := ugraph.descend[i] + [j];
                         if (ugraph.ancest^[j,i] <> notthere) then ok := false;
                      end; 
999: end; (* Calcdpath *)

procedure orientaway(i,j,k: vertextype; var nochange: boolean; 
                     var ok: boolean; var ugraph: graphrec; level: integer); 
begin
  with ugraph do
    begin
     if (i in graphvertices) and (j in graphvertices) and
        (out2 in augraph^[j,i]) and (out2 in augraph^[k,j]) and 
        (augraph^[i,k] <> []) then
          if (not (out2 in augraph^[k,i])) or (not (in2 in augraph^[i,k])) then
             begin
                 propagate(i,k,true,ok,ugraph,level);
                 nochange := false;
             end; 
    end;
end;

procedure orientdouble(var nochange,ok: boolean; var ugraph: graphrec; 
                       level: integer); 
var i,j,k: vertextype; 
begin
with ugraph do
 begin
  for i := 1 to max do
   for j := 1 to max do
     for k := 1 to max do
        if (i <> j) and (i <> k) and (j <> k) then
           if (in2 in augraph^[i,k]) and (in2 in augraph^[k,i]) and (ancest^[i,j] <> notthere)
              and (in2 in augraph^[j,k]) and (notknown in augraph^[k,j]) then
                begin
                   propagate(k,j,false,ok,ugraph,level);
                   nochange := false;
                end; 
 end;
end;

procedure orientawaycycle(var nochange,ok: boolean; 
                          var ugraph: graphrec; level: integer); 
var i,j,k: vertextype; 
begin
 for i := 1 to max-2 do
   for j := i+1 to max do
     for k := j+1 to max do
        begin
           orientaway(i,j,k,nochange,ok,ugraph,level);
           orientaway(i,k,j,nochange,ok,ugraph,level);
           orientaway(j,i,k,nochange,ok,ugraph,level);
           orientaway(j,k,i,nochange,ok,ugraph,level);
           orientaway(k,i,j,nochange,ok,ugraph,level);
           orientaway(k,j,i,nochange,ok,ugraph,level);
        end; 
end;

procedure orientedges(var ugraph: graphrec; setcoll: boolean; level: integer; oldcoll: boolean);
label 999;
var nochange,founddis,ok: boolean; curpath: vertexset; curpos: triangleptr;
    i,j: vertextype; 
begin
 with ugraph do
  begin
(*    nocollidelist := nil; *)
    nochange := true;
    ok := true;
    if setcoll then 
        if oldcoll then setcollider1(true,nochange,ok,ugraph,1)
        else setcollider(true,nochange,ok,ugraph,level);
    repeat
      if aaabort then goto 999;
      if maxtime <> -1 then checktime; 
      if toomuchtime then goto 999;
      nochange := true;
      if nochange then orientawaycycle(nochange,ok,ugraph,level); 
      if nochange then 
             discrimpath(nochange,ok,ugraph,setcoll,level);
      if nochange then orientdouble(nochange,ok,ugraph,level);
      if nochange then orientfoursome(nochange,ok,ugraph,level);
    until nochange or (not ok);
  end;
999: end;

procedure printcausalsuff(var consis: boolean; var ofilvar: text; var ugraph: endarray);
var i,j: vertextype; head,tail: vertextype; 
    written: vertexset; noconsistent: boolean;
begin
  twriteln(ofilvar,'/Pattern |');
  consis:= true;
  written := [];
  for i := 1 to numob-1 do
   for j := i+1 to numob do
    if (i in graphvertices) and (j in graphvertices) then
     if (ugraph^[i,j] <> []) or (ugraph^[j,i] <> []) then
      begin
          written := written + [i,j];
          if (knownin2 in ugraph^[i,j]) then tail := i
          else if (knownin2 in ugraph^[j,i]) then tail := j
          else if (in2 in ugraph^[i,j]) then tail := i
          else if (in2 in ugraph^[j,i]) then tail := j
          else tail := i;
          if tail = i then head := j 
          else head := i;
	  pwritename(tail,ofilvar,7);
	  twrite(ofilvar,' |');
          noconsistent :=
             ([in2,out2] <= ugraph^[tail,head]) or 
             ([in2,knownout2] <= ugraph^[tail,head]) or             
             ([out2,knownin2] <= ugraph^[tail,head]) or
             ([out2,in2] <= ugraph^[tail,head]) or 
             ([in2,out2] <= ugraph^[head,tail]) or
             ([in2,knownout2] <= ugraph^[head,tail]) or             
             ([out2,knownin2] <= ugraph^[head,tail]) or
             ([out2,in2] <= ugraph^[head,tail]);
          if noconsistent then twrite(ofilvar,'    |')
          else 
           begin
            if (in2 in ugraph^[head,tail]) and
             (not (knownout2 in ugraph^[head,tail])) then twrite(ofilvar,'<|')
            else twrite(ofilvar,'-|');
            if (notknown in ugraph^[tail,head]) then twrite(ofilvar,'- |')
            else twrite(ofilvar,'> |');
           end;
	  writename(head,ofilvar);
          if noconsistent then 
            begin
              twrite(ofilvar,'#|');
              consis:= false;
            end;
	  twriteeoln(ofilvar);
   end;
  for i := 1 to numob do if (i in graphvertices) and (not (i in written)) then
     begin
	  pwritename(i,ofilvar,7);
          twriteeoln(ofilvar);
     end;
end;

procedure disposetri(var atrilist: triangleptr);
var curpos: triangleptr;
begin
  curpos := atrilist;
  if curpos <> nil then
     if curpos^.fwd <> nil then
        begin
          curpos := curpos^.fwd;
          dispose(curpos^.back);
        end
     else
        begin
           dispose(curpos);
           curpos := nil;
        end;
  atrilist := nil;
end;

procedure checkpath(var ugraph: graphrec; var ok: boolean; var ofilvar: text);
var i,j,k: vertextype; 
begin
    ok := true;
    calcdpath(ugraph,ok);
    twriteln(ofilvar,'/Directed Paths |');
    for i := 1 to numob do 
       for j := 1 to numob do
        if (i in graphvertices) and (j in graphvertices) then
         if i <> j then
            if ugraph.ancest^[i,j] <> notthere then
               begin
	         writename(i,ofilvar);
                 twrite(ofilvar,' to |');
                 writename(j,ofilvar);
                 twriteeoln(ofilvar);
               end;
     twriteln(ofilvar,'! |');
     twriteeoln(ofilvar);
     twriteeoln(ofilvar);
end;

procedure checknopath(var ugraph: graphrec; var ofilvar: text);
var i,j,k: vertextype; 
begin
 twriteln(ofilvar,'/Not Connected by Directed Paths |');
 for i := 1 to numob do 
    for j := 1 to numob do
      if (i in graphvertices) and (j in graphvertices) then
         if i <> j then
            if ugraph.ancest^[i,j] = notthere then
                begin
                  writename(i,ofilvar);
                  twrite(ofilvar,' to |');
                  writename(j,ofilvar);
                  twriteeoln(ofilvar);
                end;
     twriteln(ofilvar,'! |');
     twriteeoln(ofilvar);
     twriteeoln(ofilvar);
end;
          
procedure printthepoipg(var ofilvar: text; var ugraph: graphrec; var nocollidelist: triangleptr; score: real);
var i,j: vertextype; head,tail: vertextype; written: vertexset; 
    curpos: triangleptr; ok,noconsistent: boolean; 
begin
  twriteln(ofilvar,'/PAG|');
  written := [];
  twriteeoln(ofilvar);
  for i := 1 to numob-1 do
    for j := i+1 to numob do
     if (i in graphvertices) and (j in graphvertices) then
      if (ugraph.augraph^[i,j] <> []) or (ugraph.augraph^[j,i] <> []) then
        begin
          written := written + [i,j];
          if (knownin2 in ugraph.augraph^[i,j]) then tail := i
          else if (knownin2 in ugraph.augraph^[j,i]) then tail := j
          else if (in2 in ugraph.augraph^[i,j]) then tail := i
          else if (in2 in ugraph.augraph^[j,i]) then tail := j
          else tail := i;
          if tail = i then head := j 
          else head := i;
          noconsistent :=
             ([in2,out2] <= ugraph.augraph^[tail,head]) or 
             ([in2,knownout2] <= ugraph.augraph^[tail,head]) or             
             ([out2,knownin2] <= ugraph.augraph^[tail,head]) or
             ([out2,in2] <= ugraph.augraph^[tail,head]) or 
             ([in2,out2] <= ugraph.augraph^[head,tail]) or
             ([in2,knownout2] <= ugraph.augraph^[head,tail]) or             
             ([out2,knownin2] <= ugraph.augraph^[head,tail]) or
             ([out2,in2] <= ugraph.augraph^[head,tail]);
	  writename(tail,ofilvar);
          if noconsistent then twrite(ofilvar,'   |')
          else
              begin
                if ((in2 in ugraph.augraph^[head,tail]) or 
                    (knownin2 in ugraph.augraph^[head,tail])) then
                       twrite(ofilvar,' <|')
                else if ((out2 in ugraph.augraph^[head,tail]) or
                         (knownout2 in ugraph.augraph^[head,tail])) then
                            twrite(ofilvar,' -|')
                else twrite(ofilvar,' o|');
                twrite(ofilvar,'-|');
                if ((in2 in ugraph.augraph^[tail,head]) or
                    (knownin2 in ugraph.augraph^[tail,head])) then
                       twrite(ofilvar,'> |')
                else if ((out2 in ugraph.augraph^[tail,head]) or
                         (knownout2 in ugraph.augraph^[tail,head])) then
                            twrite(ofilvar,'- |')
                else twrite(ofilvar,'o |');
              end;
	  writename(head,ofilvar);
          if noconsistent then twrite(ofilvar,'#|');
          twriteeoln(ofilvar);
     end;
  for i := 1 to numob do if (i in graphvertices) and (not (i in written)) then
     begin
	  pwritename(i,ofilvar,7);
          twriteeoln(ofilvar);
     end;
  tpwritereal(ofilvar,score,12,3,false);
  twriteln(ofilvar,'!|');
  twriteeoln(ofilvar);
  curpos := nocollidelist;
  if curpos <> nil then
      twriteln(ofilvar,'The following pairs of edges do not collide: |');
  while curpos <> nil do
     begin
        writename(curpos^.inout,ofilvar);
        twrite(ofilvar,' -- |');
        writename(curpos^.unori,ofilvar);
        twrite(ofilvar,'    |');
        writename(curpos^.unori,ofilvar);
        twrite(ofilvar,' -- |');
        writename(curpos^.other,ofilvar);
	twriteeoln(ofilvar);        
        curpos := curpos^.fwd;
     end;
   twriteeoln(ofilvar);   
   ok := true;     
   checkpath(ugraph,ok,ofilvar);
   checknopath(ugraph,ofilvar);
   disposetri(nocollidelist);
   disposetri(deftrilist);
   disposetri(posstrilist);
end;

procedure printlist(var ugraph: endarray);
var i,j: integer;
begin
for i := 1 to numob do
  for j := 1 to numob do
    if (i in graphvertices) and (j in graphvertices) then
     if ((in2 in ugraph^[i,j]) or (knownin2 in ugraph^[i,j])) and
        (not (out2 in ugraph^[i,j])) and (not (knownout2 in ugraph^[i,j])) and
        ((out2 in ugraph^[j,i]) or (knownout2 in ugraph^[j,i])) and
        (not (in2 in ugraph^[j,i])) and (not (knownin2 in ugraph^[j,i])) then
	begin
	    writename(i,ofilvar);
	    twrite(ofilvar,' -> |');
	    writename(j,ofilvar);
	end;
end;

procedure getzeroes(var zero1,zero2,zero3,zero4: vertexset;
                    i,j,k,l,pattern: integer);
begin
   if pattern = 1 then
      begin
        zero1 := zeroes^[i,j];
        zero2 := zeroes^[k,l];
        zero3 := zeroes^[i,k];
        zero4 := zeroes^[j,l];
      end
   else if pattern = 2 then
      begin
        zero1 := zeroes^[i,l];
        zero2 := zeroes^[j,k];
        zero3 := zeroes^[i,j];
        zero4 := zeroes^[k,l];
      end
   else
      begin
        zero1 := zeroes^[i,k];
        zero2 := zeroes^[j,l];
        zero3 := zeroes^[i,l];
        zero4 := zeroes^[j,k];
      end  
end;

function parthold(partial: partialpair; var partialedon: partialedonarray): integer;
var aval,aprob: real; held,ok: boolean;
begin
      aval := calcpart1(partial,partialedon,ok);
      if not ok then parthold := 1
      else
        begin
           calcpartialheld(aval,aprob,held,partialedon.length);
           if held then parthold := 0 else parthold := 1;
        end;
end;

procedure zerocorr(i1,j1,k1,l1,i2,j2,k2,l2: vertextype; 
                   var foundcommon: boolean);
var loc1,loc2,loc3,loc4: partialpair; nullset: partialedonarray;
begin
    loc1[1] := i1; loc1[2] := j1;
    loc2[1] := k1; loc2[2] := l1;
    loc3[1] := i2; loc3[2] := j2;
    loc4[1] := k2; loc4[2] := l2;
    nullset.length := 0;
    if (parthold(loc1,nullset) = 0) or (parthold(loc2,nullset) = 0) and
        (parthold(loc3,nullset) = 0) or (parthold(loc4,nullset) = 0) then
      foundcommon := true;
end;

procedure checkzero(i1,j1,k1,l1,i2,j2,k2,l2,k,depth: integer; 
                   var intersectarray,subarray: partialedonarray;
                   var foundcommon: boolean; var subset: vertexset; 
                   pattern: vertextype);
var next,i1j1,k1l1,i2j2,k2l2: integer; goahead: boolean;
    loc1,loc2,loc3,loc4: partialpair; nullset: partialedonarray;
begin
    goahead := true;
    subset := subset + [intersectarray.arr[k]];
    subarray.arr[depth] := intersectarray.arr[k];
    if ((pattern = 1) or (pattern = 2)) and
       (((i1 in subset) and (l1 in subset)) or 
        ((j1 in subset) and (k1 in subset))) then goahead := false
    else if
       (((i1 in subset) and (k1 in subset)) or 
        ((j1 in subset) and (l1 in subset))) then goahead := false;
    if goahead then
     if (depth = subarray.length) then
       begin
          loc1[1] := i1; loc1[2] := j1;
          loc2[1] := k1; loc2[2] := l1;
          loc3[1] := i2; loc3[2] := j2;
          loc4[1] := k2; loc4[2] := l2;
          if (i1 in subset) or (j1 in subset) then i1j1 := 0
          else i1j1 := parthold(loc1,subarray);
          if (k1 in subset) or (l1 in subset) then k1l1 := 0
          else k1l1 := parthold(loc2,subarray);
          if (i2 in subset) or (j2 in subset) then i2j2 := 0
          else i2j2 := parthold(loc3,subarray);
          if (k2 in subset) or (l2 in subset) then k2l2 := 0
          else k2l2 := parthold(loc4,subarray);
          if (i1j1 = 0) and (k1l1 = 0) and (i2j2 = 0) and (k2l2 = 0) then
             foundcommon := true;
       end
     else if (depth < subarray.length) and (not foundcommon) then
         for next := k+1 to intersectarray.length-subarray.length+depth+1 do
                checkzero(i1,j1,k1,l1,i2,j2,k2,l2,next,depth+1,intersectarray,
                          subarray,foundcommon,subset,pattern);
    subset := subset - [intersectarray.arr[k]];
end;

procedure writelatent(i,j,k,l,pattern: integer);
begin
    if nonewritten then
       begin
          twriteeoln(ofilvar);
          twriteeoln(ofilvar);
          twriteln(ofilvar,'Tests for latent variables.|');
          nonewritten := false;
       end;
    if pattern = 1 then 
	begin
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' - |');
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	twrite(ofilvar,' needs a latent variable|');
	twriteeoln(ofilvar);
	end
    else if pattern = 2 then
	begin
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' - |');
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	twrite(ofilvar,' needs a latent variable|');
	twriteeoln(ofilvar);
	end
    else
	begin
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' - |');
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	twrite(ofilvar,' needs a latent variable|');
	twriteeoln(ofilvar);
	end;
end;

procedure needlatent(i,j,k,l,pattern: vertextype);
label 999;
var zero1,zero2,zero3,zero4,intersect,subset: vertexset; n,length: integer;
    foundcommon: boolean; intersectarray,subarray: partialedonarray;
begin
     foundcommon := false;
     if pattern = 1 then zerocorr(i,j,k,l,i,k,j,l,foundcommon)
     else if pattern = 2 then zerocorr(i,l,j,k,i,j,k,l,foundcommon)
     else zerocorr(i,k,j,l,i,l,j,k,foundcommon);
     if foundcommon then goto 999;
     getzeroes(zero1,zero2,zero3,zero4,i,j,k,l,pattern);
     intersect := zero1 * zero2 * zero3 * zero4;
     intersectarray.length := 0;
     for n := 0 to numob do 
      if (n = 0) or (n in graphvertices) then
        if n in intersect then 
            begin
               intersectarray.length := intersectarray.length + 1;
               intersectarray.arr[intersectarray.length] := n;
            end;
        length := 1;
        subset := [];
        while (length <= intersectarray.length) and (not foundcommon) do
           begin
              subarray.length := length;
              for n := 1 to intersectarray.length-length+1 do
                if pattern = 1 then
                   checkzero(i,j,k,l,i,k,j,l,n,
                             1,intersectarray,subarray,foundcommon,subset,pattern)
                else if pattern = 2 then
                   checkzero(i,l,j,k,i,j,k,l,n,
                             1,intersectarray,subarray,foundcommon,subset,pattern)
                else
                   checkzero(i,k,j,l,i,l,j,k,n,
                             1,intersectarray,subarray,foundcommon,subset,pattern);
                length := length + 1;
           end;
       if (not foundcommon) and (i in graphvertices) and 
         (j in graphvertices) and (k in graphvertices) and
         (l in graphvertices) then writelatent(i,j,k,l,pattern);
999: end;

procedure checklatents;
label 999;
var i,j,k,l: vertextype; p,q,r: real; locresid,aprob: array[1..3] of real;
begin
   if aaabort then goto 999;
   for i := 1 to numob-3 do
      for j := i+1 to numob-2 do
          for k := j+1 to numob-1 do
             for l := k+1 to numob do
                if (i in graphvertices) and (j in graphvertices) and 
                   (k in graphvertices) and (l in graphvertices) then
                  begin
                     if aaabort then goto 999;
                     p := ((rho^[i,j] * rho^[k,l]));
                     q := ((rho^[i,k] * rho^[j,l]));
                     r := ((rho^[i,l] * rho^[j,k]));
                     aprob[1]:= 0; aprob[2]:= 0; aprob[3]:= 0;
                     locresid[1] := (p - q);
                     calcstat(i,l,j,k,abs(locresid[1]),aprob[1],1);
                     locresid[3] := (q - r);
                     calcstat(i,j,k,l,abs(locresid[3]),aprob[3],3);
                     locresid[2] := (r - p);
                     calcstat(i,k,j,l,abs(locresid[2]),aprob[2],2);
                     if aprob[1] > signif then needlatent(i,j,k,l,1);
                     if aprob[2] > signif then needlatent(i,j,k,l,2);
                     if aprob[3] > signif then needlatent(i,j,k,l,3);
                  end;
999: end;

procedure deleteedge(var edge,listedge: nodeptr);
begin
 if edge <> nil then
   begin
      if edge^.back <> nil then edge^.back^.fwd := edge^.fwd
      else listedge := edge^.fwd;
      if edge^.fwd <> nil then edge^.fwd^.back := edge^.back;
      dispose(edge);
   end;
end;

procedure removeedge(vert: integer; var listedge: nodeptr);
var curpos: nodeptr; notfound: boolean;
begin
   curpos := listedge;
   notfound := true;
   while (curpos <> nil) and notfound do
      if curpos^.edge = vert then
         begin
            notfound := false;
            deleteedge(curpos,listedge);
         end
      else curpos := curpos^.fwd;
end;
   

procedure leafinsert(var curnode,leaflist: nodeptr; smallest: boolean);
var curpos: nodeptr; leaveloop,done: boolean;
begin
   curpos := leaflist;
   leaveloop := false;
   done := false;
   if leaflist = nil then 
       begin
         leaflist := curnode;
         leaflist^.fwd := nil;
         leaflist^.back := nil;
       end
   else
    begin
     while (curpos <> nil) and (not leaveloop) and (not done) do
        begin
           if ((curpos^.prob < curnode^.prob) and smallest) or
              ((curpos^.prob > curnode^.prob) and (not smallest)) then
              if curpos^.fwd = nil then
                  begin
                     curpos^.fwd := curnode;
                     curnode^.back := curpos;
                     curnode^.fwd := nil;
                     done := true;
                  end
              else curpos := curpos^.fwd
          else leaveloop := true;
        end;
     if not done then
       begin
         if curpos^.back <> nil then
            begin
              curpos^.back^.fwd := curnode;
              curnode^.back := curpos^.back;
              curpos^.back := curnode;
              curnode^.fwd := curpos;
            end
         else 
            begin
              curnode^.fwd := leaflist;
              curnode^.back := nil;
              leaflist^.back := curnode;
              leaflist := curnode;
            end;
       end;
   end;
end;

procedure moveedge(var edge,listedge1,listedge2: nodeptr);
begin
 if edge <> nil then
   begin
      if edge^.back <> nil then edge^.back^.fwd := edge^.fwd
      else listedge1 := edge^.fwd;
      if edge^.fwd <> nil then edge^.fwd^.back := edge^.back;
      edge^.fwd := nil;
      edge^.back := nil;
      leafinsert(edge,listedge2,true);
   end;
end;

procedure writprobmessage;
  begin
    twriteln(ofilvar,'{------------------------------------------------------|');
    if (not discrete) and (not indepdata) then 
        twriteln(ofilvar,'List of vanishing (partial) correlations that made|')
    else twriteln(ofilvar,'List of (conditional) independence relations that made|');
    twriteln(ofilvar,'TETRAD remove adjacencies. |');
    twriteeoln(ofilvar);
    if (not discrete) and (not indepdata) then 
      begin
       twriteln(ofilvar,'  Corr. :  Sample (Partial) Correlation|');
       twriteln(ofilvar,'  Prob. :  Probability that the absolute value of the sample |');
       twriteln(ofilvar,'           (partial) correlation exceeds the observed value, |');
       twriteln(ofilvar,'           on the assumption of zero (partial) correlation in|');
       twriteln(ofilvar,'           the population, assuming a multinormal distribution.|');
      end
    else if (not indepdata) then
      begin
       twriteln(ofilvar,'  Prob. :  Probability that the absolute value of the sample |');
       twriteln(ofilvar,'           G-square statistic exceeds the observed value, |');
       twriteln(ofilvar,'           on the assumption of (conditional) independence in|');
       twriteln(ofilvar,'           the population, assuming a multinomial distribution.|');
      end;
    twriteeoln(ofilvar);
    if (not discrete) and (not indepdata) then
      begin
        twriteln(ofilvar,'Edge             (Partial)|');  
        twriteln(ofilvar,'Removed          Correlation                 Corr.     Prob.|');
        twriteln(ofilvar,'-------          -----------                 -----     -----|');
      end
    else if (not indepdata) then
      begin
        twriteln(ofilvar,'Edge             (Conditional)|');  
        twriteln(ofilvar,'Removed          Independence                G-square  Prob.|');
        twriteln(ofilvar,'-------          ------------                --------  -----|');
      end
    else
      begin
        twriteln(ofilvar,'Edge             (Conditional)|');  
        twriteln(ofilvar,'Removed          Independence|');
        twriteln(ofilvar,'-------          ------------    |');
      end;
  end;

procedure printpartial(part: partialpair; var partialedon: partialedonarray;
                    var ofilvar: text; aprob,aval: real);
var i,totalchars: integer; 
begin
   if nonewritten then writprobmessage;
   nonewritten := false;
   totalchars:=0;
   rswritename(part[1],ofilvar,totalchars);
   twrite(ofilvar,' -- |'); totalchars:= totalchars+4;
   rswritename(part[2],ofilvar,totalchars);
   for i:= 1 to (17 - totalchars) do twrite(ofilvar,' |');   
   totalchars:= 17;
   if (not indepdata) and (not discrete) then
      begin
        twrite(ofilvar,'rho(|'); totalchars:= totalchars+4;
      end;
   rswritename(part[1],ofilvar,totalchars);
   twrite(ofilvar,' |'); totalchars:= totalchars+1;
   rswritename(part[2],ofilvar,totalchars); 
   if (partialedon.length > 0) then 
     begin
       twrite(ofilvar,' . |'); 
       totalchars:= totalchars+3;
       for i := 1 to partialedon.length do
         begin
	   rswritename(partialedon.arr[i],ofilvar,totalchars);
           if i < partialedon.length then 
             begin
               twrite(ofilvar,' |');
               totalchars:= totalchars+1;
             end;
         end;
     end;
   if (not indepdata) and (not discrete) then
        begin
          twrite(ofilvar,') |'); totalchars:= totalchars+2;
        end;
   if totalchars > 45 then begin twriteeoln(ofilvar); totalchars:= 0; end;
   for i:= 1 to (45 - totalchars) do twrite(ofilvar,' |');
   if (not indepdata) then
      begin
        twritereal(ofilvar,aval); 
        twrite(ofilvar,'  |');
        twritereal(ofilvar,aprob);
        twriteeoln(ofilvar);
      end
   else twriteeoln(ofilvar);
end;

function checkindeplist(partial: partialpair; 
                         var partialedon: partialedonarray): boolean;
var curpos: indepptr; i,j: vertextype;
begin
   checkindeplist := false;
   curpos := indep1;
   while curpos <> nil do
     begin
       if ((curpos^.tail = partial[1]) and (curpos^.head = partial[2])) or
          ((curpos^.head = partial[1]) and (curpos^.tail = partial[2])) then
             begin
                checkindeplist := true;
                i := 1;
                for j := 1 to numob do 
                 if j in graphvertices then
                   if j in curpos^.cond then
                       begin
			   zeroes^[partial[1],partial[2]] := 
     			      zeroes^[partial[1],partial[2]] + [j];
                           dsep^[partial[1],partial[2]] := 
                              dsep^[partial[1],partial[2]] + [j];
                           partialedon.arr[i] := j;
                           i := i + 1;
                       end;
                 dsep^[partial[2],partial[1]] := dsep^[partial[1],partial[2]];
                 partialedon.length := i - 1;
                 curpos := nil;
             end
        else curpos := curpos^.fwd;
     end;
end;

       
procedure checkpartials(edge: nodeptr; order,depth: integer; start: nodeptr;
                        var partialedon: partialedonarray; 
                        var ugraph: endarray; internal: boolean; var corrlists: corrarray);
label 999;
var i,j: integer; aval,aprob,g2: real; held,ok: boolean; partial: partialpair;
    curpos: nodeptr;
begin
  if aaabort then goto 999;
  partialedon.length := order;
  partial[1] := edgetovert^[edge^.edge].head; 
  partial[2] := edgetovert^[edge^.edge].tail;
  curpos := start;
  while curpos <> nil do 
   begin
    if (curpos^.edge <> partial[1]) and (curpos^.edge <> partial[2]) then
     with partialedon do
      begin
        arr[depth+1] := curpos^.edge;
        if depth <> order - 1 then
           checkpartials(edge,order,depth+1,curpos^.fwd,partialedon,ugraph,internal,corrlists)
        else
          begin
           if indepdata then held := checkindeplist(partial,partialedon)
           else if (not discrete) then
             begin
               aval := calcpart1(partial,partialedon,ok);
               if ok then 
                   calcpartialheld(aval,aprob,held,partialedon.length)
               else held := false;
               if toomuchtime then goto 999;
             end
           else held := calcindep1(partial,aprob,g2,partialedon,numchecked);
           if held then
              if (ugraph^[partial[1],partial[2]] <> []) and
                 (adjacency^[partial[1],partial[2]] = notthere) and
                 (adjacency^[partial[2],partial[1]] = notthere) then
                 begin
                   numedges := numedges - 1;
(*                   if (not internal) then  *)
                     if (not discrete) then
                       printpartial(partial,partialedon,ofilvar,aprob,aval)
                     else printpartial(partial,partialedon,ofilvar,aprob,g2);
                   ugraph^[partial[1],partial[2]] := [];
                   ugraph^[partial[2],partial[1]] := [];
                   parents[partial[1]] := parents[partial[1]] - [partial[2]];
                   parents[partial[2]] := parents[partial[2]] - [partial[1]];
                   removeedge(partial[1],corrlists[partial[2]]);
                   removeedge(partial[2],corrlists[partial[1]]);
(*                   if not causalsuff then  *)
                    begin
                      readarraytoset(dsep^[partial[1],partial[2]],partialedon);
                      dsep^[partial[2],partial[1]] := 
                         dsep^[partial[1],partial[2]];
                    end;
                   for j := 1 to length do 
                     zeroes^[partial[1],partial[2]] := 
                       zeroes^[partial[1],partial[2]] + [arr[j]];
                 end;
           end;
        end; 
      curpos := curpos^.fwd;
     end;
999: end;

(*
procedure initugraph(var adjacency: adjtype; var ugraph: graphrec);
var i,j: integer; 
begin
   for i := 1 to numob do
    begin
     ugraph.ancestset[i] := [];
     ugraph.descend[i] := [];
     for j := 1 to numob do
      begin
        ugraph.ancest^[i,j] := notthere;
        if (i in graphvertices) and (j in graphvertices) then
          begin
             if adjacency^[i,j] <> notthere then
               ugraph.augraph^[i,j] := [knownin2]
             else if adjacency^[j,i] <> notthere then
               ugraph.augraph^[i,j] := [knownout2]
             else ugraph.augraph^[i,j] := [notknown];
             dsep^[i,j] := [0];
          end;
      end;
     end;
end;
*)

procedure initugraph(var adjacency: adjtype; var ugraph: graphrec);
var i,j: integer; 
begin
   for i := 1 to numob do
    begin
     ugraph.ancestset[i] := [];
     ugraph.descend[i] := [];
     for j := 1 to numob do
      begin
        ugraph.ancest^[i,j] := notthere;
        if (i in graphvertices) and (j in graphvertices) then
          begin
             if adjacency^[i,j] <> notthere then
               ugraph.augraph^[i,j] := [in2]
             else if adjacency^[j,i] <> notthere then
               ugraph.augraph^[i,j] := [out2]
             else ugraph.augraph^[i,j] := [notknown];
             dsep^[i,j] := [0];
          end;
      end;
     end;
end;

procedure orientfromknow(var ugraph: endarray);
var i,j: integer;
begin
  for i := 1 to numob -1 do
     for j := i+1 to numob do
       if (i in graphvertices) and (j in graphvertices) then
          if causalsuff then
              begin
                 if adjacency^[i,j] <> notthere then 
                   begin
                      ugraph^[i,j] := [knownin2];
                      ugraph^[j,i] := [knownout2];
                   end
                 else if adjacency^[j,i] <> notthere then 
                   begin
                      ugraph^[j,i] := [knownin2];
                      ugraph^[i,j] := [knownout2];
                   end
                 else if (forbiddenadj^[i,j] <> notthere) and
                    (forbiddenadj^[j,i] <> notthere) then 
                    begin
                      ugraph^[i,j] := [];
                      ugraph^[j,i] := [];
                    end
                 else if (forbiddenadj^[i,j] <> notthere) then
                    begin
                      ugraph^[j,i] := [knownin2];
                      ugraph^[i,j] := [knownout2];
                    end
                 else if (forbiddenadj^[j,i] <> notthere) then
                    begin
                      ugraph^[i,j] := [knownin2];
                      ugraph^[j,i] := [knownout2];
                    end
                 else
                    begin
                      ugraph^[i,j] := [notknown];
                      ugraph^[j,i] := [notknown];
                    end
              end
           else
             begin
                 if adjacency^[i,j] <> notthere then 
                      ugraph^[i,j] := [knownin2]
                 else if adjacency^[j,i] <> notthere then 
                      ugraph^[j,i] := [knownin2]
                 else if (forbiddenadj^[i,j] <> notthere) and
                    (forbiddenadj^[j,i] <> notthere) and
                    (commonfor^[i,j] <> notthere) then 
                    begin
                      ugraph^[i,j] := [];
                      ugraph^[j,i] := [];
                    end
                 else if (forbiddenadj^[i,j] <> notthere) and
                    (forbiddenadj^[j,i] <> notthere) then
                      begin
                         ugraph^[i,j] := [knownin2];
                         ugraph^[j,i] := [knownin2];
                      end
                 else if (forbiddenadj^[i,j] <> notthere) and
                    (commonfor^[i,j] <> notthere) then
                       begin
                         ugraph^[j,i] := [knownin2];
                         ugraph^[i,j] := [knownout2];
                       end
                 else if (forbiddenadj^[j,i] <> notthere) and
                    (commonfor^[i,j] <> notthere) then
                       begin
                         ugraph^[i,j] := [knownin2];
                         ugraph^[j,i] := [knownout2];
                       end
                 else if (forbiddenadj^[i,j] <> notthere) then
                      ugraph^[j,i] := [knownin2]
                 else if (forbiddenadj^[j,i] <> notthere) then
                      ugraph^[i,j] := [knownin2]
                 else
                    begin
                      ugraph^[i,j] := [notknown];
                      ugraph^[j,i] := [notknown];
                    end
          end;

end;

procedure disposeedgelist(var edgelist: nodeptr);
var curpos: nodeptr;
begin
  curpos := edgelist;
  while edgelist <> nil do
    begin
       edgelist := edgelist^.fwd;
       dispose(curpos);
       curpos := edgelist;
    end;
  edgelist := nil;
end;

procedure checkedges(var edgelist,newlist,newlist1: nodeptr; internal: boolean; var ugraph: endarray;
                     var corrlists: corrarray);
label 999;
var partial: partialpair; temp,curnode: nodeptr; 
    theseparents1,theseparents2: vertexset;
    partialedon: partialedonarray; i: integer; leaveloop: boolean;
begin;
   newlist1 := nil;
   if aaabort then goto 999;
   leaveloop := false;
   i := 0;
   repeat
      curnode := edgelist;
      i := i + 1;
      if (depth <> -1) and (i > depth) then leaveloop := true;
      while (curnode <> nil) do 
        begin
          if toomuchtime then goto 999;
          partial[1] := edgetovert^[curnode^.edge].head;
          partial[2] := edgetovert^[curnode^.edge].tail;

(* if 2 is later than 1, then 2 is not ancestor of 1, so check only 2 *)
          if (not (partial[2] in temporder[partial[1]])) then 
             begin
               theseparents1 := parents[partial[1]] - [partial[1],partial[2]];
               if (setsize(theseparents1) >= i) then
                 checkpartials(curnode,i,0,corrlists[partial[1]],partialedon,ugraph,internal,corrlists);
             end;

(* if 1 is later than 2, then 1 is not ancestor of 2, so check only 1 *)
          if (not (partial[1] in temporder[partial[2]])) then
             begin
               theseparents2 := parents[partial[2]] - [partial[1],partial[2]];
               if (setsize(theseparents2) >= i) then
                 checkpartials(curnode,i,0,corrlists[partial[2]],partialedon,ugraph,internal,corrlists);
             end;
          if (ugraph^[partial[1],partial[2]] = []) then
             begin
                temp := curnode;
                curnode := curnode^.fwd;
                moveedge(temp,edgelist,newlist1);
             end
          else if ((setsize(theseparents1) <= i) and (setsize(theseparents2) <= i)) then
             begin
                temp := curnode;
                curnode := curnode^.fwd;
                moveedge(temp,edgelist,newlist);
             end
          else curnode := curnode^.fwd;
       end;
    until (edgelist = nil) or leaveloop;
999: end;

function collide(tail,head,i: vertextype; var ugraph: endarray): boolean;
begin
   collide := 
      ((in2 in ugraph^[tail,head]) or (knownin2 in ugraph^[tail,head])) and
      ((in2 in ugraph^[i,head]) or (knownin2 in ugraph^[i,head]));
end;

procedure findcond(v,tail,head: vertextype; 
                   var condset,pathset1,pathset2: vertexset; var ugraph: endarray);
label 999;
var i: vertextype; intoi: boolean;
begin
 if (graphvertices <= (condset + [v])) then goto 999;
 for i := 1 to numob do
   if (i in graphvertices) then
     begin
      if (graphvertices <= (condset + [v])) then goto 999;
      if (ugraph^[head,i] <> []) then
        begin
         intoi := (in2 in ugraph^[head,i]) or (knownin2 in ugraph^[head,i]);
         if collide(tail,head,i,ugraph) or triangle(i,head,tail,ugraph) or (tail = 0) then
            if (intoi and (not (i in pathset1))) or
               ((not intoi) and (not (i in (pathset1 + pathset2)))) then
           begin
             condset := condset + [i];
             if intoi then pathset1 := pathset1 + [i]
             else pathset2 := pathset2 + [i];
             findcond(v,head,i,condset,pathset1,pathset2,ugraph);
             if intoi then pathset1 := pathset1 - [i]
             else pathset2 := pathset2 - [i];
           end;
        end;
     end;
999: end;

procedure unorient(var ugraph: endarray);
var i,j: vertextype;
begin
  for i := 1 to numob do
    for j := 1 to numob do
     if (i in graphvertices) and (j in graphvertices) then
        if ugraph^[i,j] <> [] then 
           ugraph^[i,j] := ugraph^[i,j] - [in2,out2] + [notknown];
end;

procedure insertnode(var corrlists : corrarray; num: real; index1,index2: vertextype);
var curnode: nodeptr;
begin
   new(curnode);
   curnode^.prob := num;
   curnode^.edge := index2;
   leafinsert(curnode,corrlists[index1],false);   
end;

procedure initparents(var edgelist: nodeptr; internal: boolean; var ugraph: endarray);
var i,j: vertextype; aprob,g2: real; held: boolean; 
    partialedon: partialedonarray; curnode1,curnode2,curnode3: 
    nodeptr; partial: partialpair;
begin
   partialedon.length := 0;
   for i := 1 to numob - 1 do
      for j := i + 1 to numob do
        if (i in graphvertices) and (j in graphvertices) then
         begin
            zeroes^[i,j] := [];
            zeroes^[j,i] := [];
            partial[1] := i; partial[2] := j; 
            if (not discrete) and (not indepdata) then 
                 calcpartialheld(rho^[i,j],aprob,held,0)
            else if (not indepdata) then
              held := calcindep1(partial,aprob,g2,partialedon,numchecked)
            else held := checkindeplist(partial,partialedon);

            (* if it holds and is not required remove edge *)
            if (held and
                (adjacency^[i,j] = notthere) and 
                (adjacency^[j,i] = notthere)) then
               begin
                  ugraph^[i,j] := [];
                  ugraph^[j,i] := [];
                  if (not indepdata) then dsep^[i,j] := [];
                  if (not indepdata) then dsep^[j,i] := [];
                  numedges := numedges - 1;
(*                  if (not internal) then  *)
                   if not discrete then
                     printpartial(partial,partialedon,ofilvar,aprob,rho^[i,j])
                  else printpartial(partial,partialedon,ofilvar,aprob,g2);
                end;

            (* if not held, not required either way, and not forbidden
               all ways, add to list of edges to check *)
            if (not held) and                             
               ((adjacency^[i,j] = notthere) and 
                (adjacency^[j,i] = notthere)) and
               (ugraph^[i,j] <> []) then
               begin
                   new(curnode1);
                   if discrete then curnode1^.prob := g2
                   else curnode1^.prob := rho^[i,j];
                   curnode1^.edge := edgenames^[i,j];
                   leafinsert(curnode1,edgelist,true);
               end;

            if (not held) and (not (knownout2 in ugraph^[j,i])) and
               (not (ugraph^[i,j] = [])) then
                  begin
                      parents[i] := parents[i] + [j];
                      if discrete then insertnode(corrlists,g2,i,j) else insertnode(corrlists,rho^[i,j],i,j);
                      if discrete then insertnode(corrlists1,g2,i,j) else insertnode(corrlists1,rho^[i,j],i,j);
                  end;
            if (not held) and (not (knownout2 in ugraph^[i,j])) and
               (not (ugraph^[j,i] = [])) then
                   begin
                      parents[j] := parents[j] + [i];
                      if discrete then insertnode(corrlists,g2,j,i) else insertnode(corrlists,rho^[i,j],j,i);
                      if discrete then insertnode(corrlists1,g2,j,i) else insertnode(corrlists1,rho^[i,j],j,i);
                   end;
           end;
end;

(*******************************************************************************************)

procedure getthenextedge(var i,j: integer; var found: boolean; var ugraph: graphrec);
var start1,start2,tempi,tempj: integer;
begin
 with ugraph do
  begin
   start1 := i;
   start2 := j;
   found := false;
   if j = max then 
      begin
        start1 := start1 + 1;
        start2 := 1;
      end
   else start2 := start2 + 1;
   tempi := start1;
   tempj := start2;
   repeat
      while (tempj <= max) and (not found) do
        begin
	 if (tempi <> tempj) and (notknown in augraph^[tempj,tempi]) then
	    begin
	       i := tempi;
               j := tempj;
               found := true;
            end
         else tempj := tempj + 1;
       end;
      tempi := tempi + 1;
      tempj := 1;
   until (tempi > max) or found;
 end;
end;

function okmag(var mag: graphrec): boolean;
label 999; 
var i,j: vertextype; tempok: boolean;
begin
  with mag do
   begin
    tempok := true;
    for i := 1 to max do
      for j := 1 to max do
       if (i <> j) then
        begin
          if (in2 in augraph^[i,j]) and (ancest^[j,i] <> notthere) then tempok := false;
          if ([in2,out2] <= augraph^[i,j]) or ([in2,out2] <= augraph^[j,i]) then    
             tempok := false;
          if (not tempok) then goto 999;
        end;
   end;
999: okmag := tempok;
end;

procedure undochanges(var pag: graphrec; start,finish: integer; startlist1,startlist2: triangleptr; level: integer);
var i,j: integer; first,second: vertextype; curpos1,curpos2: triangleptr;
begin
 with pag do
  begin 
    for i := start to finish-1 do
     if ((i mod 2) <> 0) then
      begin
        first := recchange^[i];
        second := recchange^[i+1];
        augraph^[first,second] := [notknown];
      end;
    curchange := start;
    curpos1 := pag.nocollidelist;
    while (curpos1 <> nil) and (curpos1 <> startlist1) do
       begin
          curpos2 := curpos1^.fwd;
          dispose(curpos1);
          curpos1 := curpos2;
       end;
    pag.nocollidelist := startlist1;
    if pag.nocollidelist <> nil then pag.nocollidelist^.back := nil;         
    curpos1 := pag.bothlist;
    if startlist2 <> nil then
        begin
            while (curpos1 <> nil) and (curpos1 <> startlist2) do
                  curpos1 := curpos1^.fwd;
            if curpos1 <> nil then
                begin
                  curpos2 := curpos1;
                  curpos1 := curpos1^.fwd;
                  curpos2^.fwd := nil;
                end;
        end;
    while (curpos1 <> nil) do
       begin
          curpos2 := curpos1^.fwd;
          dispose(curpos1);
          curpos1 := curpos2;
       end;
    for i := 1 to max do 
      for j := 1 to max do
        if (i <> j) then
           if ancest^[i,j] = level then 
             begin
               ancest^[i,j] := notthere;
               ancestset[j] := ancestset[j] - [i];
               descend[i] := descend[i] - [j];
             end;
  end;
end;

procedure copylist(var list1,list2: triangleptr);
var curpos1: triangleptr;
begin
  curpos1 := list1;
  while (curpos1 <> nil) do
     begin
        putinlist(curpos1^.inout,curpos1^.unori,curpos1^.other,list2);
        curpos1 := curpos1^.fwd;
     end;
end;

procedure orientanedge(i,j: vertextype; var pag: graphrec; both: boolean;
                    var ok: boolean; var start,finish: integer;
                    var startlist1,startlist2: triangleptr; level: integer);
begin
 with pag do
  begin
   ok := true;
   start := curchange;
   startlist1 := pag.nocollidelist;
   startlist2 := pag.bothlist;
   if startlist2 <> nil then
      while (startlist2^.fwd <> nil) do startlist2 := startlist2^.fwd;
   if (both and 
        ([in2,knownin2] * pag.augraph^[j,i] = []) and 
        ([out2,knownout2] * pag.augraph^[i,j] = []) and
        (forbiddenadj^[i,j] = notthere)) or
     ((not both) and ([out2,knownout2] * pag.augraph^[i,j] = [])) then
     begin
       propagate(i,j,both,ok,pag,level);
       orientedges(pag,false,level,true);
     end
   else ok := false;
   finish := curchange;
  end;
end;

procedure orientnon(var curpos: triangleptr; var pag: graphrec; 
                    var ok: boolean; var start,finish: integer;
                    var startlist1,startlist2: triangleptr; level: integer);
begin
 with pag do
  begin
   ok := true;
   start := curchange;
   startlist1 := pag.nocollidelist;
   startlist2 := pag.bothlist;
   if startlist2 <> nil then
      while (startlist2^.fwd <> nil) do startlist2 := startlist2^.fwd;
   if (not (in2 in pag.augraph^[curpos^.inout,curpos^.unori])) and 
      (not (in2 in pag.augraph^[curpos^.other,curpos^.unori])) then
     begin
       putinlist(curpos^.inout,curpos^.unori,curpos^.other,nocollidelist);
       orientedges(pag,false,level,true);
     end;
   finish := curchange;
  end;
end;

procedure maxcardsearch(var ugraph: graphrec; var alpha,alphainv: alphtype);
var i,j,v,w: vertextype; deleted,finished: boolean;
    size: alphtype;
    aset: settype;
begin
   for i := 0 to max do 
      begin
        aset[i] := [];
        size[i] := 0;
      end;
   aset[0] := [1..max];
   i := 1; j := 0;
   while i <= max do
     if (not (i in graphvertices)) then i := i + 1
     else
       begin
        v := 1;
        deleted := false;
        while (v <= max) and (not deleted) do
         if (not (v in graphvertices)) then v := v + 1
         else
          begin
            if (v in aset[j]) then
               begin
                 aset[j] := aset[j] - [v];
                 deleted := true;
                end
            else v := v + 1;
          end;
        alpha[v] := i; alphainv[i] := v;
        size[v] := -1;
        for w := 1 to max do
          if (w in graphvertices) then
           if (v <> w) and (notknown in ugraph.augraph^[v,w]) and
              (notknown in ugraph.augraph^[w,v]) and (size[w] >= 0) then
                begin
                  aset[size[w]] := aset[size[w]] - [w];
                  size[w] := size[w] + 1;
                  aset[size[w]] := aset[size[w]] + [w];
                end;
       i := i + 1;
       j := j + 1;
       finished := false;
       while (j >= 0) and (aset[j] = []) and (not finished) do if (j > 0) then j := j - 1 else finished := true;
      end;
end;

procedure fillingraph(var ugraph: graphrec);
var alpha,alphainv: alphtype;
    i,v,w,x: vertextype;
    f: array[1..maxvertex] of vertextype;
    index: array[1..maxvertex] of integer;
begin
   maxcardsearch(ugraph,alpha,alphainv);
   for i := max downto 1 do
    if (i in graphvertices) then
     begin
       w := alphainv[i];
       f[w] := w;
       index[w] := i;
       for v := 1 to max do
        if (v in graphvertices) then
         if (notknown in ugraph.augraph^[v,w]) and (notknown in ugraph.augraph^[w,v])
            and (alpha[v] > i) then
           begin
             x := v;
             while (index[x] > i) do
               begin
                 index[x] := i;
                 if ugraph.augraph^[x,w] = [] then
                   begin
                       write(ofilvar,'Adding edge '); writename(x,ofilvar); write(ofilvar,' '); 
                       writename(w,ofilvar); writeln(ofilvar);
                       ugraph.augraph^[x,w] := [notknown];
                       ugraph.augraph^[w,x] := [notknown];
                   end;
                 x := f[x];
               end;
             if f[x] = x then f[x] := w;
           end;
       end;
end;

 
procedure copypag(var pag1,pag2: graphrec; includebest: boolean);
var x,y: vertextype; ok: boolean;
begin
    for x := 1 to max do
     for y := 1 to max do
       if (reqpag.augraph^[x,y] <> []) and includebest then 
           pag2.augraph^[x,y] := reqpag.augraph^[x,y]
       else pag2.augraph^[x,y] := pag1.augraph^[x,y];
    ok := true;
    calcdpath(pag2,ok);
    disposetri(pag2.nocollidelist);
    copylist(pag1.nocollidelist,pag2.nocollidelist);
    disposetri(pag2.bothlist);
    copylist(pag1.bothlist,pag2.bothlist);
end;
 
     
procedure pagtomag(var thepag,mag: graphrec; var isok: boolean; level: integer);
var noconsistent,ok,foundone,someadded: boolean; var curpos1,curpos2: triangleptr;
    i,j: integer; start,finish: array[1..10000] of integer; 
    startlist1,startlist2: array[1..10000] of triangleptr;
begin
 fillingraph(thepag);
 copypag(thepag,mag,false);
 i := 1;
 j := 1;
 repeat
   getthenextedge(i,j,foundone,mag);
   if foundone then
      begin
        noconsistent :=
          ([in2,out2] <= mag.augraph^[i,j]) or ([in2,out2] <= mag.augraph^[j,i]);
        if not noconsistent then 
             orientanedge(i,j,mag,true,ok,start[level],finish[level],startlist1[level],startlist2[level],level);
        if ok then ok := okmag(mag);
        if (not ok) then
          begin
            ok := true;
            undochanges(mag,start[level],finish[level],startlist1[level],startlist2[level],level);
            calcdpath(mag,ok);
            if not noconsistent then 
                orientanedge(j,i,mag,true,ok,start[level],finish[level],startlist1[level],startlist2[level],level);
            if ok then ok := okmag(mag);
         end;
     end;
   level := level + 1;
 until (not ok) or (not foundone);
end;

(* procedure pagtomag(var thepag,mag: graphrec; i,j: integer; var isok: boolean; level: integer);
var noconsistent,ok,foundone: boolean; var curpos1,curpos2: triangleptr;
    x,y: vertextype; start,finish: integer; startlist1,startlist2: triangleptr;
begin
   getthenextedge(i,j,foundone,mag);
   if foundone then
      begin
        noconsistent :=
          ([in2,out2] <= mag.augraph^[i,j]) or ([in2,out2] <= mag.augraph^[j,i]);
        if not noconsistent then 
             orientanedge(i,j,mag,true,ok,start,finish,startlist1,startlist2,level);
        if (ok) and okmag(mag) then
              pagtomag(thepag,mag,i,j,isok,level)
        else 
          begin
            isok := true;
            undochanges(mag,start,finish,startlist1,startlist2,level);
            calcdpath(mag,ok);
            if not noconsistent then 
                orientanedge(i,j,mag,true,ok,start,finish,startlist1,startlist2,level);
            if (ok) and okmag(mag) then
               pagtomag(thepag,mag,i,j,isok,level)
            else isok := false;
         end;
     end;
end;
*)

(********************************************************************************)

procedure initrec(* var graphinfo: graphrec *);
begin
  with graphinfo do
    begin
      tnewadj(ancest,fatal);
      tnewend(augraph,fatal);
      nocollidelist := nil;
      bothlist := nil;
      new(recchange);
      curchange := 1;
    end;
end;


procedure dseptocond(a,b: vertextype; var varpair: partialpair; thedsep: vertexset;
                     var conditions: partialedonarray);
var i: vertextype;
begin
  varpair[1] := a;
  varpair[2] := b;
  conditions.length := 0;
  for i := 1 to max do 
     if i in thedsep then
       begin
         conditions.length := conditions.length + 1;
         conditions.arr[conditions.length] := i;
       end;
end;
   
procedure calcdsep(a,b: vertextype; var ugraph: endarray; var thedsep: vertexset; var an: adjtype);
var i,lastadded,curpos,curvertex: vertextype; expand: vertexset; 
    expandlist: array[0..maxvertex] of vertextype; 
begin
  expand := [a];
  lastadded := 1;
  expandlist[lastadded] := a;
  thedsep := [];
  curpos := 1;
  repeat
    curvertex := expandlist[curpos];    
    for i := 1 to max do
      if (i in graphvertices) then
       begin
         if (not (i in expand)) and (i <> b) then
            if ((an^[i,a] <> notthere) or (an^[i,b] <> notthere)) and
               (in2 in ugraph^[i,curvertex]) then
              begin
                thedsep := thedsep + [i];
                if (in2 in ugraph^[curvertex,i]) and (not (i in expand)) then
                   begin
                      lastadded := lastadded + 1;
                      expandlist[lastadded] := i;
                      expand := expand + [i];
                   end;
              end;
        end;
     curpos := curpos + 1;
  until curpos > lastadded;
end;

procedure repairpag(var ugraph,mag: graphrec);
label 666;
var a,b,i,x,y: vertextype; change1,change2,held,isok,ok: boolean; varpair: partialpair; 
    conditions: partialedonarray; thedsep1,thedsep2: vertexset; aprob,g2,aval: real;
begin
 with ugraph do
  begin    
    repeat
      change1 := false;
      repeat
      for a := 1 to max-1 do 
       if (a in graphvertices) then
        for b := a + 1 to max do
          if (b in graphvertices) then
	   if augraph^[a,b] = [] then
             begin
                if mag.ancest^[a,b] = notthere then
                   calcdsep(a,b,mag.augraph,thedsep1,mag.ancest);
                if mag.ancest^[b,a] = notthere then
                   calcdsep(b,a,mag.augraph,thedsep2,mag.ancest);
                
                (* choose smallest dsep for thedsep1 *)
                if mag.ancest^[a,b] <> notthere then thedsep1 := thedsep2
                else if (setsize(thedsep2) < setsize(thedsep1)) and (mag.ancest^[b,a] = notthere) then
                      thedsep1 := thedsep2;
                dseptocond(a,b,varpair,thedsep1,conditions);
                if (not discrete) then
                    begin
                      aval := calcpart1(varpair,conditions,ok);
                      calcpartialheld(aval,aprob,held,conditions.length);
                    end
                else held := calcindep1(varpair,aprob,g2,conditions,numchecked);
                if (not held) and ((not discrete) or (g2 <> 0)) then
                  begin
                    augraph^[a,b] := [notknown];
                    augraph^[b,a] := [notknown];
                    unorient(ugraph.augraph);
                    writeln(ofilvar,'Adding edge ');
                    if (not discrete) then
                      printpartial(varpair,conditions,ofilvar,aprob,aval)
                    else printpartial(varpair,conditions,ofilvar,aprob,g2);
                    orientedges(ugraph,true,1,false);
                    isok := true;
                    pagtomag(ugraph,mag,isok,1);
                    change1 := true;
                    goto 666;
                  end
              end
        until not change1;
        change2 := false;
(*        repeat       should remove a,b edge; check to see what d-separates; 
         for a := 1 to max-1 do 
          for b := a + 1 to max do
              begin
                 if mag.ancest^[a,b] = notthere then
                    calcdsep(a,b,mag.augraph,thedsep1,mag.ancest)
                 else calcdsep(b,a,mag.augraph,thedsep1,mag.ancest);
                 dseptocond(a,b,varpair,thedsep1,conditions);
                 if (not discrete) then
                    begin
                      aval := calcpart1(varpair,conditions,ok);
                      calcpartialheld(aval,aprob,held,conditions.length);
                    end
                else held := calcindep1(varpair,aprob,g2,conditions,numchecked);
                dseptocond(a,b,varpair,thedsep1,conditions);
                if (not discrete) then
                    begin
                      aval := calcpart1(varpair,conditions,ok);
                      calcpartialheld(aval,aprob,held,conditions.length);
                    end
                else held := calcindep1(varpair,aprob,g2,conditions,numchecked);
                if held then
                   begin
                    augraph^[a,b] := [];
                    augraph^[b,a] := [];
                    writeln('Removing edge ');
                    if (not discrete) then
                      printpartial(varpair,conditions,output,aprob,aval)
                    else printpartial(varpair,conditions,output,aprob,g2);
                    isok := true;
                    orientedges(ugraph,true,1,false);                    
                    unorient(ugraph.augraph);
                    orientedges(ugraph,true,1,false);
                    change2 := true;
                    pagtomag(ugraph,mag,isok,1);
                  end
               end
         until not change2; *)
666: until (not change1) and (not change2);
 end;
end;

procedure scoremagdiscrete(mag: graphrec; var thescore: real);
var i,j,oldmax,k: vertextype; oldgraphvertices: vertexset; df: integer;
begin
   oldgraphvertices := graphvertices;
   oldmax := max;
   for i := 1 to oldmax-1 do
     for j := i+1 to oldmax do
        begin
           adjacency^[i,j] := notthere;
           adjacency^[j,i] := notthere;
           if (in2 in mag.augraph^[i,j]) and (in2 in mag.augraph^[j,i]) then
               begin
                  max := max + 1;
                  for k := 1 to max-1 do 
                      begin
                         adjacency^[max,k] := notthere;
                         adjacency^[k,max] := notthere;
                      end;
                  adjacency^[max,i] := 0;
                  adjacency^[max,j] := 0;
               end
           else if (in2 in mag.augraph^[i,j]) then adjacency^[i,j] := 0
           else if (in2 in mag.augraph^[j,i]) then adjacency^[j,i] := 0;
       end;
   graphvertices := [1..max];
   em(tempcats,cats,thescore);
   score := score - df * ln(sampsize * max);
   max := oldmax;
   graphvertices := oldgraphvertices;
end;

procedure magtopag(var mag,pag: graphrec);
var ok: boolean; i,j: vertextype; thedsep: vertexset;
begin
   ok := true;
   calcdpath(mag,ok);
   for i := 1 to max-1 do
     for j := i+1 to max do
       if (mag.augraph^[i,j] = []) then 
         if (mag.ancest^[i,j] = notthere) then
           begin
             calcdsep(i,j,mag.augraph,thedsep,mag.ancest);              
             pag.augraph^[i,j] := [];
             pag.augraph^[j,i] := [];
             dsep^[i,j] := thedsep;
             dsep^[j,i] := thedsep;
           end
         else
           begin
             calcdsep(j,i,mag.augraph,thedsep,mag.ancest);              
             pag.augraph^[i,j] := [];
             pag.augraph^[j,i] := [];
             dsep^[i,j] := thedsep;
             dsep^[j,i] := thedsep;
           end
       else
           begin
             pag.augraph^[i,j] := [notknown];
             pag.augraph^[j,i] := [notknown];
           end;
    calcdpath(pag,ok);
    disposetri(pag.bothlist);
    disposetri(pag.nocollidelist);
    orientedges(pag,true,1,true);
end;

procedure printpoipg(var ugraph: graphrec);
var i,j: vertextype; head,tail: vertextype; written: vertexset; 
    curpos: triangleptr; noconsistent: boolean; 
begin
with ugraph do
 begin
  written := [];
  twriteeoln(output);
  for i := 1 to numob-1 do
    for j := i+1 to numob do
     if (i in graphvertices) and (j in graphvertices) then
      if (augraph^[i,j] <> []) or (augraph^[j,i] <> []) then
        begin
          written := written + [i,j];
          if (knownin2 in augraph^[i,j]) then tail := i
          else if (knownin2 in augraph^[j,i]) then tail := j
          else if (in2 in augraph^[i,j]) then tail := i
          else if (in2 in augraph^[j,i]) then tail := j
          else tail := i;
          if tail = i then head := j 
          else head := i;
          noconsistent :=
             ([in2,out2] <= augraph^[tail,head]) or 
             ([in2,knownout2] <= augraph^[tail,head]) or             
             ([out2,knownin2] <= augraph^[tail,head]) or
             ([out2,in2] <= augraph^[tail,head]) or 
             ([in2,out2] <= augraph^[head,tail]) or
             ([in2,knownout2] <= augraph^[head,tail]) or             
             ([out2,knownin2] <= augraph^[head,tail]) or
             ([out2,in2] <= augraph^[head,tail]);
	  writename(tail,output);
          if noconsistent then twrite(output,'   |')
          else
              begin
                if ((in2 in augraph^[head,tail]) or 
                    (knownin2 in augraph^[head,tail])) then
                       twrite(output,' <|')
                else if ((out2 in augraph^[head,tail]) or
                         (knownout2 in augraph^[head,tail])) then
                            twrite(output,' -|')
                else twrite(output,' o|');
                twrite(output,'-|');
                if ((in2 in augraph^[tail,head]) or
                    (knownin2 in augraph^[tail,head])) then
                       twrite(output,'> |')
                else if ((out2 in augraph^[tail,head]) or
                         (knownout2 in augraph^[tail,head])) then
                            twrite(output,'- |')
                else twrite(output,'o |');
              end;
	  writename(head,output);
          if noconsistent then twrite(output,'#|');
          twriteeoln(output);
     end;
  for i := 1 to numob do if (i in graphvertices) and (not (i in written)) then
     begin
	  pwritename(i,output,7);
          twriteeoln(output);
     end;
  twriteeoln(output);
  curpos := nocollidelist;
  if curpos <> nil then
      twriteln(output,'The following pairs of edges do not collide: |');
  while curpos <> nil do
     begin
        writename(curpos^.inout,output);
        twrite(output,' -- |');
        writename(curpos^.unori,output);
        twrite(output,'    |');
        writename(curpos^.unori,output);
        twrite(output,' -- |');
        writename(curpos^.other,output);
	twriteeoln(output);        
        curpos := curpos^.fwd;
     end;
   twriteeoln(output);        
 end;
end;

procedure getnext(start,x,y: vertextype; var next,endpoint: vertextype; var bothends: boolean; var modified: graphrec);
var i: vertextype; 
begin
  next := 0;
  i := start;
  with modified do
    while (next = 0) and (i <= max) do
      if (augraph^[y,i] <> []) and (augraph^[x,i] = []) and
         (not (in2 in augraph^[y,i])) and (not (out2 in augraph^[y,i])) then
           begin
             next := i;
             endpoint := y;
           end
      else if bothends and (augraph^[x,i] <> []) and 
           (augraph^[y,i] = []) and (not (in2 in augraph^[x,i])) and
           (not (out2 in augraph^[x,i])) then
              begin
                next := i;
                endpoint := x;
              end
      else i := i + 1;
end;

function samepags(var pag1,pag2: graphrec): boolean;
label 999;
var i,j: integer; curpos1,curpos2: triangleptr;
begin
   samepags := true;
   for i := 1 to numob do
      for j := 1 to numob do
        if pag1.augraph^[i,j] <> pag2.augraph^[i,j] then
            begin
              samepags := false;
              goto 999;
            end;
   curpos1 := pag1.nocollidelist;
   curpos2 := pag2.nocollidelist;
   while (curpos1 <> nil) do
      begin
         if (curpos1^.inout <> curpos2^.inout) or
            (curpos1^.unori <> curpos2^.unori) or
            (curpos1^.other <> curpos2^.other) then
                 begin
                    samepags := false;
                    goto 999;
                 end;
         curpos1 := curpos1^.fwd;
         curpos2 := curpos2^.fwd;
      end;
999: end;

procedure insertpag(var score,lowestbest: real; var thispag: graphrec);
(* this assumes that if the score are equal, then they are the same model - 
   should this be changed? *)
label 999; 
var i,j: integer; 
begin
  if (score > lowestbest) then
    for i := 1 to maxbest do
       if score >= thebest[i].score then 
         begin
           if score = thebest[i].score then 
             goto 999;
           lowestbest := thebest[maxbest-1].score;
           if thebest[maxbest].augraph <> nil then
              begin
                dispose(thebest[maxbest].augraph);
                disposetri(thebest[maxbest].nocollidelist);
                dispose(thebest[maxbest].ancest);
                dispose(thebest[maxbest].augraph);
                dispose(thebest[maxbest].recchange);
                disposetri(thebest[maxbest].bothlist);
              end;
           for j := maxbest downto i+1 do
               thebest[j] := thebest[j-1];
           thebest[i].score := score;
           initrec(thebest[i]);
           magtopag(thispag,thebest[i]);
           goto 999;
         end;
999: end;


procedure getcombos(thestart,x,y: vertextype; bothends: boolean; var changedscore: boolean; firsttime: boolean; 
                    var best,modified,mag: graphrec; var bestscore: real; curpos: triangleptr; var changed: boolean; 
                    level: integer; var lowestbest: real; var aux: auxiliary);
var next,endpoint: vertextype; ok,isok: boolean; start,start2,finish,finish2: integer; 
                   startlist1,startlist2,startlist1a,startlist2a: triangleptr;
begin
   startlist1 := nil; startlist2 := nil; startlist1a := nil; startlist2a := nil;
   ok := true;
   getnext(thestart,x,y,next,endpoint,bothends,modified);
   if next <> 0 then
     begin
      changed := true;
      ok := true;
      orientanedge(next,endpoint,modified,false,ok,start,finish,startlist1,startlist2,level);
      if ok then getcombos(next+1,x,y,bothends,changedscore,firsttime,best,modified,mag,bestscore,modified.bothlist,changed,level+1,lowestbest,aux);
      undochanges(modified,start,finish,startlist1,startlist2,level);
      ok := true;
      orientanedge(endpoint,next,modified,true,ok,start,finish,startlist1,startlist2,level);
      if ok then getcombos(next+1,x,y,bothends,changedscore,firsttime,best,modified,mag,bestscore,modified.bothlist,changed,level+1,lowestbest,aux);
      undochanges(modified,start,finish,startlist1,startlist2,level);
    end
  else 
    begin
      if curpos <> nil then
         begin
           changed := true;
           ok := true;

(* so not oriented away from curpos^.inout,curpos^.unori *)
           modified.augraph^[curpos^.other,curpos^.unori] := [in2];
           orientanedge(curpos^.inout,curpos^.unori,modified,false,ok,start,finish2,startlist1,startlist2,level);

(* so can record changes with orientanedge *)
           modified.augraph^[curpos^.other,curpos^.unori] := [notknown];
           orientanedge(curpos^.other,curpos^.unori,modified,false,ok,start2,finish,startlist1a,startlist2a,level);  
           if ok then getcombos(max+1,x,y,bothends,changedscore,firsttime,best,modified,mag,bestscore,curpos^.fwd,changed,level+1,lowestbest,aux);
           undochanges(modified,start,finish,startlist1,startlist2a,level);
           ok := true;
           orientnon(curpos,modified,ok,start,finish,startlist1,startlist2,level);
           if ok then getcombos(max+1,x,y,bothends,changedscore,firsttime,best,modified,mag,bestscore,curpos^.fwd,changed,level+1,lowestbest,aux);
           undochanges(modified,start,finish,startlist1,startlist2,level);
           if ok then getcombos(max+1,x,y,bothends,changedscore,firsttime,best,modified,mag,bestscore,curpos^.fwd,changed,level+1,lowestbest,aux);
         end
      else if (changed or firsttime) and ok then
         begin
            mag.curchange := 1;
            isok := true;
            pagtomag(modified,mag,isok,3);
            if isok then 
              begin
                 writeln(output,'Modified');
                 printpoipg(modified);
                 writeln(output,'MAG');
                 printpoipg(mag);
                 scoremag(mag,score,aux);
                 writeln(output,'Score =',score:8:4);
                 insertpag(score,lowestbest,mag);
                 if (score > bestscore) then
                      begin
                         magtopag(mag,best);
                         writeln(output,'New Best');
                         printpoipg(best);
                         bestscore := score;
                         changedscore := true;
                      end;
             end;
         end;
   end;
end;
       
procedure copycolliders(var current,modified: graphrec; end1,end2: vertextype;
                        both: boolean);
var i,j,k: vertextype; var ok: boolean;
begin
 disposetri(modified.nocollidelist);
 disposetri(modified.bothlist);
 with current do
  begin
   for i := 1 to max do
     for j := 1 to max do
        if augraph^[i,j] = [] then modified.augraph^[i,j] := []
        else if reqpag.augraph^[i,j] <> [] then modified.augraph^[i,j] := reqpag.augraph^[i,j]
        else modified.augraph^[i,j] := [notknown];
   for i := 1 to max-1 do
     for j := i+1 to max do
       for k := 1 to max do
         if (k <> i) and (k <> j) and (in2 in augraph^[i,k]) and
            (augraph^[i,j] = []) and (in2 in augraph^[j,k]) and
            (not ((i = end1) and (j = end2))) and
            (not ((i = end2) and (j = end1))) then
               begin
                  modified.augraph^[i,k] := [in2];
                  modified.augraph^[j,k] := [in2];
               end;
   modified.augraph^[end1,end2] := [in2];
   if both then modified.augraph^[end2,end1] := [in2] 
   else modified.augraph^[end2,end1] := [out2];
   calcdpath(modified,ok);
  end;
end;

       
procedure copysome(var current,modified: graphrec; end1,end2: vertextype;
                        both: boolean);
var i,j,k: vertextype; ok: boolean;
begin
 disposetri(modified.nocollidelist);
 disposetri(modified.bothlist);
 with current do
  begin
   for i := 1 to max do
     for j := 1 to max do
        if reqpag.augraph^[i,j] <> [] then modified.augraph^[i,j] := reqpag.augraph^[i,j]
        else if augraph^[i,j] <> [] then modified.augraph^[i,j] := [notknown]
        else modified.augraph^[i,j] := [];
   for i := 1 to max-1 do
     for j := i+1 to max do
       for k := 1 to max do
         if (k <> i) and (k <> j) then
           if (in2 in augraph^[i,k]) and (augraph^[i,j] = []) and 
              (in2 in augraph^[j,k]) and 
              (not ((i = end1) and (k = end2))) and
              (not ((i = end2) and (k = end1))) and 
              (not ((j = end1) and (k = end2))) and 
              (not ((j = end2) and (k = end1))) then
                begin
                  modified.augraph^[i,k] := [in2];
                  modified.augraph^[j,k] := [in2];
                end
           else if (augraph^[i,k] <> []) and (augraph^[j,k] <> []) and
                   ((notknown in augraph^[i,k]) or (notknown in augraph^[j,k])) and
                   ((i = end1) and (j = end2)) or ((i = end2) and (j = end1)) then
                  addtolist(i,k,j,modified.bothlist);
   modified.augraph^[end1,end2] := [];
   modified.augraph^[end2,end1] := [];
   calcdpath(modified,ok);
  end;
end;

procedure asearch(* var temp1pag: graphrec *);
var best,current,modified,mag: graphrec; bestscore,score: real;
    changedscore,ok,changed1,changed2: boolean;
    i,j: vertextype; start,finish,start2,finish2,splitindex,getsplit: integer; 
    startlist,startlist2: triangleptr;
    tempend1,tempend2: endset; lowestbest: real; aux: auxiliary;    
begin
 tnewadj(aux.parent,fatal);
 tnewadj(aux.doublehead,fatal);
 tnewadj(aux.tempadj,fatal);
 tnewcov(aux.localparm,fatal);
 tnewcov(aux.localcov,fatal);
 tnewcov(aux.localestcov,fatal);
 tnewcov(aux.estcov,fatal); 
 tnewcov(aux.tempcov,fatal);
 tnewcov(aux.theinverse,fatal);
 initrec(aux.apag); (* used in scoremag *)
 for i := 1 to maxbest do
   begin
     thebest[i].augraph := nil;
     thebest[i].score := -1000000000;
     splits[i].augraph := nil;
   end;
 lowestbest := -1000000000;
 tnewdsep(dsep,fatal);
 initrec(best);
 initrec(modified);
 initrec(mag);
 initrec(current);
 tnewadj(collidera,fatal);
 initrec(splits[1]);
 splitindex := 1;
 getsplit := 1;
 copypag(temp1pag,splits[1],true);
 with best do
  begin
   repeat
    copypag(splits[getsplit],best,true);
    mag.curchange := 1;
    ok := true;
    pagtomag(best,mag,ok,3);
    if ok then 
     begin
      scoremag(mag,score,aux);
      if (getsplit = 1) then bestscore := score;
      if (getsplit = 1) then insertpag(bestscore,lowestbest,mag);
      repeat
       changedscore := false;
       copypag(best,current,true);
       ok := true;
       calcdpath(current,ok);
       if getsplit = 1 then
         for i := 2 to maxbest do
          if (thebest[i].score >= (1.005 * thebest[1].score)) and (splitindex < maxbest) then
             begin
                initrec(splits[splitindex]);
                copypag(thebest[i],splits[splitindex],true);
                splitindex := splitindex + 1;
             end;
       for i := 1 to max-1 do 
         for j := i+1 to max do
           begin
            changed1 := false;
            changed2 := false;
            if (augraph^[i,j] = []) and 
               (not ((notknown in forbidpag.augraph^[i,j]) and (notknown in forbidpag.augraph^[j,i]))) then
              begin
                 if (current.ancest^[j,i] = notthere) and (not ((forbidpag.augraph^[j,i] * [out2,notknown] <> []) and
                    (forbidpag.augraph^[i,j] * [in2] <> []))) and (forbiddenadj^[i,j] = notthere) then
                   begin
                     copycolliders(current,modified,i,j,false);
                     orientedges(modified,false,1,true);
                     getcombos(1,i,j,false,changedscore,true,best,modified,mag,bestscore,modified.bothlist,changed1,3,lowestbest,aux);
                   end;
                 if (current.ancest^[i,j] = notthere)  and (not ((forbidpag.augraph^[i,j] * [out2,notknown] <> []) and
                    (forbidpag.augraph^[j,i] * [in2] <> []))) and (forbiddenadj^[j,i] = notthere) then
                   begin
                      copycolliders(current,modified,j,i,false);
                      orientedges(modified,false,1,true);
                      getcombos(1,j,i,false,changedscore,false,best,modified,mag,bestscore,modified.bothlist,changed2,3,lowestbest,aux);
                   end;
                 if (current.ancest^[i,j] = notthere) and (current.ancest^[j,i] = notthere) and changed1 and changed2 
(*                    and ((out2 in forbidpag.augraph^[j,i]) or (out2 in forbidpag.augraph^[i,j]))  *)
                    and (commonfor^[i,j] = notthere) then
                    begin
                       copycolliders(current,modified,i,j,true);
                       orientedges(modified,false,1,true);
                       getcombos(1,i,j,true,changedscore,false,best,modified,mag,bestscore,modified.bothlist,changed1,3,lowestbest,aux);
                    end;
                 modified.augraph^[j,i] := [];
                 modified.augraph^[i,j] := [];                 
               end;
          end;
      until (not changedscore);
      repeat
       changedscore := false;
       copypag(best,current,true);
       ok := true;
       calcdpath(current,ok);
       for i := 1 to max-1 do 
         for j := i+1 to max do
          if (reqpag.augraph^[i,j] = []) and (reqpag.augraph^[j,i] = []) then
           begin
            changed1 := false;
            changed2 := false;
            if (augraph^[i,j] <> []) then
              begin
                 tempend1 := augraph^[i,j];
                 tempend2 := augraph^[j,i];
                 copysome(current,modified,i,j,false);
                 orientedges(modified,false,1,true);
                 getcombos(1,i,j,false,changedscore,true,best,modified,mag,bestscore,modified.bothlist,changed1,3,lowestbest,aux);
                 modified.augraph^[j,i] := tempend1;
                 modified.augraph^[i,j] := tempend2;           
               end;
          end;
       until (not changedscore);
     end;
   getsplit := getsplit + 1;
   until (getsplit > maxbest) or (splits[getsplit].augraph = nil);
 end;
  for i := 1 to maxbest do
   begin
    if (thebest[i].augraph <> nil) then
     begin
        printthepoipg(ofilvar,thebest[i],thebest[i].nocollidelist,thebest[i].score);
        dispose(thebest[i].ancest);
        dispose(thebest[i].augraph);
        disposetri(thebest[i].nocollidelist);
     end;
   if (splits[i].augraph <> nil) then
     begin
        dispose(splits[i].ancest);
        dispose(splits[i].augraph);
        disposetri(splits[i].nocollidelist);
     end;
   end;
  dispose(best.augraph);
  dispose(mag.augraph);
  dispose(modified.augraph);
  dispose(current.augraph);
  dispose(aux.apag.augraph);
  dispose(best.ancest);
  dispose(mag.ancest);
  dispose(modified.ancest);
  dispose(current.ancest);
  dispose(aux.apag.ancest);
  disposetri(best.nocollidelist);
  disposetri(modified.nocollidelist);
  disposetri(mag.nocollidelist);
  disposetri(aux.apag.nocollidelist);
  dispose(aux.tempcov);
  dispose(aux.theinverse);
  dispose(aux.parent);
  dispose(aux.tempadj);
  dispose(aux.estcov);
  dispose(aux.doublehead);
  dispose(aux.localparm);
  dispose(aux.localcov);
  dispose(aux.localestcov);
end;

procedure restoreedges(var ugraph: graphrec; var edgelist: nodeptr);
var curnode: nodeptr;
begin
   curnode := edgelist;
   while (curnode <> nil) do
     begin
       ugraph.augraph^[edgetovert^[curnode^.edge].head,edgetovert^[curnode^.edge].tail] := [notknown];
       ugraph.augraph^[edgetovert^[curnode^.edge].tail,edgetovert^[curnode^.edge].head] := [notknown];
       curnode := curnode^.fwd;
     end;
end;

procedure insertthepag(ugraph: graphrec; var thebest: ugrapharray);
var i: integer; finished: boolean;
begin
    i := 1;
    finished := false;
    repeat
       if thebest[i].augraph <>  nil then i := i + 1
       else
         begin
            initrec(thebest[i]);
            copypag(ugraph,thebest[i],false);
            finished := true;
         end
    until (i > maxbest) or finished;
end;

procedure getpartialnonsuff(var ugraph: graphrec; crossvalidate: boolean);
var i,j: vertextype; curpos: edgeptr; partial: partialpair; num: integer;
    aprob,g2: real; partialedon,vertices: partialedonarray; c: char;
    held: boolean; pathset1,pathset2: vertexset; condset: parentstype;
    nochange,ok: boolean; temp,edgelist,newlist,newlist1,curnode1,curnode2,curnode3: nodeptr; 
    adj: adjtype; mag: graphrec; isok: boolean; x,y: vertextype; 
label 999;
begin
   tnewadj(ugraph.ancest,fatal);

(*******************************************************)
   initrec(mag);
   tnewadj(collidera,fatal);
   tnewadj(tempadj,fatal);
(*******************************************************)


   if fatal in errors then goto 999;
   for i := 1 to numob do 
      begin
        corrlists[i] := nil;
        corrlists1[i] := nil;
      end;
   numchecked := 1;
   edgelist := nil;
   newlist := nil;
   newlist1 := nil;
   num := 1;
   ok := true;
   toomuchtime := false;
   start := tgettime;
   numedges := tround(setsize(graphvertices) * (setsize(graphvertices) - 1) / 2);
(*   if discrete then 
      if sizeof(cellsarr) < heapavail then new(cells)
      else 
         begin
            errormess('Out of memory.|',fatal,false);
            goto 999;
         end;  *)
   initugraph(adjacency,ugraph);
   for i := 1 to numob do parents[i] := [];
   orientfromknow(ugraph.augraph);
   initparents(edgelist,false,ugraph.augraph);
   checkedges(edgelist,newlist,newlist1,false,ugraph.augraph,corrlists);
   restoreedges(ugraph,newlist1);
   disposeedgelist(edgelist);
   disposeedgelist(newlist);
   checkedges(newlist1,edgelist,newlist,false,ugraph.augraph,corrlists);   
   disposeedgelist(edgelist);
   disposeedgelist(newlist1);
   if aaabort then goto 999;
   if toomuchtime then 
	twriteln(ofilvar,'Search aborted because time limits exceeded.|');
   if not heuristic then
       begin
          setcollider(false,nochange,ok,ugraph,1);    
          for i := 1 to numob do 
            if i in graphvertices then
              begin
                 condset[i] := [];
                 pathset1 := [i];
                 pathset2 := [i];
                 findcond(i,0,i,condset[i],pathset1,pathset2,ugraph.augraph);
              end;      
          for i := 1 to numob do
           if i in graphvertices then
            begin            
              for j := 1 to numob do
                if j in graphvertices then
                   if j in (condset[i] - parents[i]) then
                      begin
                        if aaabort then goto 999;
                        new(curnode1);
                        if discrete then 
                           begin
                             partial[1] := i;
                             partial[2] := j;
                             partialedon.length := 0;
                             held := calcindep1(partial,aprob,g2,partialedon,numchecked);
                             curnode1^.prob := g2;
                           end
                        else curnode1^.prob := rho^[i,j];
                        curnode1^.edge := j;
                        leafinsert(curnode1,corrlists[i],false);
                     end;
              parents[i] := parents[i] + condset[i];
            end;
          checkedges(newlist,newlist1,edgelist,false,ugraph.augraph,corrlists1);
          disposeedgelist(newlist);
          disposeedgelist(newlist1);
          disposeedgelist(edgelist);
          for i := 1 to max do
             begin
               disposeedgelist(corrlists[i]);
               disposeedgelist(corrlists1[i]);
             end;
    if toomuchtime then 
	 twriteln(ofilvar,'Search aborted because time limits exceeded.|');

          unorient(ugraph.augraph);
       end;
    disposeedgelist(edgelist);
    disposeedgelist(newlist);
    for i := 1 to numob do
     for j := 1 to numob do
      if (i in graphvertices) and (j in graphvertices) then
          if adjacency^[i,j] <> notthere then
             ugraph.augraph^[i,j] := [knownin2]
          else if adjacency^[j,i] <> notthere then
             ugraph.augraph^[i,j] := [knownout2];
    if not heuristic then orientedges(ugraph,true,1,false)
    else 
      begin
         setcollider(true,nochange,ok,ugraph,1);
         calcdpath(ugraph,ok);
      end;
    twriteeoln(ofilvar);
999:if aaabort then twriteln(ofilvar,'Search aborted by interrupt.|');
    twriteln(ofilvar,'---------------------------------------------------|');
    twriteln(ofilvar,'NOT assuming causal sufficiency|');
    twriteln(ofilvar,'The Partial Ancestral Graph (PAG):|');
    twrite(ofilvar,'Significance Level = |');
    twritereal(ofilvar,signif);
    twriteeoln(ofilvar);
    isok := true;
(*    pagtomag(ugraph,mag,isok,1);
    if debug then repairpag(ugraph,mag); *)
    if (not discrete) and (not indepdata) and (not celldata) then
       debug := getreply('Do you want to perform a MAG search? |','y',debug); 
    if debug then asearch(ugraph); 
    printpoipg(ugraph);
    if crossvalidate then insertthepag(ugraph,thebest);
    printthepoipg(ofilvar,ugraph,ugraph.nocollidelist,ugraph.score); 
    dispose(tempadj);
    dispose(collidera);
    dispose(mag.augraph);
    disposetri(mag.nocollidelist);    
end;


procedure fillpartialedon(var partialedon: partialedonarray;  
                          partial: partialpair; k,depth,length: integer);
label 999;
var next,i: vertextype; aval,aprob: real; held,ok: boolean;
begin
   partialedon.arr[depth] := k;
   if (length = 0) or (depth = length) then
      begin
         aval := calcpart1(partial,partialedon,ok);
         if not ok then held := false
         else calcpartialheld(aval,aprob,held,partialedon.length);
         if held then 
           begin
            for i := 1 to partialedon.length do 
                zeroes^[partial[1],partial[2]] := 
                   zeroes^[partial[1],partial[2]] + [partialedon.arr[i]];
            zeroes^[partial[2],partial[1]] :=  zeroes^[partial[1],partial[2]];
           end;
         if toomuchtime then goto 999;
      end
   else 
      for next := k + 1 to numob do
       if (k in graphvertices) then
        if (next <> partial[1]) and (next <> partial[2]) and 
           (next in graphvertices) then
         fillpartialedon(partialedon,partial,next,depth+1,length);
999: end;

procedure findlatents;
var i,j,k,thelength,length: vertextype; partialedon: partialedonarray; 
    partial: partialpair;
begin
  for i := 1 to numob do
    for j := 1 to numob do
     if (i in graphvertices) and (j in graphvertices) then
        zeroes^[i,j] := [i,j];
  thelength := setsize(graphvertices);
  for length := 0 to thelength - 2 do
    begin
       partialedon.length := length;
       for i := 1 to numob - 1 do
         for j := i + 1 to numob do
           if (i in graphvertices) and (j in graphvertices) then
             begin
                partial[1] := i;
                partial[2] := j;
                if (length = 0) then                     
                   fillpartialedon(partialedon,partial,k,1,length)
                else
                  for k := 1 to numob do 
                    if (k <> partial[1]) and (k <> partial[2]) and
                       (k in graphvertices) then
                         fillpartialedon(partialedon,partial,k,1,length);
             end;
     end;
  if not aaabort then checklatents;
end;


procedure getnextedge(var i,j: integer; var found: boolean;
		      var newadj: adjtype;
		      var ugraph: endarray);
var start1,start2,tempi,tempj: integer;
begin
   start1 := i;
   start2 := j;
   found := false;
   if j = max then 
      begin
        start1 := start1 + 1;
        start2 := 1;
      end
   else start2 := start2 + 1;
   tempi := start1;
   tempj := start2;
   repeat
      while (tempj <= max) and (not found) do
        begin
	 if (tempi <> tempj) and (newadj^[tempi,tempj] = notthere) and
	    (newadj^[tempj,tempi] = notthere) and
            ((ugraph^[tempi,tempj] <> []) or (ugraph^[tempi,tempj] <> [])) then
	    begin
	       i := tempi;
               j := tempj;
               found := true;
            end
         else tempj := tempj + 1;
       end;
      tempi := tempi + 1;
      tempj := tempi + 1;
   until (tempi >= max) or found;
end;
   
procedure undo(var newadj: adjtype; depth: integer);
var i,j: integer;
begin
   for i := 1 to max do
     for j := 1 to max do
        if newadj^[i,j] = depth then newadj^[i,j] := notthere;
end;


function newiscycle(*var newadj: modeltypeptr): boolean;*);
label 999;
var j,k,i: integer;

begin (* calcpath *)
    newiscycle := false;
    if fatal in errors then goto 999;
    for i := 1 to numobgraph do 
      for j := 1 to numobgraph do
        if i <> j then 
            begin
             if graphlist[i] in newadj^.amodel[graphlist[j]] then 
               path^[graphlist[i],graphlist[j]] := 0
             else path^[graphlist[i],graphlist[j]] := notthere
            end
          else path^[graphlist[i],graphlist[i]] := notthere;
    for k := 1 to numobgraph do 
       for i := 1 to numobgraph do 
           for j := 1 to numobgraph do        
               if (path^[graphlist[i],graphlist[k]] <> notthere) and
                  (path^[graphlist[k],graphlist[j]] <> notthere) then
                      path^[graphlist[i],graphlist[j]] := 1;
    for i := 1 to numobgraph do 
       if (path^[graphlist[i],graphlist[i]] <> notthere) then
          begin
            newiscycle := true;
(*            twriteln(output,'Warning: Graph is cyclic.|'); *)
            goto 999;
          end;
999:end; (* Calcdpath *)

procedure newpropagate(var newadj: adjtype;
		       var ugraph: endarray;
		       i,j,depth: integer; var ok: boolean);
var k: integer; noconsistent: boolean;
begin
   for k := 1 to max do
     if k <> i then
       if (ugraph^[i,k] = []) and (ugraph^[j,k] <> []) then
         begin
             if ((ugraph^[j,k] <> [out2]) and (newadj^[k,j] = notthere)) then
                  begin
                    if newadj^[j,k] = notthere then
                        begin
                           newadj^[j,k] := depth;
                           noconsistent :=
                              ([in2,out2] <= ugraph^[j,k]) or 
                              ([in2,knownout2] <= ugraph^[j,k]) or             
                              ([out2,knownin2] <= ugraph^[j,k]) or
                              ([out2,in2] <= ugraph^[j,k]) or 
                              ([in2,out2] <= ugraph^[k,j]) or
                              ([in2,knownout2] <= ugraph^[k,j]) or             
                              ([out2,knownin2] <= ugraph^[k,j]) or
                              ([out2,in2] <= ugraph^[k,j]);
                           if not noconsistent then
                              newpropagate(newadj,ugraph,j,k,depth,ok);
                        end
                  end
           else ok := false;
        end;
end;

procedure addmodtolist(var newadj: adjtype; patlist,patend: modeltypeptr);
var currentmod: modeltypeptr;
begin
   new(currentmod);
   currentmod^.fwd := nil; currentmod^.back := nil;
   currentmod^.score := 0;
   currentmod^.signif := signif;
   zreadadjtomod(newadj,currentmod);
   zinsertmodel(currentmod,zmodellist,zmodelend);
end;

procedure formgraph(var ugraph: endarray;
		    var newadj: adjtype; i,j: integer;
		    ok: boolean; newdepth: integer);
var foundone: boolean; noconsistent: boolean; 
begin
   getnextedge(i,j,foundone,newadj,ugraph);
   if foundone then
      begin
         noconsistent :=
          ([in2,out2] <= ugraph^[i,j]) or 
          ([in2,knownout2] <= ugraph^[i,j]) or             
          ([out2,knownin2] <= ugraph^[i,j]) or
          ([out2,in2] <= ugraph^[i,j]) or 
          ([in2,out2] <= ugraph^[j,i]) or
          ([in2,knownout2] <= ugraph^[j,i]) or             
          ([out2,knownin2] <= ugraph^[j,i]) or
          ([out2,in2] <= ugraph^[j,i]);
        newadj^[i,j] := newdepth;
        if not noconsistent then newpropagate(newadj,ugraph,i,j,newdepth,ok);
        if (not iscycle(newadj)) and ok then 
           formgraph(ugraph,newadj,i,j,ok,newdepth+1);
        if iscycle(newadj) or (not ok) then
          begin
            undo(newadj,newdepth);
            newadj^[j,i] := newdepth;
            ok := true;
            if not noconsistent then newpropagate(newadj,ugraph,j,i,newdepth,ok);
            if (not iscycle(newadj)) and ok then
                formgraph(ugraph,newadj,i,j,ok,newdepth+1);
            undo(newadj,newdepth);
         end;
     end
   else addmodtolist(newadj,zmodellist,zmodelend);
end;

procedure newwriteadj(var graph: adjtype; var outfile: text);
var i,j,k,l: integer; written: vertexset;
begin
   written := [];
   twriteln(outfile,'/graph|');
   for k := 1 to max do
      for l := 1 to max do
        if (k in graphvertices) and (l in graphvertices) and
           (graph^[k,l] <> notthere) then
		begin
		writename(k,outfile); twrite(outfile,' |');
		writename(l,outfile); twrite(outfile,' |');
		twriteeoln(outfile);
                written := written + [k,l];
		end;
   for i := 1 to max do 
      if (not (i in written)) then 
        begin
          writename(i,outfile);
          twriteeoln(outfile);
        end;  
   twriteeoln(outfile);
   twriteeoln(outfile);
end;

procedure writemod(var ofilvar: text; var amod: modeltypeptr);
var i,j: vertextype;
begin
   for i := 1 to max do
     for j := 1 to max do
        if (i in graphvertices) and (i in amod^.amodel[j]) then
           begin
             writename(i,ofilvar);
             twrite(ofilvar,' -> |');
             writename(j,ofilvar);
             twriteeoln(ofilvar);
           end;
end;

procedure writelist(var ofilvar: text; var zmodellist: modeltypeptr);
var curpos: modeltypeptr;
begin
   curpos := zmodellist;
   while curpos <> nil do
     begin
       writemod(ofilvar,curpos);
       twriteeoln(ofilvar);
       curpos := curpos^.fwd;
     end;
end;

procedure readmodtoadj(* var adjacency: adjtype; var currentmod: modeltypeptr *);
var i,j: vertextype;
begin
 if currentmod <> nil then
   for i := 1 to max do 
     for j := 1 to max do
       begin
          if i = 1 then
            begin
                theparents[j] := [];
                numparents[j] := 0;
            end;
          if (i in currentmod^.amodel[j]) then 
             begin
                adjacency^[i,j] := 0;
                theparents[j] := theparents[j] + [i];
                numparents[j] := numparents[j] + 1;
             end
          else adjacency^[i,j] := notthere;
       end;
end;

function samemod(mod1,mod2: modeltypeptr): boolean;
label 999;
var i: integer; 
begin
   samemod := true;
   if (mod1 = nil)or (mod2 = nil) then 
      begin
        samemod := false;
        goto 999;
      end;
   for i := 1 to max do 
     if mod1^.amodel[i] <> mod2^.amodel[i] then samemod := false;
999: end;

procedure makegraph(var ugraph : endarray; firsttime,lastone: boolean; var newadj: adjtype);
var i,j: integer; var goahead,ok:  boolean; curpos: modeltypeptr;
    gfile: text; boo: boolean; mybuf: inbuf;
begin
   for i:=1 to max do
      for j := 1 to max do   
         newadj^[i,j] := adjacency^[i,j];
   for i:=1 to max do
      for j := 1 to max do
         if (i <> j) and (((in2 in ugraph^[i,j]) or (knownin2 in ugraph^[i,j])) and
            (adjacency^[j,i] = notthere)) then
                 adjacency^[i,j] := 3;
   i := 1;
   j := 1;
   ok := true;
   formgraph(ugraph,adjacency,i,j,ok,4);
   goahead := not samemod(oldmod,zmodellist);
   zreadadjtomod(newadj,oldmod);   
(*   writelist(output,zmodellist); *)
   curpos := zmodellist;
   mybuf := '';
   if (goahead or firsttime) then newsearch(adjacency)
   else
      begin
         twrite(ofilvar,'Significance Level = |');
         twritereal(ofilvar,signif);
         twriteeoln(ofilvar);
         twriteln(ofilvar,'Same as model at previous significance level. |');
      end;
   for i:=1 to max do
      for j := 1 to max do   
         adjacency^[i,j] := newadj^[i,j];
end;

procedure removetheedges(var ugraph: graphrec; var edgelist,edgelist1: nodeptr; var corrlists: corrarray;
                      var internal: boolean);
label 999;
var leaveloop: boolean; i: vertextype; curnode1,temp: nodeptr; partial: partialpair; partialedon: partialedonarray;
    theseparents1,theseparents2: vertexset;
begin
   leaveloop := false;
   i := 0;
   repeat
      if aaabort then goto 999;
      curnode1 := edgelist;
      i := i + 1;
      if (depth <> -1) and (i > depth) then leaveloop := true;
      while (curnode1 <> nil) and (not toomuchtime) do
        begin
          if aaabort then goto 999;
          partial[1] := edgetovert^[curnode1^.edge].head;
          partial[2] := edgetovert^[curnode1^.edge].tail;
          theseparents1 := parents[partial[1]] - [partial[1],partial[2]];
          if (setsize(theseparents1) >= i) and
             (not (partial[2] in temporder[partial[1]])) then 
               checkpartials(curnode1,i,0,corrlists[partial[1]],partialedon,ugraph.augraph,internal,corrlists);
          theseparents2 := parents[partial[2]] - [partial[1],partial[2]];
          if (setsize(theseparents2) >= i) and
             (not (partial[1] in temporder[partial[2]])) then 
               checkpartials(curnode1,i,0,corrlists[partial[2]],partialedon,ugraph.augraph,internal,corrlists);
          if (ugraph.augraph^[partial[1],partial[2]] = []) or
             ((setsize(theseparents1) <= i) and (setsize(theseparents2) <= i)) then
             begin
                temp := curnode1;
                curnode1 := curnode1^.fwd;
                moveedge(temp,edgelist,edgelist1);
             end
          else curnode1 := curnode1^.fwd;
       end;
    until (edgelist = nil) or toomuchtime or leaveloop;
    disposeedgelist(edgelist);
999: end;

procedure getpartialssuff(internal,firsttime,lastone,combined,lookforlatent: boolean; var ugraph: graphrec;
                          crossvalidate: boolean);
label 999;
var i,j: vertextype; curpos: edgeptr; 
    k,num: integer; aprob,g2: real; c: char;
    held,nochange,consis, leaveloop,ok: boolean; 
    edgelist,edgelist1: nodeptr; newadj: adjtype;
begin
   zmodellist := nil;
   zmodelend := nil;

(*******************************************************)
   tnewadj(newadj,fatal);
   tnewadj(collidera,fatal);
   tnewadj(tempadj,fatal);
(*******************************************************)


   if fatal in errors then goto 999;
   for i := 1 to numob do
      begin
         corrlists[i] := nil;
         corrlists1[i] := nil;
      end;
   ugraph.nocollidelist := nil;
   numchecked := 1;
   edgelist := nil;
   edgelist1 := nil;
   num := 1;
   toomuchtime := false;
   start := tgettime;
   numedges := tround(setsize(graphvertices) * (setsize(graphvertices) - 1) / 2);
   if discrete and (not internal) then
      if sizeof(cellsarr) < heapavail then new(cells)
      else 
         begin
            writeln(heapavail);
            errormess('Out of memory.|',fatal,false);
            goto 999;
         end;
   initugraph(adjacency,ugraph);
   orientfromknow(ugraph.augraph); 
   initparents(edgelist,internal,ugraph.augraph);
   removetheedges(ugraph,edgelist,edgelist1,corrlists,internal);
   restoreedges(ugraph,edgelist1);
   disposeedgelist(edgelist);
   removetheedges(ugraph,edgelist1,edgelist,corrlists,internal);
   disposeedgelist(edgelist);
   disposeedgelist(edgelist1);
   for i := 1 to max do
      begin
         disposeedgelist(corrlists[i]);
         disposeedgelist(corrlists1[i]);
     end;
    if toomuchtime then 
	 twriteln(ofilvar,'Search aborted because time limits exceeded.|');
    ok := true;
    setcollider(true,nochange,ok,ugraph,1);
(*    calcdpath(ugraph.augraph,ugraph.ancest); *)
(*    repeat
        setcollider(true,nochange,ugraph.ancest,ugraph.augraph,ugraph.nocollidelist,1);
        if nochange then orientawaycycle(nochange,adj);
    until nochange; *)
999:if not internal then
    begin
      twriteeoln(ofilvar);
      if aaabort then twriteln(ofilvar,'Search aborted by interrupt.|');
      if (not consis) then 
      twriteln(ofilvar,'  #: no orientation consistent with assumptions|');
      twriteeoln(ofilvar);
    end;
    if discrete and (not internal) then dispose(cells);
    if lookforlatent then 
              begin
                 nonewritten := true;
                 findlatents;
              end;
    twriteeoln(ofilvar);
    if not combined then 
      begin
         twrite(ofilvar,'Significance Level = |');
         twritereal(ofilvar,signif);
         twriteeoln(ofilvar);
         printcausalsuff(consis,ofilvar,ugraph.augraph);
      end
    else makegraph(ugraph.augraph,firsttime,lastone,newadj);
    if crossvalidate then insertthepag(ugraph,thebest);
    dispose(tempadj);
    dispose(collidera);
end;



procedure chi;
type bin = 0..1;
     binarray = array[bin,bin,bin,bin] of real;
     indarray = array[1..4,1..4,bin,bin] of real;
     margtype = array[1..4,bin] of real;
     difftype = array[1..4,1..4] of real;

var i,j,k,l: integer; g2,expected,c2,observed: real; file1name,file2name: inbuf;
   file1,file2: text; changed: boolean; thearr1,thearr2: binarray;
   index1,index2,index3,index4: bin; aprob: real; expectedpair,joint: indarray;
   marginal: margtype; diffc2,diffg2: difftype;

   procedure init;
   var i,j: integer; k,l: bin;
   begin
      for i := 1 to 4 do 
         for j := 1 to 4 do
           begin
             for k := 0 to 1 do
               for l := 0 to 1 do
                  begin
                     marginal[i,k] := 0;
                     expectedpair[i,j,k,l] := 0;
                     joint[i,j,k,l] := 0;
                  end;
             diffg2[i,j] := 0;
             diffc2[i,j] := 0;
          end;
   end;

   procedure fillinjoint(var1,var2: integer);
   var i,j,k,l: bin; index1,index2: integer;
   begin
     for i := 0 to 1 do
       for j := 0 to 1 do
         for k := 0 to 1 do
           for l := 0 to 1 do
             begin
               if var1 = 1 then index1 := i
               else if var1 = 2 then index1 := j
               else if var1 = 3 then index1 := k
               else index1 := l;
               if var2 = 1 then index2 := i
               else if var2 = 2 then index2 := j
               else if var2 = 3 then index2 := k
               else index2 := l;
               joint[var1,var2,index1,index2] := 
                   joint[var1,var2,index1,index2] + thearr1[i,j,k,l];
             end;
    end;

    procedure fillinexpect(var1,var2: integer);
    var i,j: bin;
    begin
      for i := 0 to 1 do
         for j := 0 to 1 do
            expectedpair[var1,var2,i,j] := marginal[var1,i] * marginal[var2,j]/
               sampsize;
    end;
   
   procedure fillinmarginal;
   var i,j,k,l: bin; 
   begin
     for i := 0 to 1 do
       for j := 0 to 1 do
         for k := 0 to 1 do
           for l := 0 to 1 do
             begin
               marginal[1,i] := marginal[1,i] + thearr1[i,j,k,l];
               marginal[2,j] := marginal[2,j] + thearr1[i,j,k,l];
               marginal[3,k] := marginal[3,k] + thearr1[i,j,k,l];
               marginal[4,l] := marginal[4,l] + thearr1[i,j,k,l];
             end;
     end;

       
    procedure getdiff(var1,var2: integer);
    var i,j: bin;
    begin
       for i := 0 to 1 do
         for j := 0 to 1 do
           begin
            diffg2[var1,var2] := joint[var1,var2,i,j] * 
               ln(joint[var1,var2,i,j]/expectedpair[var1,var2,i,j]);
            diffc2[var1,var2] := 
               diffc2[var1,var2] + 
                  sqr(joint[var1,var2,i,j] - expectedpair[var1,var2,i,j])/
                  expectedpair[var1,var2,i,j];
           end;
       end;

     procedure finddiff;
     var i,j: integer; k,l: bin;
     begin
       init;
       fillinmarginal;
       for i := 1 to 3 do
         for j := i + 1 to 4 do 
           begin
              fillinjoint(i,j);
              fillinexpect(i,j);
              getdiff(i,j);
           end;
        end;

                  

begin
  file1name := nullfilename;
  getfile(file1,'file 1: ',know,file1name,changed);
  i := 0;
  repeat
     i := i + 1;
     read(file1,index1);
     read(file1,index2);
     read(file1,index3);
     read(file1,index4);
     readln(file1,aprob);
     thearr1[index1,index2,index3,index4] := aprob * sampsize;
  until i = 16;
  file2name := nullfilename;
  getfile(file2,'file 2: ',know,file2name,changed); 
  i := 0;
  repeat
     i := i + 1;
     read(file2,index1);
     read(file2,index3);
     read(file2,index2);
     readln(file2,index4);
     readln(file2,aprob);
     thearr2[index1,index2,index3,index4] := aprob;
  until i = 16;
  i := 0;
  g2 := 0;
  c2 := 0;
  for i := 0 to 1 do
   for j := 0 to 1 do
    for k := 0 to 1 do
      for l := 0 to 1 do
        begin
           expected := thearr1[i,j,k,l];
           observed := thearr2[i,j,k,l];
           c2 := c2 + sqr(observed - expected)/expected;
           g2 := g2 + (observed * ln(observed/expected)); 
        end;
     finddiff;
end;

procedure disposenodes(var node: nodeptr);
var curpos,nextpos: nodeptr;
begin
  curpos := node;
  if curpos <> nil then
    repeat
      nextpos := curpos^.fwd;
      dispose(curpos);
      curpos := nextpos;
    until curpos = nil;
end;


procedure getpartials1(internal,firsttime,lastone,crossvalidate: boolean);
label 999;
var madedsep, madeugraph, madezeroes, combined,lookforlatent: boolean; i: vertextype; ugraph: graphrec;
begin
   lookforlatent := false;
   if isacycle then 
      begin
        myerror('Build cannot be used on cyclic graph.|',fatal,false);
        goto 999;
      end;
   nonewritten := true;
   madedsep := false;
   madeugraph := false;
   madezeroes := false;
   tnewdsep(dsep,fatal);
   if (not (fatal in errors)) then madedsep := true;
   initrec(ugraph);
   if (not (fatal in errors)) then madeugraph := true;
   tnewdsep(zeroes,fatal);
   if (not (fatal in errors)) then madezeroes := true;
   if (not (fatal in errors)) then
      begin
         debug := false;
         if debug then chi
           else
              begin
               if firsttime then
                   causalsuff := not getreply('Assume latent common causes? |','y',false);
               if causalsuff then
                   begin
                      if firsttime then 
                        begin
                         combined := getreply('Combined search? |','y',false);
                         if (setsize(graphvertices) <= 15) and (not discrete) and (not indepdata) then 
                            lookforlatent := getreply('Test the assumption of no latent variables? |','y',false);
                        end;
                      getpartialssuff(internal,firsttime,lastone,combined,lookforlatent,ugraph,crossvalidate);
                   end
               else 
                  begin
                    if firsttime then
                      begin
                        heuristic := not getreply('Use the exact algorithm? |','y',true);
                        if heuristic then 
                          twriteln(ofilvar,'Using heuristic algorithm.|')
                        else twriteln(ofilvar,'Using exact algorithm.|');
                      end;
                    getpartialnonsuff(ugraph,crossvalidate); 
                  end;
             end;
       end;
     dispose(dsep);
     if ugraph.augraph <> nil then dispose(ugraph.augraph); 
     dispose(ugraph.ancest);
     disposetri(ugraph.nocollidelist);
     disposetri(ugraph.bothlist);
     dispose(zeroes);
     for i := 1 to max do disposenodes(corrlists[i]); 
999: end;

procedure generatesubsample(subsample: integer; var samparray: sampletypearray);
var i,randnum: integer; 
begin
   initrand(newseed);
   for i := 1 to round(sampsize) do
       if samparray[i] = train then samparray[i] := trainest;
   i := 0;
   repeat
       randnum := round(thisrand * (sampsize-1)) + 1;
       if samparray[randnum] = trainest then
          begin
            samparray[randnum] := train;
            i := i + 1;
          end;
   until i = subsample;
end;            

procedure writecombined(var thebest: ugrapharray; percentcut: real);
type recrange = 1..maxbest; 
     adjarec =  record
                  rarrow,rline,rcircle,larrow,lline,lcircle: integer;
               end;
     combinedarray = array[1..maxvertex,1..maxvertex] of adjarec;
var combined: combinedarray; i,j,k,numsamp: integer; wroteone: boolean;
begin
  for j := 1 to max-1 do
    for k := j+1 to max do 
       begin
          combined[j,k].rarrow := 0;
          combined[j,k].rline := 0;
          combined[j,k].rcircle := 0;
          combined[k,j].larrow := 0;
          combined[k,j].lline := 0;
          combined[k,j].lcircle := 0;
       end;
  numsamp := 0;
  for i := 1 to maxbest do if thebest[i].augraph <> nil then numsamp := numsamp + 1;
  for j := 1 to max-1 do
    for k := j+1 to max do
       for i := 1 to maxbest do
          if thebest[i].augraph <> nil then
                begin
                   if (in2 in thebest[i].augraph^[j,k]) then combined[j,k].rarrow := combined[j,k].rarrow + 1;       
                   if (out2 in thebest[i].augraph^[j,k]) then combined[j,k].rline := combined[j,k].rline + 1;        
                   if (notknown in thebest[i].augraph^[j,k]) then combined[j,k].rcircle := combined[j,k].rcircle + 1;
                   if (in2 in thebest[i].augraph^[k,j]) then combined[k,j].larrow := combined[k,j].larrow + 1;      
                   if (out2 in thebest[i].augraph^[k,j]) then combined[k,j].lline := combined[k,j].lline + 1;      
                   if (notknown in thebest[i].augraph^[k,j]) then combined[k,j].lcircle := combined[k,j].lcircle + 1;
                end;
  twriteln(ofilvar,'Var 1    Var 2    o    <    -    -    >    o|');
  for j := 1 to max-1 do
    for k := j+1 to max do 
       if (combined[j,k].rarrow > 0) or (combined[j,k].rline > 0) or (combined[j,k].rcircle > 0) or 
          (combined[k,j].larrow > 0) or (combined[k,j].lline > 0) or (combined[k,j].lcircle > 0) then
       begin
          pwritename(j,ofilvar,5);
          twrite(ofilvar,'    |');
          pwritename(k,ofilvar,5);
          twrite(ofilvar,'    |');
          tpwriteint(ofilvar,combined[j,k].rarrow,5,false);
          tpwriteint(ofilvar,combined[j,k].rline,5,false);
          tpwriteint(ofilvar,combined[j,k].rcircle,5,false);
          tpwriteint(ofilvar,combined[k,j].larrow,5,false);
          tpwriteint(ofilvar,combined[k,j].lline,5,false);
          tpwriteint(ofilvar,combined[k,j].lcircle,5,false);
          twriteeoln(ofilvar);
       end;
  twriteeoln(ofilvar);
  twriteln(ofilvar,'/Indep |');
  for j := 1 to max-1 do
    for k := j+1 to max do 
       if ((combined[j,k].rarrow + combined[j,k].rline + combined[j,k].rcircle) >= percent * numsamp) then
       begin
          wroteone := false;
          pwritename(k,ofilvar,5);
          twrite(ofilvar,'    |');
          if (combined[j,k].rarrow >= percentcut * numsamp) then begin twrite(ofilvar,'<|'); wroteone := true; end;
          if (combined[j,k].rline >= percentcut * numsamp) then begin twrite(ofilvar,'-|'); wroteone := true; end;
          if (not wroteone) then twrite(ofilvar,'o|');
          wroteone := false;
          twrite(ofilvar,'-|');
          if (combined[k,j].larrow >= percentcut * numsamp) then begin twrite(ofilvar,'>|'); wroteone := true; end;
          if (combined[k,j].lline >= percentcut * numsamp) then  begin twrite(ofilvar,'-|'); wroteone := true; end;
          if (not wroteone) then  twrite(ofilvar,'o|'); 
          twrite(ofilvar,'    |');
          pwritename(j,ofilvar,5);
          twriteeoln(ofilvar);
       end;
       twriteln(ofilvar,'!|');
end;

procedure selectsubset;
var i,j: integer; classcat: catrange;
begin
    effsampsize := 0;
    if havereadraw then
       for i := 1 to round(sampsize) do
         begin
          for j := 1 to max do
             if (j in graphvertices) then
                 begin
                    classcat := getbigdata(i,j);
                    if classcat = 0 then samparray[i] := test;
                 end;
          if (samparray[i] <> test) then effsampsize := effsampsize + 1;
         end;
end;

procedure getpartials;
var i: integer; sig: array[1..6] of real; curmod: modeltype; percentcut: real;
    numbers: numberset; maxnum,minnum: integer; autosig,crossvalidate: boolean; subsample,numsample: real;
begin
    for i := 1 to maxbest do 
       begin
          thebest[i].augraph := nil;
          thebest[i].ancest := nil;
          thebest[i].nocollidelist := nil;
       end;
    autosig := getreply('Automatic choice of significance levels? |','y',false);    
    new(cells);
    new(oldmod);
    for i := 1 to max do oldmod^.amodel[i] := [];
    sig[1] := 0.001; sig[2] := 0.005; sig[3] := 0.01; sig[4] := 0.05; sig[5] := 0.1; sig[6] := 0.2;
    if autosig then
      for i := 1 to 6 do 
          begin
            signif := sig[i];
            getpartials1(autosig,i=1,i=6,false);
          end
    else
        begin
             crossvalidate := getreply('Cross validate? |','y',false);    
             if crossvalidate then 
                  begin
	              subsample := getnumber(1,round(sampsize),false,
                      'Subsample |                    ',subsample);
	              numsample := getnumber(1,maxbest,false,
                      'Number of samples |                    ',numsample);
                      for i := 1 to round(numsample) do
                        begin
                           getpartials1(autosig,i=1,i=round(numsample),true);
		           generatesubsample(round(subsample),samparray);
                        end;
                      percentcut := getnumber(0,1,false,
                      'Percent cutoff|                    ',percentcut);
                      writecombined(thebest,percentcut);
                  end
             else getpartials1(autosig,true,true,true);
        end;
    dispose(oldmod);
    for i := 1 to maxbest do 
        begin
            if thebest[i].augraph <> nil then dispose(thebest[i].augraph);
            if thebest[i].nocollidelist <> nil then disposetri(thebest[i].nocollidelist);
            if thebest[i].ancest <> nil then dispose(thebest[i].ancest);
            thebest[i].augraph := nil;
            thebest[i].ancest := nil;
            thebest[i].nocollidelist := nil;
        end;
end;

procedure getregressee(var regressee: vertextype);
var name: vertexnametype; wasinserted: boolean;
begin
    if (currentin = inp) then twrite(output,'Name of regressee: |');
    if (currentin = inp) then readttyline(buf) else readfileline(knowfile,buf);
    repeat
       errors := [];
       readname(name,ignoreall);
       regressee := listposition(name,false,vernames,wasinserted);
    until errors = [];
end;

procedure getregressors;
var leaveloop,wasinserted,firsttime: boolean; name: vertexnametype; k: vertextype; 
   begin
     if (currentin = inp) then twrite(output,'List of regressors: |');
     if (currentin = inp) then readttyline(buf) else readfileline(knowfile,buf);
     initinput([graph,raw,cellcounts,covariance,independence],false);
     firsttime := true;
     leaveloop := false;
     while (not myeoln) and (not leaveloop) do 
      while (not leaveloop) do 
       begin
         readwhitespace;
         if (not myeoln) then
          begin
           if alphabetic(buf[curpos]) then
            begin
             firsttime := false;
             readwhitespace;
             if (not myeoln) then
	       begin
                 readname(name,warning);
                 if name[1] in ['A'..'Z'] then
   	           myerror('No theoreticals allowed.|  ',fatal,true);
                 k := listposition(name,false,vernames,wasinserted);
                 graphvertices := graphvertices + [k];
	       end; (* not myeoln *)
             if (myeoln) then
              if (currentin = inp) then 
                begin
                   twrite(output,'List of regressors: |');
                   readttyline(buf)
                 end
              else readfileline(knowfile,buf);
            end (* if alphabetic *)
          end (* not myeoln *)
         else 
              begin
               leaveloop := true;
               if firsttime then errormess('List of variables required here. |',fatal,true);
              end
        end;
       calcnum(numob,numtheo,max,vernames);
end;

procedure orienttime(regressee: vertextype);
var i: vertextype;
begin
   temptiers[1] := [];
   temptiers[2] := [regressee];  
   for i := 1 to max do
     if (i in graphvertices) and (i <> regressee) then 
          temptiers[1] := temptiers[1] + [i];    
   calcorder;
end;

procedure forbidedges(regressee: vertextype);
var i,j: vertextype;
begin
  for i := 1 to max do
    if (i in graphvertices) and (i <> regressee) then
      for j := 1 to max do
         if (j in graphvertices) then
           if (j <> regressee) then
             begin
               forbiddenadj^[i,j] := 1;
               forbiddenadj^[j,i] := 1;
             end
           else forbiddenadj^[i,j] := notthere;
end;

procedure doregress;
label 999;
var regressee: vertextype;
begin
   getregressee(regressee);
   getregressors;
   graphvertices := graphvertices + [regressee];
   orienttime(regressee);
   forbidedges(regressee);
   getpartials;
999: end;

procedure getsignificances(var lowersig,uppersig,increment: real);
begin
   lowersig := getnumber(0.0,1.0,false,'Lower significance level: |',0.01);
   uppersig := getnumber(lowersig,1.0,false,'Upper significance level: |',0.13);
   increment := getnumber(0.01,1.0,false,'Increment: |',0.02);
end;

procedure getesttypes(var estimates: estset);
var numset: numberset; i: vertextype;
begin
   estimates := [];
   twriteln(output,'1: Maximum likelihood |');
   twriteln(output,'2: Uniform |');
   twriteln(output,'3: Noisy Or |');
   twrite(output,'Type in numbers of estimate types to run: |');
   getnumbers(numset,3);
   twriteeoln(output);
   for i := 1 to 3 do
      if i in numset then
         if i = 1 then estimates := estimates + [MLinternal]
         else if i = 2 then estimates := estimates + [uniform]
         else if i = 3 then estimates := estimates + [noisyormeth];
end;

procedure getprecisions(var lowerpre,upperpre,incr: integer);
begin
   lowerpre := round(getnumber(0,100,true,'Lower precision: |',1));
   upperpre := round(getnumber(lowerpre,100,true,'Upper significance level: |',21));
   incr := round(getnumber(1,100,true,'Increment: |',5));
end;

procedure initmod(*var curmodel: modeltypeptr *);
begin
   curmodel^.acc := 0.0; curmodel^.sens := 0.0; curmodel^.spec := 0.0;
   curmodel^.pospred := 0.0; curmodel^.negpred := 0.0; curmodel^.acccut := 0.0;
   curmodel^.senscut := 0.0; curmodel^.speccut := 0.0;
   curmodel^.pospredcut := 0.0; curmodel^.negpredcut := 0.0; 
   curmodel^.mostposcut := 0; curmodel^.mostnegcut := 0; 
   curmodel^.mostpos := 0; curmodel^.mostneg := 0;
end;

procedure fillnoisyor(classvar: vertextype; var adjacency: adjtype; 
                      precision: integer; numratio: real; poscat: vertextype);
var coltotal,i,nonnormal,j,numnonnormal: integer; normal: parray; 
    nonnormcat,classcat: catrange; testnum: integer;
begin
  testnum := 0;
  for i := 1 to max do
   if  i in graphvertices then
    begin
       normal[i] := 0;
       for j := 1 to maxval[i] do
           if varcatvalue[i,j] = 0 then normal[i] := j;
       noisyor[i,1].pos := 0; noisyor[i,1].total := 0;
(*       for j := 2 to cats[i].num do 
           begin
             noisyor[i,j].pos := numratio * precision;
             noisyor[i,j].total := precision;
           end; *)
       noisyor[i,2].pos := precision /2 ; noisyor[i,2].total := precision;
    end;
  for i := 1 to tround(sampsize) do
   if (samparray[i] = trainest) or (samparray[i] = est) then
    begin
      numnonnormal := 0;
      j := 1;
      while (j <= max) and (numnonnormal < 2) do
        begin
           if adjacency^[j,classvar] <> notthere then
             begin
              classcat := getbigdata(i,j);
              if classcat <> normal[j] then
                 begin
 	            numnonnormal := numnonnormal +1;
                    nonnormal := j;
                    nonnormcat := classcat;
                 end;
              end;
           j := j + 1;
         end;
      if (numnonnormal = 1) then
         begin
             if nonnormal = 50 then
                 begin
                   testnum := testnum + 1;
                   writeln('i = ',i);
                 end;
(* assume that classvar = 2 is the value being turned on by noisy or *)
            if (getbigdata(i,classvar) = 2) then
               noisyor[nonnormal,nonnormcat].pos := noisyor[nonnormal,nonnormcat].pos + 1;
            noisyor[nonnormal,nonnormcat].total := noisyor[nonnormal,nonnormcat].total + 1;               
         end;
    end;
  writeln(testnum);
  for i := 1 to max do
    if adjacency^[i,classvar] <> notthere then
          for j := 1 to maxval[i] do
             if noisyor[i,j].total <> 0 then
             noisyor[i,j].pos := noisyor[i,j].pos / noisyor[i,j].total;
(*   makenoisyor(classvar,cats[classvar].condprob,cats,1,noisyor); *)
end;

procedure maximizeest(var adjacency: adjtype; var score: real; 
                      var bestmodel,curmodel: modeltypeptr; maxtype: infotype;
                      estimates: estset; lowerpre,upperpre,incr: integer;
                      poscat: vertextype; var numratio: real;
                      pospercent,percentpop: real);
var estimate: esttype; k,l: integer; numtot,numpos: integer; 
begin
  if uniform in estimates then
    if numratio = -1 then
       begin
         numpos := 0;
         numtot := 0;
         for l := 1 to tround(sampsize) do
            if (samparray[l]=trainest) or (samparray[l]=est) then
               begin
                 numtot := numtot + 1;
                 if getbigdata(l,classvar) = poscat then numpos := numpos + 1;
                end;
         if numtot = 0 then numratio := 0
         else numratio := numpos/numtot;
       end;
  for estimate := MLinternal to noisyormeth do
     if estimate in estimates then
        begin
           k := lowerpre;
           repeat
              if estimate <> noisyormeth then
                doestimate(adjacency,output,true,estimate,k)
              else fillnoisyor(classvar,adjacency,k,numratio,poscat);
              initmod(curmodel);
              curmodel^.estimate := estimate;
              curmodel^.precision := k;
              doautocrossvalidate(poscat,curmodel,pospercent,percentpop,estimate,maxtype);
              comparemodel(bestmodel,curmodel,maxtype,true);
              disposebayesnet(cats);
              k := k + incr;
           until (k > upperpre) or (estimate = MLinternal) or (estimate = MLexternal);
         end;
end;

procedure gettypetomaximize(* var maxtype: infotype; var pospercent,percentpop: real *);
var num: integer; realnum: real;
begin
   twriteln(output,'1: Accuracy |');
   twriteln(output,'2: Sensitivity |');
   twriteln(output,'3: Specificity |');
   twriteln(output,'4: Positive Predictive Value |');
   twriteln(output,'5: Negative Predictive Value |');
   twriteln(output,'6: Most Positive |');
   twriteln(output,'7: Biggest Population |');
   twrite(output,'Type in number corresponding to type of score to be maximized: |');
   readttyline(buf);
   realnum := readnum(true,fatal);
   num := tround(realnum);
   twriteeoln(output);
   if num = 1 then maxtype := accuracy
   else if num = 2 then maxtype := sensitivity
   else if num = 3 then maxtype := specificity
   else if num = 4 then maxtype := positivepredvalue
   else if num = 5 then maxtype := negativepredvalue
   else if num = 6 then 
        begin
            maxtype := mostpos;
            twrite(output,'Percentage: |');
            readttyline(buf);
            pospercent := readnum(true,fatal);            
        end
   else if num = 7 then
        begin
            maxtype := mostneg;
            twrite(output,'Percentage: |');
            readttyline(buf);
            percentpop := readnum(true,fatal);            
        end;
end;

procedure findbestreg(* var ugraph: endarray *);
var regressee: vertextype; lowersig,uppersig,increment,k,score,pospercent,
   percentpop,numratio: real; i,j: vertextype; 
   lowerpre,upperpre,incr,poscat,numpos: integer;
   maxtype: infotype; estimates: estset; 
   oldgraphvert,newgraphvert,lastvert: vertexset;
   curmodel,bestmodel: modeltypeptr;
label 999; 
begin
   if sizeof(cellsarr) < heapavail then new(cells)
   else 
      begin
         writeln(heapavail);
         errormess('Out of memory.|',fatal,false);
         goto 999;
       end;
   dosplit;
   getregressee(regressee);
   classvar := regressee;
   getregressors;
   getsignificances(lowersig,uppersig,increment);
   getesttypes(estimates);
   getprecisions(lowerpre,upperpre,incr);
   getposcat(regressee,poscat);
   gettypetomaximize(maxtype,pospercent,percentpop);
   graphvertices := graphvertices + [regressee];
   oldgraphvert := graphvertices;
   orienttime(regressee);
   forbidedges(regressee);
   k := lowersig;
   new(curmodel);
   new(bestmodel);
   initmod(bestmodel);
   numratio := -1;
   lastvert := [];
   repeat
      signif := k;
      getpartials;
      for i := 1 to max do curmodel^.amodel[i] := [];
      curmodel^.signif := k;
      newgraphvert := [regressee];
      for i := 1 to max do 
        if i in graphvertices then
         for j := 1 to max do
           if j in graphvertices then
            if (knownin2 in ugraph^[i,j]) and (i <> j) then
              begin
                adjacency^[i,j] := 0;
                curmodel^.amodel[j] := curmodel^.amodel[j] + [i];
                newgraphvert := newgraphvert + [i];
              end
            else adjacency^[i,j] := notthere;
      if lastvert <> newgraphvert then
         begin
           graphvertices := newgraphvert;
           maximizeest(adjacency,score,bestmodel,curmodel,maxtype,estimates,
                  lowerpre,upperpre,incr,poscat,numratio,
                  pospercent,percentpop);
         end;
      lastvert := newgraphvert;
      k := k + increment;
      graphvertices := oldgraphvert;
  until k >= uppersig;
  printstats(ofilvar,bestmodel);
  dispose(bestmodel); 
  dispose(curmodel);
  dispose(cells);
999: end;

procedure writevertexset(aset: vertexset);
var i: vertextype;
begin
  for i := 1 to max do
     if (i in graphvertices) and (i in aset) then
        write(i);
  writeln;
end;

procedure zwriteadj(var graph: adjtype;var outfile: text);
var k,l: integer;
begin
   for k := 1 to max do
      for l := 1 to max do
        if (k in graphvertices) and (l in graphvertices) and
           (graph^[k,l] <> notthere) then
		begin
		writename(k,outfile); twrite(outfile,' |');
		writename(l,outfile); twrite(outfile,' |');
		tpwritereal(outfile,graph^[k,l],8,3,false);
		twriteeoln(outfile);
		end;
   twriteeoln(outfile);
   twriteeoln(outfile);
end;

procedure writedseparray(var ugraph: endarray);
var i,j: vertextype;
begin
   for i := 1 to max do
     for j := 1 to max do
       if (i in graphvertices) and (j in graphvertices) and
          (ugraph^[i,j] = []) then
              begin
                 writeln(i,' ',j,' ');
                 writevertexset(dsep^[i,j]);
              end;
end;

            

procedure partialcheck;
var partial: partialpair; partialedon: partialedonarray; curname: vertexnametype;
    tempvar,i: vertextype; wasinserted,emptyline: boolean;
    condset: vertexset; aval,aprob,g2: real; held,ok: boolean; numchecked: integer;
begin
   new(cells);
   write(output,'First Variable: ');
   readttyline(buf);
   readwhitespace;
   readname(curname, ignoreall);
   partial[1] := listposition(curname,false,vernames,wasinserted);
   write(output,'Second Variable: ');
   readttyline(buf);
   readwhitespace;
   readname(curname, ignoreall);
   partial[2] := listposition(curname,false,vernames,wasinserted);
   emptyline := false;
   condset := [];
   repeat
      write(output,'Conditioning Variable: ');
      readttyline(buf);
      readwhitespace;
      if myeoln then emptyline := true;
      if not emptyline then
        begin
          readname(curname, ignoreall);
          tempvar := listposition(curname,false,vernames,wasinserted);
          condset := condset + [tempvar];
        end;
   until emptyline = true;
   partialedon.length := 0;
   for i := 1 to max do
     if i in condset then
       begin
         partialedon.length := partialedon.length + 1;
         partialedon.arr[partialedon.length] := i;
     end;
   if (not discrete) then
      begin
         aval := calcpart1(partial,partialedon,ok);
         calcpartialheld(aval,aprob,held,partialedon.length);
      end
   else held := calcindep1(partial,aprob,g2,partialedon,numchecked);
   if (not discrete) then writeln(output,'partial = ',aval:7:3,' prob = ',aprob:7:3)
   else writeln(output,'g2 = ',g2:7:3,' prob = ',aprob:7:3);
   dispose(cells);
end;
(*++++++++++++++++++++++++++++++++++++++++++++++++++++++*)

procedure writebest;
var num,i: vertextype;
begin
  write(output,'Number: ');
  readln(input,num);
  writeln(output,'Best in');
  for i := 1 to max do
     begin
       if i in bestin[num] then writename(i,output);
       writeln(output);
     end;
  writeln(output,'Best out');
  for i := 1 to max do
     begin
       if i in bestout[num] then writename(i,output);
       writeln(output);
     end;
end;


procedure writeset(v: vertexset);
var i: vertextype; first: boolean;
begin
   first := true;
   write('[');
   for i := 1 to max do 
    if i in v then
       begin
         if (not first) then write(',');
         first := false;
         writename(i,output);
       end;
   write(']');
end;

procedure writetempset1;
var i: vertextype;
begin
   for i := 1 to max do
      if t[i] <> [] then
        begin
          writeln(output,'i = ',i:4,' ');
          writeset(t[i]);
        end;
end;

procedure printdsep;
var i,j: vertextype;
begin
   for i := 1 to max do
     for j := 1 to max do
        if (i <> j) then
          begin
            write('i = ',i,' j = ',j,' dsep = ');
            writeset(dsep^[i,j]);
            writeset(dsep^[j,i]);
            writeln;
          end;
end;

procedure calcbicscore;
var mag: graphrec; score: real; isok: boolean; aux: auxiliary;
begin
 tnewadj(aux.parent,fatal);
 tnewadj(aux.doublehead,fatal);
 tnewadj(aux.tempadj,fatal);
 tnewcov(aux.localparm,fatal);
 tnewcov(aux.localcov,fatal);
 tnewcov(aux.localestcov,fatal);
 tnewcov(aux.estcov,fatal); 
 tnewcov(aux.tempcov,fatal);
 tnewcov(aux.theinverse,fatal);
 initrec(forbidpag);
 initrec(temppag);
 initrec(aux.apag); 
 initrec(mag);
 getknowledge;
 copypag(temppag,mag,true);
 mag.curchange := 1;
 isok := true;
 pagtomag(temppag,mag,isok,3);
 printpoipg(mag);
 scoremag(mag,score,aux);
 writeln('score = ',score:7:3);
 dispose(aux.apag.ancest);
 dispose(aux.apag.augraph);
 dispose(mag.ancest);
 dispose(mag.augraph);
 dispose(forbidpag.ancest);
 dispose(forbidpag.augraph);
 dispose(aux.parent);
 dispose(aux.doublehead);
 dispose(aux.tempadj);
 dispose(aux.localparm);
 dispose(aux.localcov);
 dispose(aux.localestcov);
 dispose(aux.estcov); 
 dispose(aux.tempcov);
 dispose(aux.theinverse);
end;

procedure getindep;
var partial: partialpair; partialedon: partialedonarray; aprob,g2: real; numchecked: integer;
    held: boolean; partialedset: vertexset; wasinserted,leaveloop: boolean; i,k: integer;
    name: vertexnametype;
begin
     write(output,'Variables: ');
     readttyline(buf);
     i := 1;
     leaveloop := false;
     while (i < 3) and (not leaveloop) do
       begin
         readwhitespace;
         if (not myeoln) then
           begin
             readwhitespace;
             if (not myeoln) then
	       begin
                 readname(name,fatal);
                 if name[1] in ['A'..'Z'] then
   	           myerror('No theoreticals allowed in selection.|  ',
                           fatal,true)
                 else i := i + 1;
                 k := listposition(name,true,vernames,wasinserted);
                 partial[i-1] := k;
	       end;
           end
          else leaveloop := true;
        end;
     readvarlist(partialedset);
     partialedon.length := 0;
     for i := 1 to max do 
       if i in partialedset then
          begin
             partialedon.length := partialedon.length + 1;
             partialedon.arr[partialedon.length] := i;
          end;
     held := calcindep1(partial,aprob,g2,partialedon,numchecked);
     writeln('Prob = ',aprob:6:2,' g2 = ',g2:6:2);
end;

procedure getindep1;
var partial: partialpair; partialedon: partialedonarray; aprob,g2: real; numchecked: integer;
    held: boolean; partialedset: vertexset; wasinserted,leaveloop: boolean; i,k: integer;
    name: vertexnametype;
begin
     write(output,'Variables: ');
     readttyline(buf);
     i := 1;
     leaveloop := false;
     while (i < 3) and (not leaveloop) do
       begin
         readwhitespace;
         if (not myeoln) then
           begin
             readwhitespace;
             if (not myeoln) then
	       begin
                 readname(name,fatal);
                 if name[1] in ['A'..'Z'] then
   	           myerror('No theoreticals allowed in selection.|  ',
                           fatal,true)
                 else i := i + 1;
                 k := listposition(name,true,vernames,wasinserted);
                 partial[i-1] := k;
	       end;
           end
          else leaveloop := true;
        end;
     readvarlist(partialedset);
     partialedon.length := 0;
     for i := 1 to max do 
       if i in partialedset then
          begin
             partialedon.length := partialedon.length + 1;
             partialedon.arr[partialedon.length] := i;
          end;
     held := calcindep1(partial,aprob,g2,partialedon,numchecked);
     writeln('Prob = ',aprob:6:2,' g2 = ',g2:6:2);
end;

procedure getindep2;
var partial: partialpair; partialedon: partialedonarray; aval,aprob,g2: real; numchecked: integer;
    held,ok: boolean; partialedset: vertexset; wasinserted,leaveloop: boolean; i,k: integer;
    name: vertexnametype;
begin
     write(output,'Variables: ');
     readttyline(buf);
     i := 1;
     leaveloop := false;
     while (i < 3) and (not leaveloop) do
       begin
         readwhitespace;
         if (not myeoln) then
           begin
             readwhitespace;
             if (not myeoln) then
	       begin
                 readname(name,fatal);
                 if name[1] in ['A'..'Z'] then
   	           myerror('No theoreticals allowed in selection.|  ',
                           fatal,true)
                 else i := i + 1;
                 k := listposition(name,true,vernames,wasinserted);
                 partial[i-1] := k;
	       end;
           end
          else leaveloop := true;
        end;
     readvarlist(partialedset);
     partialedon.length := 0;
     for i := 1 to max do 
       if i in partialedset then
          begin
             partialedon.length := partialedon.length + 1;
             partialedon.arr[partialedon.length] := i;
          end;
      aval := calcpart1(partial,partialedon,ok);
      if not ok then held := false
      else calcpartialheld(aval,aprob,held,partialedon.length);
      writeln('Prob = ',aprob:6:2,' partial = ',aval:6:2);
end;

procedure getpag(var ugraph: graphrec);
var pagfile: text; pagfilename: inbuf; var changed: boolean; i,j: vertextype; ok: boolean;
begin
   for i := 1 to max do 
     for j := 1 to max do
       ugraph.augraph^[i,j] := [];
   getknowledge;
   calcdpath(ugraph,ok);
   
end;

procedure setedge(var ugraph: graphrec);
var i,j: vertextype;
begin
   write('vertices: ');
   readln(i,j);
   ugraph.augraph^[i,j] := [];
   ugraph.augraph^[j,i] := [];
end;


procedure selectvars;
var threshold: real; missingval,nummissing: array[-1..maxvertex] of integer; thevars: vertexset; i,j: integer;
    maxmissing: vertextype; dat: real; missing: real; totmissing: integer; foundone: boolean;
begin
    threshold := getnumber(0,1,false,'threshold |                    ',threshold);
    missing := round(getnumber(1,maxcategories,false,'missing |',missing));
    thevars := [1..max]; 
    for i := 1 to max do missingval[i] := -1;
    for i := 1 to max do
       for j := 1 to maxcategories do
          if (cats[i].value[j] = missing) then
              missingval[i] := j;
    repeat
       maxmissing := 1;
       totmissing := 0;
       for j := 1 to max do nummissing[j] := 0;
       for i := 1 to round(sampsize) do
        begin
         foundone := false;
         for j := 1 to max do
           begin
               if (j in thevars) then 
                    begin
                        dat := getbigdata(i,j);
                        if dat = missingval[j] then 
                           begin
                              nummissing[j] := nummissing[j] + 1;
                              if (not foundone) then
                                  begin
                                      totmissing := totmissing + 1;
                                      foundone := true;
                                  end;
                           end;
                        if (nummissing[j] > nummissing[maxmissing]) then
                            maxmissing := j;
                     end;
           end;
          end;
         thevars := thevars -  [maxmissing];
      until (round(sampsize) - totmissing) > (threshold * round(sampsize));
      for i := 1 to max do
         if (not (i in thevars)) then
            begin
               writename(i,output);
               writeln(output);
            end;
end;
       
(*end.*)
(*-----------#pc-version-----------------------------------------------------*)
(*end.*)
(*-----------#pc-version-----------------------------------------------------*)

procedure setgraph(verset: vertexset; var adjacency: adjtype);
var i,j: vertextype;
begin
  for i := 1 to max do 
     for j := 1 to max do
        adjacency^[i,j] := notthere;
  graphvertices := [];
  numobgraph := 0;
  for i := 1 to max do
     if (i in verset) then
        begin
           graphvertices := graphvertices + [i];
           numobgraph := numobgraph + 1;
           graphlist[numobgraph] := i;
        end;
end;

procedure findy;
var i,j: vertextype; aprob,aprob1,aprob2,aprob3,aprob4,g2,highsig,lowsig: real; held: boolean; 
    edgelist: edgetovertarr; ugraph: graphrec; ok,bic: boolean; oldgraphvert: vertexset;
    partialedon: partialedonarray; partial: partialpair; highcorr: tempset; curindex,k: integer;
begin
   initrec(ugraph);
   tnewadj(tempadj,fatal);
   highsig := getnumber(0,1,false,'Highsig = |',0.5);
   lowsig := getnumber(0,1,false,'Lowsig = |',0.005);
   bic := getreply('BIC score? |','y',false);    
   for i := 1 to max do
     for j := 1 to max do
       ugraph.augraph^[i,j] := [];
   partialedon.length := 0;
   for i := 1 to numobgraph - 1 do highcorr[graphlist[i]] := [];
   curindex := 0;
   for i := 1 to numobgraph - 1 do
      for j := i + 1 to numobgraph do
         begin
            partial[1] := graphlist[i]; partial[2] := graphlist[j]; 
            if (not discrete) and (not indepdata) then 
                 calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob,held,0)
            else if (not indepdata) then
              held := calcindep1(partial,aprob,g2,partialedon,numchecked)
            else held := checkindeplist(partial,partialedon);
            if aprob > highsig then 
               begin
		   edgelist[curindex].tail := graphlist[i];
		   edgelist[curindex].head := graphlist[j];
                   curindex := curindex + 1;
               end
            else if aprob < lowsig then 
               begin
		  highcorr[graphlist[i]] := highcorr[graphlist[i]] + [graphlist[j]];
		  highcorr[graphlist[j]] := highcorr[graphlist[j]] + [graphlist[i]];
               end;
            (* if it holds and is not required remove edge *)
            if (held and
                (adjacency^[graphlist[i],graphlist[j]] = notthere) and 
                (adjacency^[graphlist[j],graphlist[i]] = notthere)) then
               begin
                  ugraph.augraph^[graphlist[i],graphlist[j]] := [];
                  ugraph.augraph^[graphlist[j],graphlist[i]] := [];
                  numedges := numedges - 1;
                  if not discrete then
                     printpartial(partial,partialedon,ofilvar,aprob,rho^[graphlist[i],graphlist[j]])
                  else printpartial(partial,partialedon,ofilvar,aprob,g2);
                end;
          end;
   for k := 1 to curindex - 1 do
      for i := 1 to numobgraph-1 do
        if (graphlist[i] in highcorr[edgelist[k].tail]) and 
           (graphlist[i] in highcorr[edgelist[k].head]) then
              begin
                partial[1] := edgelist[k].tail;
                partial[2] := edgelist[k] .head;
                partialedon.length := 1;
                partialedon.arr[1] := graphlist[i];
                if (not discrete) and (not indepdata) then 
                   calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob,held,0)
                else if (not indepdata) then
                   held := calcindep1(partial,aprob,g2,partialedon,numchecked)
                else held := checkindeplist(partial,partialedon);
                if (aprob < signif) then
                      begin
                         oldgraphvert := graphvertices;
                         setgraph([edgelist[k].tail,edgelist[k].head,graphlist[i]],adjacency);
                         constructgraphs(tempadj,bic);
                         graphvertices := oldgraphvert; 
                         globalindex := 0;
                         setgraphlist;
                      end;
                if (tempadj^[edgelist[k].tail,graphlist[i]] <> notthere) and
                   (tempadj^[edgelist[k].head,graphlist[i]] <> notthere) and
                   (tempadj^[edgelist[k].tail,edgelist[k].head] = notthere) and
                   (tempadj^[edgelist[k].head,edgelist[k].tail] = notthere) then
                   begin
		      writeln(output,'V structure');
		      writename(edgelist[k].head,output);
		      twrite(output,'  |');
		      writename(edgelist[k].tail,output);
		      twriteeoln(output);
		      writename(graphlist[i],output);
 		      twriteeoln(output);
 		      twriteeoln(output);
                      for j := i+1 to numobgraph do
                       if (graphlist[j] in highcorr[edgelist[k].tail]) and 
                          (graphlist[j] in highcorr[edgelist[k].head]) and
                          (graphlist[j] in highcorr[graphlist[i]]) then
				begin
				   partialedon.length := 1;
				   partialedon.arr[1] := graphlist[i];
				   partial[1] := graphlist[j];
				   partial[2] := edgelist[k].tail;
		                   if (not discrete) and (not indepdata) then 
			              calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob1,held,0)
		                   else if (not indepdata) then
			              held := calcindep1(partial,aprob1,g2,partialedon,numchecked)
		                   else held := checkindeplist(partial,partialedon);

				   partialedon.arr[1] := graphlist[j];
				   partial[1] := graphlist[i];
		                   if (not discrete) and (not indepdata) then 
			              calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob2,held,0)
		                   else if (not indepdata) then
			              held := calcindep1(partial,aprob2,g2,partialedon,numchecked)
		                   else held := checkindeplist(partial,partialedon);

				   partial[2] := edgelist[k].head;
		                   if (not discrete) and (not indepdata) then 
			              calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob3,held,0)
		                   else if (not indepdata) then
			              held := calcindep1(partial,aprob3,g2,partialedon,numchecked)
		                   else held := checkindeplist(partial,partialedon);

				   partialedon.arr[1] := graphlist[i];
				   partial[1] := graphlist[j];
		                   if (not discrete) and (not indepdata) then 
			              calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob4,held,0)
		                   else if (not indepdata) then
			              held := calcindep1(partial,aprob4,g2,partialedon,numchecked)
		                   else held := checkindeplist(partial,partialedon);

			           if ((aprob1 > highsig) and (aprob4 > highsig) and 
                                        (aprob2 < lowsig) and (aprob3 < lowsig)) then
					  begin
					     writename(graphlist[i],output);
					     twrite(output,' really causes |');
					     writename(graphlist[j],output);
					     twriteeoln(output);
					     writename(edgelist[k].head,output);
					     twrite(output,'  |');
					     writename(edgelist[k].tail,output);
					     twriteeoln(output);
					  end;
				   
                                   if ((aprob1 < highsig) or (aprob4 < highsig)) and
				      ((aprob2 < lowsig) and (aprob3 < lowsig)) then
				       begin
				             partialedon.length := 2;
				             partialedon.arr[2] := edgelist[k].tail;
				             if (not discrete) and (not indepdata) then 
			                          calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob4,held,0)
		                             else if (not indepdata) then
			                         held := calcindep1(partial,aprob4,g2,partialedon,numchecked)
				             else held := checkindeplist(partial,partialedon);
				             if (aprob4 > highsig) then
				                begin
    					           writename(graphlist[i],output);
					           twrite(output,' causes |');
					           writename(graphlist[j],output);
					           twriteeoln(output);
					           writename(edgelist[k].head,output);
					           twrite(output,'  |');
					           writename(edgelist[k].tail,output);
					           twriteeoln(output);
				                 end;
				             partialedon.arr[2] := edgelist[k].head;
				             partial[2] := edgelist[k].tail;
				             if (not discrete) and (not indepdata) then 
			                          calcpartialheld(rho^[graphlist[i],graphlist[j]],aprob4,held,0)
		                             else if (not indepdata) then
			                         held := calcindep1(partial,aprob4,g2,partialedon,numchecked)
				             else held := checkindeplist(partial,partialedon);
				             if (aprob4 > highsig) then
				                begin
					           writename(graphlist[i],output);
					           twrite(output,' causes |');
					           writename(graphlist[j],output);
					           twriteeoln(output);
					           writename(edgelist[k].head,output);
					           twrite(output,'  |');
					           writename(edgelist[k].tail,output);
					           twriteeoln(output);
				                end;
					  end;
                                end;
       end;
    end;
  dispose(tempadj);
  dispose(ugraph.ancest);
  dispose(ugraph.augraph);
  dispose(ugraph.recchange);
end;

